{*************************************************************}
{Regrouper les principales fonctions de la bibliothque PROB20}
{*************************************************************}

UNIT LOISTAT;

INTERFACE

{Inverse}
FUNCTION ILS_CHI2(prb,df1: Extended): Extended;
FUNCTION ILS_NORMAL(prb: Extended): Extended;
FUNCTION ILS_FISCHER(prb,df1,df2: Extended): Extended;
FUNCTION ILS_STUDENT(prb,df1: Extended): Extended;
{Valeur critique -- new -- 05/05/2007 -- masques pour ne pas gner FMATH}
(*
FUNCTION LS_CHI2(x,df1: Extended): Extended;
FUNCTION LS_NORMAL(x: Extended): Extended;
FUNCTION LS_FISCHER(x,df1,df2: Extended): Extended;
FUNCTION LS_STUDENT(x,df1: Extended): Extended;
*)

IMPLEMENTATION

uses acrcy,dlgama_u,ppchi2_u,ppnd_u,xinbta_u,xap_u,pax_u,
     SYSUTILS;

const
  tol = 1.0e-7;
  ptest = 1.0e-5;

{************************************************************************************************}
{***************************** Les calculs inverses *********************************************}
{************************************************************************************************}

{Fonction inverse du Chi-2}
{prb: proba critique}
{df1: degree of freedom}
FUNCTION ILS_CHI2(prb,df1: Extended): Extended;
Var Tmp: Extended;
    ifault: Integer;
Begin
 Tmp:= Pax(1,ifault,prb,df1,0);
 If (ifault<>0)
  Then ILS_CHI2:= -1
  Else ILS_CHI2:= Tmp;
End;


{Inverse Fischer}
{prb: proba critique}
{df1,df2: degree of freedom}
FUNCTION ILS_FISCHER(prb,df1,df2: Extended): Extended;
Var Tmp: Extended;
    ifault: Integer;
Begin
 Tmp:= Pax(2,ifault,prb,df1,df2);
 If (ifault<>0)
  Then ILS_FISCHER:= -1
  Else ILS_FISCHER:= Tmp;
End;


{Inverse du Student}
{Prb: proba critique}
{df1: degree of freedom}
FUNCTION ILS_STUDENT(prb,df1: Extended): Extended;
Var Tmp: Extended;
    ifault: Integer;
Begin
 Tmp:= Pax(3,ifault,prb,df1,0);
 If (ifault<>0)
  Then ILS_STUDENT:= -1E37
  Else ILS_STUDENT:= Tmp;
End;

{Inverse de la fonction de rpartition normale}
{prb: proba critique}
FUNCTION ILS_NORMAL(prb: Extended): Extended;
Var ifault: Integer;
    Tmp: Extended;
Begin
 ifault := 0;
 Tmp := ppnd(prb,ifault);
 if(ifault <> 0)
  Then ILS_NORMAL:= -1E37
  Else ILS_NORMAL:= Tmp;
End;


{************************************************************************************************}
{***************************** Les calculs des proba ********************************************}
{************************************************************************************************}

{Chi2}
FUNCTION LS_CHI2(x,df1: Extended): Extended;
Var ifault: Integer;
    p: Extended;
Begin
 p:= xap(1,ifault,x,df1,0);
 if (ifault<>0)
  Then LS_CHI2:= -1
  Else LS_CHI2:= p;
End;

{Normal}
FUNCTION LS_NORMAL(x: Extended): Extended;
Var ifault: Integer;
    p: Extended;
Begin
 p:= xap(4,ifault,x,0,0);
 if (ifault<>0)
  Then LS_NORMAL:= -1
  Else LS_NORMAL:= p;
End;

{Fischer}
FUNCTION LS_FISCHER(x,df1,df2: Extended): Extended;
Var ifault: Integer;
    p: Extended;
Begin
 p:= xap(2,ifault,x,df1,df2);
 if (ifault<>0)
  Then LS_FISCHER:= -1
  Else LS_FISCHER:= p;
End;

{Student}
FUNCTION LS_STUDENT(x,df1: Extended): Extended;
Var ifault: Integer;
    p: Extended;
Begin
 p:= xap(3,ifault,x,df1,0);
 if (ifault<>0)
  Then LS_STUDENT:= -1
  Else LS_STUDENT:= p;
End;


END.