(**********************************************************************)
(* UCalcStatDesCrossTab3D.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Tableaux croiss, avec les indicateurs stats associs)
@author(Ricco)
@created(18/05/2007)

Tableau croiss  3 dim. pour le calcul des associations partielles.
Rf. Ritschard, Kellerhals, Olszak et Sardi -- 1994. (Cahiers du Dpartement d'Economtrie)
Voir aussi publication des mmes auteurs dans "The Statisticians"

Seule la version partielle des (U)ncertainty Coef. est programm -- 18/05/2007 

}
unit UCalcStatDesCrossTab3D;

interface

USES
    Contnrs,
    UCalcDistribution,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcStatDes,
    UCalcCrossTab,
    UCalcRndGenerator;

TYPE
    TTab3D = array of array of array of TFreqCountType;

    TCalcCrossTab3D = class(TCalcStatDes)
    private
    //attributs
    A,B,E: TAttribute;
    //tableau de comptage
    FTab: TTab3D;
    //U(ab) et U(ab/e)
    FU,FUp: double;
    //ecart-type
    FSigma: double;
    //liste des rusltats par modalit de la variable de contrle
    FLstU: TObjectList;
    procedure   calcStats();
    function    getValue(i,j,k: integer): TFreqCountType;
    function    getProba(i,j,k: integer): double;
    protected
    procedure   BeginUpdate(); override;
    procedure   AddValue(prmExample: integer); override;
    procedure   EndUpdate(); override;
    public
    //mthodes
    constructor Create(prmA,prmB, prmE: TAttribute; prmExamples: TExamples = nil);
    destructor  Destroy(); override;
    function    getHTMLResult(prmOption: integer = -1): string; override;
    procedure   RefreshStat(prmExamples: TExamples); override;
    //proprits
    property    n[i,j,k: integer]: TFreqCountType read getValue;
    property    p[i,j,k: integer]: double read getProba;
    end;

    TLstCalcStatDesCrossTab3D = class(TLstCalcStatDes)
    public
    {description HTML des rsultats}
    function    getHeaderHTML(): string; override;
    {comparaison pour les stats conditionnelles}
    procedure SortStats(); override;
    end;

implementation

USES
    Math, Classes, UConstConfiguration, SysUtils, UCalcStatDesCrossTab;

{ TCalcCrossTab3D }

procedure TCalcCrossTab3D.AddValue(prmExample: integer);
var i,j,k: integer;
begin
 //identifier la case
 i:= A.dValue[prmExample];
 j:= B.dValue[prmExample];
 k:= E.dValue[prmExample];
 //ajouter
 FTab[i][j][k]:= FTab[i][j][k] + 1;
 //sans oublier les marges cond.
 FTab[i][j][0]:= FTab[i][j][0] + 1;
 FTab[i][0][k]:= FTab[i][0][k] + 1;
 FTab[0][j][k]:= FTab[0][j][k] + 1;
 //et doublement cond.
 FTab[i][0][0]:= FTab[i][0][0] + 1;
 FTab[0][j][0]:= FTab[0][j][0] + 1;
 FTab[0][0][k]:= FTab[0][0][k] + 1;
 //et globale
 FTab[0][0][0]:= FTab[0][0][0] + 1;
end;

procedure TCalcCrossTab3D.BeginUpdate;
var i,j,k: integer;
begin
 for i:= 0 to A.nbValues do
  for j:= 0 to B.nbValues do
   for k:= 0 to E.nbValues do
    FTab[i][j][k]:= 0;
end;

procedure TCalcCrossTab3D.calcStats;
var s: double;
    i,j,k: integer;
    delta,pPhi,phi,phiBar: double;
begin
 //calculer le U marginal, q.6
 FU:= 0.0;
 for i:= 1 to A.nbValues do
  for j:= 1 to B.nbValues do
   if (p[i,j,0] > 0) and (p[i,0,0] > 0) and (p[0,j,0] > 0)
    then FU:= FU + p[i,j,0] * math.Log2(p[i,0,0]*p[0,j,0]/p[i,j,0]);
 s:= 0.0;
 if (FU <> 0)
  then for i:= 1 to A.nbValues do
   if (p[i,0,0] > 0) then s:= s + p[i,0,0] * math.Log2(p[i,0,0]);
 if (s <> 0) then FU:= FU / s;
 //le U partiel, q.20
 FUp:= 0.0;
 for k:= 1 to E.nbValues do
  for i:= 1 to A.nbValues do
   for j:= 1 to B.nbValues do
    if (p[i,j,k] > 0) and (p[i,0,k] > 0) and (p[0,j,k] > 0) and (p[0,0,k] > 0)
     then FUp:= FUp + p[i,j,k] * math.Log2((p[i,0,k]*p[0,j,k])/(p[i,j,k]*p[0,0,k]));
 pPhi:= FUp;//pour le calcul de la variance
 s:= 0.0;
 if (FUp <> 0)
  then
   for k:= 1 to E.nbValues do
    for i:= 1 to A.nbValues do
     if (p[i,0,k] > 0) and (p[0,0,k] > 0)
      then s:= s + p[i,0,k] * math.Log2(p[i,0,k]/p[0,0,k]);
 delta:= s;
 if (s <> 0) then FUp:= FUp / s;
 //calculer la variance, eq.48 etc.
 phiBar:= 0.0;
 for k:= 1 to E.nbValues do
  for j:= 1 to B.nbValues do
   for i:= 1 to A.nbValues do
    if (p[i,j,k] > 0) and (p[i,0,k] > 0) and (p[0,0,k] > 0) and (p[0,j,k] > 0)
     then
      begin
       phi:= math.Log2(p[i,0,k]/p[0,0,k]) * pPhi - delta * math.Log2((p[i,0,k]*p[0,j,k])/(p[i,j,k]*p[0,0,k]));
       phiBar:= phiBar + p[i,j,k] * phi;
      end;
 FSigma:= 0.0;
 for k:= 1 to E.nbValues do
  for j:= 1 to B.nbValues do
   for i:= 1 to A.nbValues do
    if (p[i,j,k] > 0) and (p[i,0,k] > 0) and (p[0,0,k] > 0) and (p[0,j,k] > 0)
     then
      begin
       phi:= math.Log2(p[i,0,k]/p[0,0,k]) * pPhi - delta * math.Log2((p[i,0,k]*p[0,j,k])/(p[i,j,k]*p[0,0,k]));
       FSigma:= FSigma + p[i,j,k] * math.Power(phi-phiBar,2.0);
      end;

 if (delta <> 0)
  then FSigma:= FSigma / (1.0 * n[0,0,0] * math.Power(delta,4.0))
  else FSigma:= 0.0;
 if (FSigma > 0) then FSigma:= sqrt(FSigma);

end;

constructor TCalcCrossTab3D.Create(prmA, prmB, prmE: TAttribute;
  prmExamples: TExamples);
begin
 A:= prmA;
 B:= prmB;
 E:= prmE;
 setLength(FTab,A.nbValues+1,B.nbValues+1,E.nbValues+1);
 inherited create(prmA,prmExamples);
 FLstU:= TObjectList.Create(true);
end;

destructor TCalcCrossTab3D.Destroy;
begin
 FLstU.Free();
 finalize(FTab);
 inherited;
end;

procedure TCalcCrossTab3D.EndUpdate;
begin
 self.calcStats();
end;

const
    VAL_CRIT_LOI_NORMALE_5pc : double = 1.95996108232;

function TCalcCrossTab3D.getHTMLResult(prmOption: integer): string;
var s: string;
    k: integer;
    stat: TCalcSDCrossTab;
begin
 //former le tableau intermdiaire
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>U(AB)</TH><TH>U(AB / E)</TH><TH>Sigma</TH><TH>95% C.I.</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD %s align="right">%.6f</TD><TD %s align="right">%.6f</TD><TD align="right">%.6f</TD><TD align="right">%.4f ; %.4f</TD></TR>',
                                               [HTML_BGCOLOR_DATA_GREEN,FU,HTML_BGCOLOR_DATA_GREEN,FUp,FSigma,FUp-VAL_CRIT_LOI_NORMALE_5pc*FSigma,FUp+VAL_CRIT_LOI_NORMALE_5pc*FSigma]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH colspan=4>Details</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Value</TH><TH>U(AB / E=k)</TH><TH>Sigma</TH><TH>95% C.I.</TH></TR>';
 for k:= 0 to pred(FLstU.Count) do
  begin
   stat:= FLstU[k] as TCalcSDCrossTab;
   s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>E=%s</TD><TD align="right">%.4f</TD><TD align="right">%.4f</TD><TD align="right">%.4f ; %.4f</TD></TR>',
                                               [E.LstValues.getDescription(succ(k)),stat.U,stat.SigmaU,stat.U-VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaU,stat.U+VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaU]);
  end;
 s:= s + '</table>';
 //et envoyer la sauce...
 result:= format('<TD>%s</TD><TD>%s</TD><TD>%s</TD><TD align="right">%s</TD>',[A.Name,B.name,E.Name,s]);
end;

function TCalcCrossTab3D.getProba(i, j, k: integer): double;
begin
 result:= (1.0 * FTab[i][j][k])/(1.0 * FTab[0][0][0]);
end;

function TCalcCrossTab3D.getValue(i, j, k: integer): TFreqCountType;
begin
 result:= FTab[i][j][k];
end;

procedure TCalcCrossTab3D.RefreshStat(prmExamples: TExamples);
var lstExamples: TObjectList;
    k: integer;
    stat: TCalcSDCrossTab;
begin
 inherited;//pour dclencher les calculs standards
 //initier maintenant les calculs par modalit de la variable de contrle
 FLstU.Clear();
 lstExamples:= prmExamples.DispatchExamples(E);
 for k:= 0 to pred(lstExamples.Count) do
  begin
   stat:= TCalcSDCrossTab.Create(A,B,lstExamples[k] as TExamples);
   FLstU.Add(stat);
  end;
 lstExamples.Free();
end;

{ TLstCalcStatDesCrossTab3D }

function TLstCalcStatDesCrossTab3D.getHeaderHTML: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY +
          '<TH>Target. A</TH><TH>Input. B</TH><TH>Control. E</TH><TH>Partial U(AB / E)</TH>';  
end;

{trier selon le nom des attributs en ligne}
function ListSortCompareRowAttName(item1,item2: pointer): integer;
var st1,st2: TCalcCrossTab3D;
begin
 st1:= TCalcCrossTab3D(item1);
 st2:= TCalcCrossTab3D(item2);
 if (st1.A.Name<st2.A.Name)
  then result:= -1
  else
   if (st1.A.Name>st2.A.Name)
    then result:= +1
    else result:= 0;
end;

{trier selon le nom des attributs en colonne}
function ListSortCompareColumnAttName(item1,item2: pointer): integer;
var st1,st2: TCalcCrossTab3D;
begin
 st1:= TCalcCrossTab3D(item1);
 st2:= TCalcCrossTab3D(item2);
 if (st1.B.Name<st2.B.Name)
  then result:= -1
  else
   if (st1.B.Name>st2.B.Name)
    then result:= +1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortCompareTheil(item1,item2: pointer): integer;
var st1,st2: TCalcCrossTab3D;
begin
 st1:= TCalcCrossTab3D(item1);
 st2:= TCalcCrossTab3D(item2);
 //tri invers !!!
 if (st1.FU<st2.FU)
  then result:= +1
  else
   if (st1.FU>st2.FU)
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortComparePartialTheil(item1,item2: pointer): integer;
var st1,st2: TCalcCrossTab3D;
begin
 st1:= TCalcCrossTab3D(item1);
 st2:= TCalcCrossTab3D(item2);
 //tri invers !!!
 if (st1.FUp<st2.FUp)
  then result:= +1
  else
   if (st1.FUp>st2.FUp)
    then result:= -1
    else result:= 0;
end;

procedure TLstCalcStatDesCrossTab3D.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareTheil;
     3: funcCompare:= ListSortComparePartialTheil;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

end.
