(******************************************************************************)
(* UCompAssocRuleOutlierDetection.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(******************************************************************************)

{
@abstract(Dtection des individus atypiques  partir des rgles d'association)
@author(Ricco)
@created(13/06/2007)

L'ide est simple : dtecter les rgles comportant un nombre "trs faible" de contre-exemples.

Pourquoi ?

Si pas de contre-exemples => dpendance fonctionnelle
Si quelques contre-exemples => rgle statistique (avec un marge d'erreur dans la prcision)

Si trs peu de contre-exemples => ne sont-ce pas des anomalies ?
}
unit UCompAssocRuleOutlierDetection;

interface

USES
        Forms, Classes, IniFiles,
        UCompDefinition,
        UOperatorDefinition,
        UCompAssociationRuleDefinition,
        UDatasetExamples,
        UCalcAssocStructure;

TYPE
        {gnrateur de composant}
        TGenCompAssocOutlier = class(TMLGenComp)
                               protected
                               procedure   GenCompInitializations(); override;
                               public
                               function    GetClassMLComponent: TClassMLComponent; override;
                               end;

        {composant}
        TMLCompAssocOutlier = class(TMLCompAssocRule)
                              protected
                              function    getClassOperator: TClassOperator; override;   
                              end;

        {oprateur}
        TOpAssocOutlier = class(TOpAssocRule)
                          protected
                          function    getClassParameter: TClassOperatorParameter; override;
                          function    getClassAssocRule(): TClassCalcAssocRule; override;
                          end;

        {paramtre de l'oprateur}
        TOpPrmAssocOutlier = class(TOpPrmAssocRule)
                             protected
                             function    CreateDlgParameters(): TForm; override;
                             procedure   SetDefaultParameters(); override;
                             public
                             //nombre de contre-exemples pour dcider que c'est un outlier
                             NbOutliers: integer;
                             //description HTML
                             function    getHTMLParameters(): string; override;
                             procedure   LoadFromStream(prmStream: TStream); override;
                             procedure   SaveToStream(prmStream: TStream); override;
                             procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                             procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                             end;

        {calculateur}
        TCalcAssocOutlier = class(TCalcAssocRule)
                            protected
                            function    classRuleStructure(): TClassAssocRuleStructure; override;
                            function    classLstAssocRules(): TClassLstAssocRules; override;
                            public
                            procedure   buildRules(prmExamples: TExamples); override;    
                            end;

implementation

uses
    UConstConfiguration, Sysutils, UCalcAssocStrucOutlier,
  UDlgOpPrmAssocRuleOutlier;

{ TGenCompAssocOutlier }

procedure TGenCompAssocOutlier.GenCompInitializations;
begin
 FMLComp:= mlcAssociation;
end;

function TGenCompAssocOutlier.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompAssocOutlier; 
end;

{ TMLCompAssocOutlier }

function TMLCompAssocOutlier.getClassOperator: TClassOperator;
begin
 result:= TOpAssocOutlier; 
end;

{ TOpAssocOutlier }

function TOpAssocOutlier.getClassAssocRule: TClassCalcAssocRule;
begin
 result:= TCalcAssocOutlier;
end;

function TOpAssocOutlier.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmAssocOutlier; 
end;

{ TOpPrmAssocOutlier }

function TOpPrmAssocOutlier.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmAssocOutlier.CreateFromOpPrm(self);
end;

function TOpPrmAssocOutlier.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>A-Priori parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Support min</TD><TD align="right">%.2f</TD></TR>',[MinSupport]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Confidence min</TD><TD align="right">%.2f</TD></TR>',[MinConfidence]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Max rule length</TD><TD align="right">%d</TD></TR>',[MaxRuleLength]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Lift filtering</TD><TD align="right">%.2f</TD></TR>',[MinLift]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Nb Outliers</TD><TD align="right">%d</TD></TR>',[NbOutliers]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmAssocOutlier.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 NbOutliers:= prmINI.ReadInteger(prmSection,'nb_outliers',NbOutliers);
end;

procedure TOpPrmAssocOutlier.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(NbOutliers,sizeof(NbOutliers));
end;

procedure TOpPrmAssocOutlier.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'nb_outliers',NbOutliers);
end;

procedure TOpPrmAssocOutlier.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(NbOutliers,sizeof(NbOutliers));
end;

procedure TOpPrmAssocOutlier.SetDefaultParameters;
begin
 inherited;
 NbOutliers:= 3;
end;

{ TCalcAssocOutlier }

procedure TCalcAssocOutlier.buildRules(prmExamples: TExamples);
var seuil: integer;
    i: integer;
    rule: TAssocRuleStructureOutlier;
begin
 //construire les rgles normalement
 inherited;
 //post-filtrer les rgles sur la base des outliers
 seuil:= (self.prmAssocRule as TOpPrmAssocOutlier).NbOutliers;
 for i:= pred(self.rules.count) downto 0 do
  begin
   rule:= self.rules.getRule(i) as TAssocRuleStructureOutlier;
   //on supprime la rgle s'il n'y a pas de contre-exemple ou s'ils sont trop nombreux
   if (rule.nANotC = 0) or (rule.nANotC > seuil)
    then self.rules.deleteRule(i);
  end;
end;

function TCalcAssocOutlier.classLstAssocRules: TClassLstAssocRules;
begin
 result:= TLstAssocRulesOutlier;
end;

function TCalcAssocOutlier.classRuleStructure: TClassAssocRuleStructure;
begin
 result:= TAssocRuleStructureOutlier;
end;

initialization
 RegisterClass(TGenCompAssocOutlier);
end.
