(*****************************************************************)
(* UCompClusteringEM.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(Implmentation des EM -- Trs simple)
@author(Ricco)
@created(26/08/2006)

Lance un K-Means pour initialiser les centres de classes, puis optimise la vraisemblance.
La rfrence utilise est la suivante : http://fr.wikipedia.org/wiki/Algorithme_esp%C3%A9rance-maximisation

}
unit UCompClusteringEM;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UOperatorDefinition,
        UCompClusteringDefinition,
        UDatasetDefinition,
        UDatasetExamples,
        UCalcStatDes,
        contnrs,
        D6_MATRICES, D6_FMATH,
        UCompClusteringKMeans,
        UDatasetImplementation;

TYPE
        {gnrateur}
        TGenClusEM = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant}
        TMLCompClusEM = class(TMLCompClustering)
        protected
        function    getClassOperator: TClassOperator; override;
        function    getGenericAttName(): string; override;
        end;

        {oprateur}
        TOpClusEM = class(TOperatorClusteringContinue)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassCalcClustering(): TClassCalcClustering; override;
        end;

        {paramtre oprateur}
        TOpPrmClusEM = class(TOpPrmClustering)
        private
        {nombre de clusters dsir}
        FNbClusters: integer;
        {standardisation des donnes pour le K-Means initial : 0 : rien ; 1 : par la variance}
        FStandardization: integer;
        {nombre d'itrations maximum sur la convergence}
        FMaxIter: integer;
        {ridge coef.}
        FRidgeCorrection: double;
        protected
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        function    getHTMLParameters(): string; override;
        property    NbClusters: integer read FNbClusters write FNbClusters;
        property    MaxIter: integer read FMaxIter write FMaxIter;
        property    RidgeCorrection: double read FRidgeCorrection write FRidgeCorrection;
        property    Standardization: integer read FStandardization write FStandardization;
        end;

        {One cluster EM}
        TOneClusterEM = class
        private
        //identifiant du cluster
        FClusterId: TTypeDiscrete;
        //liste des attributs
        FAtts: TLstAttributes;
        //vecteur du centre de gravit
        FAvg: TVector;
        //matrice de variance co-variance -- qu'il faut inverser par la suite
        FVCV: TMatrix;
        //inverse de la matrice de variance co-variance
        FInvVCV: TMatrix;
        //dterminant de la matrice de variance-covar
        FDetVCV: double;
        //poids du cluster
        FPoids: double;
        public
        //constructeur -- prparer les structures internes
        constructor create(prmAtts: TLstAttributes; prmClusterId: TTypeDiscrete);
        //dtruire
        destructor  destroy(); override;
        //mj des vecteurs et matrices -- true si calcul ok.
        function    update(prmExamples: TExamples; prmAppartenance: TMatrix; ridgeCoef: double = 0.0): boolean;
        //calcul du log de la fonction de densit pour une observation
        function    logDensity(example: integer): double;
        end;

        TCalcClusEM = class;

        {structure de clusters}
        TStrucClusterEM = class
        private
        //tableau d'appartenance aux diffrents clusters
        FAppartenance: TMatrix;
        //liste des clusters
        FLstClusters: TObjectList;
        //oprateur de calcul des K-Means
        FCalcEM: TCalcClusEM;
        public
        //calcul avec les clusters de KMeans
        constructor  create(prmCalcClus: TCalcClusEM);
        //destructor
        destructor   destroy(); override;
        //initialisation des clusters
        function     initialize(prmExamples: TExamples): boolean;
        //mj des param. des clusters
        function     updateParameters(prmExamples: TExamples): boolean;
        //mj de l'appartenance aux groupes
        function     updateGroupMembership(prmExamples: TExamples): boolean;
        //calcul de la log-vraisemblance Q de calcul (sur les donnes de calcul)
        function     workingLQ(prmExamples: TExamples): double;
        //nombre de clusters
        function     getNbClusters(): integer;
        end;

        {oprateur de calcul}
        TCalcClusEM = class(TCalcClusteringContinue)
        private
        //paramtre du K-Means
        FPrmKMeans: TOpPrmClusKMeans;
        //oprateur de calcul K-Means
        FCalcKMeans: TCalcClusKMeans;
        //structure de EM
        FStrucEM: TStrucClusterEM;
        //log-vraisemblance max. de calcul
        FLQMax: double;
        //log-vraisemblance max. sur ensemble  part (qui est l'ensemble d'apprentissage ici)
        FLLkMax, FAIC, FBIC: double;
        protected
        function    getHTMLClustering(): string; override;
        public
        //on en profite pour construire le paramtre de calcul K-Means
        constructor Create(prmTargets, prmInputs: TLstAttributes; OpPrm: TOpPrmClustering; prmAttClus: TAttDiscrete); override;
        //dtruire
        destructor  destroy(); override;
        //calculer les clusters
        procedure   BuildClusters(prmExamples: TExamples); override;
        //modalits de la variable cre
        procedure   FillClusAttDef(); override;
        //affecter un individu  un cluster
        function    SetClusterExample(example: integer): TTypeDiscrete; override;
        //calcul de la log-vraisemblance sur un ensemble test (on n'utilise plus le tableau d'appartenance) et les autres critres
        function    logLikelihood(prmTestSet: TExamples): double;
        //calculer l'AIC  partir de la LLK
        function    getAIC(nbClus: integer; LLK: double; n: double): double;
        //calculer le BIC  partir de LLK
        function    getBIC(nbClus: integer; LLK: double; n: double): double;
        //calcul du nombre de paramtres  estimer
        function    getNbParameters(nbClus: integer): integer;
        //nombre de cluster
        function    getNbClus(): integer;
        end;


implementation

uses
    Math,
    Sysutils, ULogFile, UConstConfiguration, UCalcRndGenerator,
    UDlgOpPrmClusEM;

Const
    TOLERANCE_FOR_DET_VCV = 1.0e-6;    

{ TGenClusEM }

procedure TGenClusEM.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

function TGenClusEM.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusEM; 
end;

{ TMLCompClusEM }

function TMLCompClusEM.getClassOperator: TClassOperator;
begin
 result:= TOpClusEM;
end;

function TMLCompClusEM.getGenericAttName: string;
begin
 result:= 'EM';
end;

{ TOpClusEM }

function TOpClusEM.getClassCalcClustering: TClassCalcClustering;
begin
 result:= TCalcClusEM;
end;

function TOpClusEM.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmClusEM; 
end;

{ TOpPrmClusEM }

function TOpPrmClusEM.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmClusEM.CreateFromOpPrm(self);
end;

function TOpPrmClusEM.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>EM parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%s</TD><TD align="right">%d</TD></TR>',['Clusters',FNbClusters]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%s</TD><TD align="right">%d</TD></TR>',['Max Iteration',FMaxIter]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%s</TD><TD align="right">%.6f</TD></TR>',['Ridge',FRidgeCorrection]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%s</TD><TD align="right">%s</TD></TR>',['Seed random generator',StartSeedDescription[self.FModeRndGenerator]]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmClusEM.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FNbClusters:= prmINI.ReadInteger(prmSection,'nb_clusters',FNbClusters);
 FMaxIter:= prmINI.ReadInteger(prmSection,'max_iter',FMaxIter);
 FStandardization:= prmINI.ReadInteger(prmSection,'standardization',FStandardization);
 FRidgeCorrection:= prmINI.ReadFloat(prmSection,'ridge',FRidgeCorrection);
end;

procedure TOpPrmClusEM.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FNbClusters,sizeof(FNbClusters));
 prmStream.ReadBuffer(FMaxIter,sizeof(FMaxIter));
 prmStream.ReadBuffer(FStandardization,sizeof(FStandardization));
 prmStream.ReadBuffer(FRidgeCorrection,sizeof(FRidgeCorrection));
end;

procedure TOpPrmClusEM.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'nb_clusters',FNbClusters);
 prmINI.WriteInteger(prmSection,'max_iter',FMaxIter);
 prmINI.WriteInteger(prmSection,'standardization',FStandardization);
 prmINI.WriteFloat(prmSection,'ridge',FRidgeCorrection);
end;

procedure TOpPrmClusEM.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FNbClusters,sizeof(FNbClusters));
 prmStream.WriteBuffer(FMaxIter,sizeof(FMaxIter));
 prmStream.WriteBuffer(FStandardization,sizeof(FStandardization));
 prmStream.WriteBuffer(FRidgeCorrection,sizeof(FRidgeCorrection));
end;

procedure TOpPrmClusEM.SetDefaultParameters;
begin
 inherited;
 FNbClusters:= 2;
 FMaxIter:= 50;
 FStandardization:= 1;
 FRidgeCorrection:= 0.001;
end;

{ TCalcClusEM }

procedure TCalcClusEM.BuildClusters(prmExamples: TExamples);
var iter: integer;
    LQ,LQMax: double;
begin
 //appel de la mthode de l'anctre pour les stat.des.
 inherited;
 TRY
 //oprateur de calcul K-Means
 if assigned(FCalcKMeans) then FCalcKMeans.Free();
 FCalcKMeans:= TCalcClusKMeans.Create(self.Targets,self.Inputs,FPrmKMeans,self.AttCluster);
 //calcul des clusters avec les K-Means
 FCalcKMeans.BuildClusters(prmExamples);
 //structure des EM
 if assigned(FStrucEM) then FStrucEM.Free();
 FStrucEM:= TStrucClusterEM.create(self);
 FStrucEM.initialize(prmExamples);//les paramtres sont calculs ici au premier passage
 //maintenant on itre
 LQMax:= -1.0 * MATH.MaxDouble;
 iter:= 0;
 WHILE (iter < (self.PrmCalc as TOpPrmClusEM).FMaxIter) do
  begin
   //re-calcul de l'appartenance aux groupes -- (E)xpectation
   FStrucEM.updateGroupMembership(prmExamples);
   //recalcul des paramtres -- (M)aximization
   FStrucEM.updateParameters(prmExamples);
   //calcul de la log-vraisemblance
   LQ:= FStrucEM.workingLQ(prmExamples);
   //TraceLog.WriteToLogFile(format('[EM Clustering] :: iteration %d, log-likelihood = %.4f',[iter,LQ]));
   //test si volution
   if (LQ > LQMax)
    then LQMax:= LQ
    //halte-l dans le cas contraire
    else BREAK;
   //on passe au suivant
   inc(iter);
  end;
 //rcuprer la vraisemblance max. utilise dans les calculs
 FLQMax:= LQMax;
 //et la vraisemblance
 FLLkMax:= self.logLikelihood(prmExamples);
 FAIC:= self.getAIC(FStrucEM.getNbClusters(),FLLKMax,1.0*prmExamples.Size);
 FBIC:= self.getBIC(FStrucEM.getNbClusters(),FLLKMax,1.0*prmExamples.Size);
 //dans le log.
 TraceLog.WriteToLogFile(format('[EM Clustering] :: number of iterations = %d, max. Q = %.4f and max. LLK = %.4f',[iter,FLQMax,FLLkMax]));
 EXCEPT
 on e: Exception do TraceLog.WriteToLogFile(format('[ERROR] -- "BuildClusters of TCalcClusEM" : %s',[e.Message]));
 END;
end;

constructor TCalcClusEM.Create(prmTargets, prmInputs: TLstAttributes;
  OpPrm: TOpPrmClustering; prmAttClus: TAttDiscrete);
begin
 inherited;
 //paramtre oprateur K-Means
 FPrmKMeans:= TOpPrmClusKMeans.Create(NIL);
 FPrmKMeans.NbClusters:= (self.PrmCalc as TOpPrmClusEM).NbClusters;
 FPrmKMeans.Normalization:= (self.PrmCalc as TOpPrmClusEM).FStandardization;
 FPrmKMeans.ModeRndGenerator:= (self.PrmCalc as TOpPrmClusEM).ModeRndGenerator;
end;

destructor TCalcClusEM.destroy;
begin
 if assigned(FCalcKMeans) then FCalcKMeans.Free();
 if assigned(FStrucEM) then FStrucEM.Free();
 if assigned(FPrmKMeans) then FPrmKMeans.Free();
 inherited;
end;

procedure TCalcClusEM.FillClusAttDef;
var k: integer;
begin
 if assigned(FAttClus)
  then
   begin
    FattClus.LstValues.clear;
    for k:= 1 to (PrmCalc as TOpPrmClusEM).NbClusters do
     //le fonctionnement particulier de getValue permet cette manip.
     FAttClus.LstValues.getValue('c_em_'+IntToStr(k));
   end;
end;

function TCalcClusEM.getAIC(nbClus: integer; LLK, n: double): double;
var k: double;
    AIC: double;
begin
 //>>nombre de paramtres
 //cf. --> http://www.cs.uoi.gr/~galatsanos/PAPERS/IPL-Conferance_papers/icip03_segmDNA.pdf#search=%22GMM%20clustering%22
 //et cf. --> http://www.uic.edu/classes/idsc/ids472/clustering.htm
 k:= 1.0 * self.getNbParameters(nbClus);
 //http://en.wikipedia.org/wiki/Akaike_Information_Criterion
 AIC:= -2.0 * LLK + 2.0 * k;
 result:= AIC;
end;

function TCalcClusEM.getBIC(nbClus: integer; LLK, n: double): double;
var k: double;
    BIC: double;
begin
 //>>nombre de paramtres
 //cf. --> http://www.cs.uoi.gr/~galatsanos/PAPERS/IPL-Conferance_papers/icip03_segmDNA.pdf#search=%22GMM%20clustering%22
 //et cf. --> http://www.uic.edu/classes/idsc/ids472/clustering.htm
 k:= 1.0 * self.getNbParameters(nbClus);
 //http://en.wikipedia.org/wiki/Bayesian_information_criterion
 BIC:= -2.0 * LLK + k * LN(n);
 //
 result:= BIC;
end;

function TCalcClusEM.getHTMLClustering: string;
var s: string;
    j,k: integer;
begin
 s:= '<H3>Clustering quality criterion</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Criterion</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align="right">%.4f</TD></TR>',['Log-likelihood',FLLkMax]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align="right">%.4f</TD></TR>',['AIC',FAIC]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align="right">%.4f</TD></TR>',['BIC',FBIC]);
 s:= s + '</table>';
 s:= s + '<H3>Mean of clusters</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 //les en-tte
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Attribute</TH>';
 for k:= 1 to self.FStrucEM.FLstClusters.Count do
  s:= s + format('<TH>Cluster_%d</TH>',[k]);
 s:= s + '</TR>';
 //la descritption
 for j:= 1 to self.Inputs.Count do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TH>%s</TH>',[self.Inputs.Attribute[pred(j)].Name]);
   //et pour chaque cluster
   for k:= 1 to self.FStrucEM.FLstClusters.Count do
    s:= s + format('<TD align="right">%.4f</TD>',[(self.FStrucEM.FLstClusters.Items[pred(k)] as TOneClusterEM).FAvg[j]]);
  end;
 s:= s + '</TR>';
 s:= s +  '</table>';
 result:= s;
end;

function TCalcClusEM.getNbClus: integer;
begin
 result:= self.FStrucEM.FLstClusters.Count;
end;

function TCalcClusEM.getNbParameters(nbClus: integer): integer;
var k: integer;
begin
 k:= nbClus;//nombre de clusters
 k:= k * (Inputs.Count + ((Inputs.Count * (Inputs.Count + 1)) div 2));//moyennes et cov. matrices
 k:= k + (nbClus - 1);//les probas d'appartenance (-1 parce la somme des probas doit tre gale  1)
 result:= k;
end;

function TCalcClusEM.logLikelihood(prmTestSet: TExamples): double;
var total,value: double;
    i,example: integer;
    cluster: TTypeDiscrete;
begin
 total:= 0.0;
 for i:= 1 to prmTestSet.Size do
  begin
   example:= prmTestSet.Number[i];

   cluster:= self.SetClusterExample(example);
   if (cluster > 0) and (cluster <= self.FStrucEM.getNbClusters()) And ((self.FStrucEM.FLstClusters.Items[pred(cluster)] as TOneClusterEM).FPoids > 0.0)
    then
     begin
      value:= LN((self.FStrucEM.FLstClusters.Items[pred(cluster)] as TOneClusterEM).FPoids);
      value:= value + (self.FStrucEM.FLstClusters.Items[pred(cluster)] as TOneClusterEM).logDensity(example);
      total:= total + value;
     end;

  end;
 result:= total;
end;

function TCalcClusEM.SetClusterExample(example: integer): TTypeDiscrete;
var k,kmax: integer;
    f,fMax: double;
begin
 kMax:= -1;
 fMax:= -1.0 * MATH.MaxDouble;
 for k:= 1 to self.FStrucEM.FLstClusters.Count do
  begin
   f:= (self.FStrucEM.FLstClusters.Items[pred(k)] as TOneClusterEM).logDensity(example);
   if (f > fMax)
    then
     begin
      kMax:= k;
      fMax:= f;
     end;
  end;
 if(kMax < 0)
  //affectation au hasard
  then result:= succ(random(self.FStrucEM.FLstClusters.Count))
  //affectation
  else result:= kMax;
end;

{ TOneClusterEM }

constructor TOneClusterEM.create(prmAtts: TLstAttributes;
  prmClusterId: TTypeDiscrete);
begin
 inherited create();
 FClusterId:= prmClusterId;
 FAtts:= prmAtts;
 DimVector(FAvg,prmAtts.Count);
 DimMatrix(FVCV,prmAtts.Count,prmAtts.Count);
 DimMatrix(FInvVCV,prmAtts.Count,prmAtts.Count);
end;

destructor TOneClusterEM.destroy;
begin
 Finalize(FAvg);
 Finalize(FVCV);
 Finalize(FInvVCV);
 inherited;
end;

function TOneClusterEM.logDensity(example: integer): double;
var sum1, sum2: double;
    j,jprim: integer;
begin
 //la partie exponentielle
 sum1:= 0.0;
 for j:= 1 to FAtts.Count do
  begin
   sum2:= 0.0;
   for jprim:= 1 to FAtts.Count do
    sum2:= sum2 + (FAtts.Attribute[pred(jprim)].cValue[example] - FAvg[jprim]) * FInvVCV[jprim,j];
   sum1:= sum1 + (FAtts.Attribute[pred(j)].cValue[example] - FAvg[j]) * sum2;
  end;
 //et les scalaires
 if (FDetVCV > TOLERANCE_FOR_DET_VCV)
  then result:= -0.5 * FAtts.Count * LN(2.0 * D6_FMATH.PI) - 0.5 * LN(FDetVCV) - 0.5 * sum1
  else result:= 0.0;
end;

function TOneClusterEM.update(prmExamples: TExamples;
  prmAppartenance: TMatrix; ridgeCoef: double): boolean;
var sumPoids: double;
    sum: double;
    i,j,jprim: integer;
    example: integer;
begin
 TRY
 //calcul du poids et de la moyenne du cluster
 sumPoids:= 0.0;
 for j:= 1 to FAtts.Count do FAvg[j]:= 0.0;
 for i:= 1 to prmExamples.Size do
  begin
   example:= prmExamples.Number[i];
   sumPoids:= sumPoids + prmAppartenance[i,self.FClusterId];
   for j:= 1 to FAtts.Count do
    FAvg[j]:= FAvg[j] + prmAppartenance[i,self.FClusterId] * FAtts.Attribute[pred(j)].cValue[example];
  end;
 //et donc...
 FPoids:= sumPoids/(1.0 * prmExamples.Size);
 if (sumPoids > 0)
  then for j:= 1 to FAtts.Count do FAvg[j]:= FAvg[j] / sumPoids;
 //la matrice de variance co-variance maintenant -- on ne s'embarasse pas, on va au plus simple
 for j:= 1 to FAtts.Count do
  begin
   for jprim:= j to FAtts.Count do
    begin
     sum:= 0.0;
     for i:= 1 to prmExamples.Size do
      begin
       example:= prmExamples.Number[i];
       sum:= sum + prmAppartenance[i,self.FClusterId] * (FAtts.Attribute[pred(j)].cValue[example] - FAvg[j]) * (FAtts.Attribute[pred(jprim)].cValue[example] - FAvg[jprim]);
      end;
     if (sumPoids > 0.0)
      then FVCV[j,jprim]:= sum / sumPoids;
     //le triangle infrieur
     FVCV[jprim,j]:= FVCV[j,jprim];
    end;
  end;
 //correction Ridge
 for j:= 1 to FAtts.Count do
  FVCV[j,j]:= FVCV[j,j] + ridgeCoef; 
 //calculer le dterminant et l'inverse...
 result:= (D6_MATRICES.InvDet(FVCV,1,FAtts.Count,FInvVCV,FDetVCV) = D6_MATRICES.MAT_OK);  
 EXCEPT
 result:= false;
 END;
end;

{ TStrucClusterEM }

constructor TStrucClusterEM.create(prmCalcClus: TCalcClusEM);
var k: integer;
begin
 inherited create();
 FCalcEM:= prmCalcClus;
 FLstClusters:= TObjectList.Create(TRUE);
 //construire les noyaux de EM  partir de ceux des K-Means
 for k:= 0 to pred(prmCalcClus.FCalcKMeans.Clusters.Count) do
  FLstClusters.Add(TOneClusterEM.create(prmCalcClus.Inputs,succ(k)));
end;

destructor TStrucClusterEM.destroy;
begin
 FLstClusters.Free();
 Finalize(FAppartenance);
 inherited;
end;

function TStrucClusterEM.initialize(prmExamples: TExamples): boolean;
var k,i,cluster: integer;
begin
 TRY
 //initialiser le tableau d'appartenance
 DimMatrix(FAppartenance,prmExamples.Size,FLstClusters.Count);
 //puis le remplir
 for i:= 1 to prmExamples.Size do
  begin
   cluster:= FCalcEM.FCalcKMeans.SetClusterExample(prmExamples.Number[i]);
   FAppartenance[i,cluster]:= 1.0;
  end;
 //initialiser chaque noyau de EM
 for k:= 0 to pred(FLstClusters.Count) do
  (FLstClusters.Items[k] as TOneClusterEM).update(prmExamples,FAppartenance,(FCalcEM.PrmCalc as TOpPrmClusEM).FRidgeCorrection);
 //si tout s'est pass sans soucis...
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

function TStrucClusterEM.workingLQ(prmExamples: TExamples): double;
var i,k: integer;
    total, sum: double;
begin
 total:= 0.0;
 for i:= 1 to prmExamples.Size do
  begin
   sum:= 0.0;
   for k:= 1 to FLstClusters.Count do
    if ((FLstClusters.Items[pred(k)] as TOneClusterEM).FPoids > 0.0)
     then sum:= sum + FAppartenance[i,k] * (LN((FLstClusters.Items[pred(k)] as TOneClusterEM).FPoids) + (FLstClusters.Items[pred(k)] as TOneClusterEM).logDensity(prmExamples.Number[i]));
   total:= total + sum;
  end;
 result:= total;
end;

function TStrucClusterEM.updateGroupMembership(
  prmExamples: TExamples): boolean;
var i,k: integer;
    sum: double;
begin
 TRY
 for i:= 1 to prmExamples.Size do
  begin
   sum:= 0.0;
   for k:= 1 to self.FLstClusters.Count do
    begin
     FAppartenance[i,k]:= (FLstClusters.Items[pred(k)] as TOneClusterEM).FPoids * EXP((FLstClusters.Items[pred(k)] as TOneClusterEM).logDensity(prmExamples.Number[i]));
     sum:= sum + FAppartenance[i,k];
    end;
   //normalisation
   if (sum > 0.0)
    then for k:= 1 to self.FLstClusters.Count do FAppartenance[i,k]:= FAppartenance[i,k] / sum;
  end;
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

function TStrucClusterEM.updateParameters(prmExamples: TExamples): boolean;
var k: integer;
begin
 TRY
 for k:= 0 to pred(FLstClusters.Count) do
  (FLstClusters.Items[k] as TOneClusterEM).update(prmExamples,FAppartenance,(FCalcEM.PrmCalc as TOpPrmClusEM).FRidgeCorrection);
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

function TStrucClusterEM.getNbClusters: integer;
begin
 result:= self.FLstClusters.Count;
end;

initialization
 RegisterClass(TGenClusEM);
end.
