(**************************************************************************)
(* UCompClusteringEMSelection.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**************************************************************************)

{
@abstract(Dtermination du nombre de clusters pour un EM -- Recherche sur une plage de valeur)
@author(Ricco)
@created(27/08/2006)

Teste "val.dpart"  "val.fin" et slectionne celui qui optimise le critre choisi.
Deux types d'valuation : resubstitution et validation croise

Ce composant doit tre branch exclusivement aprs un Clustering EM, il le manipule directement.
Il ne produit pas en lui-mme de prdiction, il cherche tout simplement  slectionner le bon nombre de clusters pour le composant sur lequel il est branch.
}
unit UCompClusteringEMSelection;

interface

USES
    Forms,Classes,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetExamples,
    UCompClusteringEM,
    UCalcRndGenerator;
    
TYPE
    {gnrateur}
    TGenClusEMSelection = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant -- attention, l'anctre n'est pas un clustering}
    TMLCompClusEMSelection = class(TMLCompLocalData)
    protected
    function    getClassOperator: TClassOperator; override;
    public
    //n'est connectable que sur un composant clustering EM
    class function isConnectableTo(destination: TMLComponent): boolean; override;
    end;

    {les critres  utiliser}
    TEnumCriterionEMSelection = (emsAIC, emsBIC, emsLLikelihood);

    {oprateur}
    TOpClusEMSelection = class(TOpLocalData)
    private
    //tableau des valeurs pour chaque nombre de clusters
    FTabValues: array of double;
    //oprateur prcdent
    FOpEM: TOpClusEM;
    //Nombre de cluster optimal
    FNbClusSolution: integer;
    //calculer un EM
    function computeAndAssess(training, test: TExamples; prmEM: TOpPrmClusEM; criterion: TEnumCriterionEMSelection): double;
    function crossvalidation(prmExamples: TExamples; prmEM: TOpPrmClusEM): double;
    //valuer un EM
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    CoreExecute(): boolean; override;

    public
    //rcuprer au passage l'oprateur EM
    constructor Create(AOwner: TObject); override;
    destructor  destroy(); override;
    function    getHTMLResultsSummary(): string; override;
    end;

CONST
    {description des critres}
    STR_CRIT_EM_SELECTION : array[TEnumCriterionEMSelection] of string = ('AIC','BIC','Log-Likelihood');

    {description du mode d'valuation}
    STR_EVAL_TYPE_EM_SELECTION : array[0..1] of string = ('Resubstitution','Cross-validation');

TYPE
    {paramtrage}
    TOpPrmClusEMSelection = class(TOperatorParameter)
    private
    //critre d'valuation -- df. AICc
    FCriterion: TEnumCriterionEMSelection;
    //val. de dpart et de fin -- 1  10
    FStartClus, FEndClus: integer;
    //mode d'valuation {0: resubstitution, "1": validation-croise}
    FEvalType: integer;
    //folds dans le cas de la validation croise -- 5
    FNbFolds: integer;
    //gnrateur de nombre alatoire pour la validation croise
    FRandom: TStartSeed;
    //mj du composant EM associ -- df. TRUE
    FUpdateEMClustering: boolean;
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    property    Criterion: TEnumCriterionEMSelection read FCriterion write FCriterion;
    property    StartClus: integer read FStartClus write FStartClus;
    property    EndClus: integer read FEndClus write FEndClus;
    property    EvalType: integer read FEvalType write FEvalType;
    property    NbFolds: integer read FNbFolds write FNbFolds;
    property    Random: TStartSeed read FRandom write FRandom;
    property    UpdateEMClustering: boolean read FUpdateEMClustering write FUpdateEMClustering; 
    end;


implementation

uses
    Sysutils, Math,
    UConstConfiguration, ULogFile, UDlgOpPrmClusEMSelection;

{ TGenClusEMSelection }

procedure TGenClusEMSelection.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

function TGenClusEMSelection.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusEMSelection; 
end;

{ TMLCompClusEMSelection }

function TMLCompClusEMSelection.getClassOperator: TClassOperator;
begin
 result:= TOpClusEMSelection; 
end;

class function TMLCompClusEMSelection.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 result:= (destination is TMLCompClusEM);
end;

{ TOpClusEMSelection }

function TOpClusEMSelection.computeAndAssess(training,
  test: TExamples; prmEM: TOpPrmClusEM; criterion: TEnumCriterionEMSelection): double;
var calcEM: TCalcClusEM;
    value: double;
    llk: double;
    eval: TExamples;
begin
 //pour viter les warnings... du compilateur...
 llk:= 0.0;
 //donnes test
 eval:= test;
 if (eval = nil) then eval:= training;
 TRY
 //apprentissage
 calcEM:= TCalcClusEM.Create(FOpEM.Targets,FOpEM.Inputs,prmEM,NIL);
 calcEM.BuildClusters(training);
 //test
 llk:= calcEM.logLikelihood(eval);
 //rcupration du critre
 case Criterion of
  emsAIC: value:= calcEM.getAIC(calcEM.getNbClus(),llk,1.0*eval.size);
  emsBIC: value:= calcEM.getBIC(calcEM.getNbClus(),llk,1.0*eval.Size);
  //puisqu'on cherche  minimiser
  else value:= -1.0 * llk;
 end;
 calcEM.Free();
 result:= value;
 EXCEPT
 on e: Exception do
  begin
   TraceLog.WriteToLogFile(format('[ERROR] EM SEL "ComputeAndAssess" = %s -- eval.size = %d -- llk = %.4f',[e.Message,eval.Size,llk]));
   result:= MATH.MaxDouble;
  end;
 END;
end;

function TOpClusEMSelection.CoreExecute: boolean;
var
    prm: TOpPrmClusEMSelection;
    prmEM: TOpPrmClusEM;
    nClus: integer;
    criterion: double;
    nClusInit: integer;
    minCriterion: double;
begin
 //paramtre de l'oprateur EM
 prmEM:= FOpEM.PrmOp as TOpPrmClusEM;
 nClusInit:= prmEM.NbClusters;
 //paramtre courant
 prm:= self.PrmOp as TOpPrmClusEMSelection;

 //recherche de la partition optimale...
 TRY
 //prparer le tableau de rsultats
 setLength(FTabValues,succ(prm.FEndClus-prm.FStartClus));
 //initialisation
 minCriterion:= MATH.MaxDouble;
 self.FNbClusSolution:= nClusInit;
 //grener les solutions
 for nClus:= prm.FStartClus to prm.FEndClus do
  begin
   //fixer le nombre de cluster
   prmEm.NbClusters:= nClus;
   //selon le mode d'valuation
   case prm.FEvalType of
    1: criterion:= self.crossvalidation(self.WorkData.Examples,prmEm)
    else criterion:= self.computeAndAssess(self.WorkData.Examples,nil,prmEm,prm.FCriterion);
   end;
   TraceLog.WriteToLogFile(format('[EM Selection] # clusters = %d, criterion = %.4f',[nClus,criterion]));
   //rcuprer tout cela
   FTabValues[nClus-prm.FStartClus]:= criterion;
   //et vrifier si on est  min.
   if (criterion < minCriterion)
    then
     begin
      FNbClusSolution:= nClus;
      minCriterion:= criterion;
     end;
  end;
 result:= true;
 EXCEPT
 result:= false;
 END;

 //et donc... que faire de ce rsultat ?
 if not(prm.FUpdateEMClustering)
  then prmEM.NbClusters:= nClusInit
  else
   begin
    prmEM.NbClusters:= FNbClusSolution;
    if (nClusInit <> FNbClusSolution)
     then (FOpEM.MLOwner as TMLComponent).Execute(true); 
   end;

end;

constructor TOpClusEMSelection.Create(AOwner: TObject);
begin
 inherited;
 FOpEM:= (MLOwner as TMLComponent).Predecessor.Operator as TOpClusEM;
end;

function TOpClusEMSelection.crossvalidation(
  prmExamples: TExamples; prmEM: TOpPrmClusEM): double;
var nbFolds,nbExamples,foldSize: integer;
    sample,train,test: TExamples;
    f,bb,bh,i: integer;
    curLLk: double;
    llk: double;
begin
 //vrifier compatibilit du nombre de folds avec la taille de l'chantillon
 nbExamples:= prmExamples.Size;
 nbFolds:= (self.PrmOp as TOpPrmClusEMSelection).FNbFolds;
 if (nbFolds > prmExamples.Size)
  then nbFolds:= prmExamples.Size;
 foldSize:= nbExamples div nbFolds;
 //prparer les chantillons
 sample:= TExamples.Create(prmExamples.Size);
 sample.Copy(prmExamples);
 //et...
 train:= TExamples.Create(prmExamples.Size);
 test:= TExamples.Create(prmExamples.Size);
 //mlanger selon le mode voulu par l'utilisateur
 sample.procRandomizeExamples((self.PrmOp as TOpPrmClusEMSelection).FRandom);
 //construire les sessions d'valuation
 llk:= 0.0;
 for f:= 1 to nbFolds do
  begin
   //construire la liste des individus
   bb:= pred(f)*foldSize;
   bh:= f*foldSize+1;
   train.BeginAdd();
   test.BeginAdd();
   for i:= 1 to sample.Size do
    begin
     if (i>bb) and (i<bh)
      then test.AddExample(sample.Number[i])
      else train.AddExample(sample.Number[i]);
    end;
   train.EndAdd();
   test.EndAdd();
   //et donc..., attention toujours LLK demand parce que lui est additif, ce n'est pas le cas de AIC et BIC !!!
   curLLk:= self.computeAndAssess(train,test,prmEM,emsLLikelihood);
   //et le total -- attention computeAndAssess renvoie (-1.0 * LLk), d'o la soustraction ici
   llk:= llk - curLLk;
  end;
 //renvoyer la valeur de retour
 case (self.PrmOp as TOpPrmClusEMSelection).FCriterion of
  emsAIC: result:= (self.FOpEM.CalcClustering as TCalcClusEM).getAIC(prmEM.NbClusters,llk,1.0*sample.size);
  emsBIC: result:= (self.FOpEM.CalcClustering as TCalcClusEM).getBIC(prmEM.NbClusters,llk,1.0*sample.size);
  else result:= -1.0 * llk;
 end;
 //librer la mmoire
 sample.Free();
 train.Free();
 test.Free();
end;

destructor TOpClusEMSelection.destroy;
begin
 finalize(FTabValues);
 inherited;
end;

function TOpClusEMSelection.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmClusEMSelection;
end;

function TOpClusEMSelection.getHTMLResultsSummary: string;
var s: string;
    nClus: integer;
    prm: TOpPrmClusEMSelection;
begin
 //paramtre courant
 prm:= self.PrmOp as TOpPrmClusEMSelection;
 //zoo...
 s:= '<H3>Criterion values w.r.t. # of clusters</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH># of clusters</TH><TH>Criterion</TH></TR>';
 for nClus:= prm.FStartClus to prm.FEndClus do
  begin
   if (nClus = FNbClusSolution)
    then s:= s + HTML_TABLE_COLOR_DATA_RED + format('<TD align="center">%d</TD><TD align="right">%.6f</TD></TR>',[nClus,FTabValues[nClus-prm.FStartClus]])
    else s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="center">%d</TD><TD align="right">%.6f</TD></TR>',[nClus,FTabValues[nClus-prm.FStartClus]]);
  end;
 s:= s + '</table>';
 //and then...
 result:= s;
end;

{ TOpPrmClusEMSelection }

function TOpPrmClusEMSelection.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmClusEMSelection.CreateFromOpPrm(self);
end;

function TOpPrmClusEMSelection.getHTMLParameters: string;
var s: string;
begin
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Parameter</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%s</TD></TR>',['Criterion',STR_CRIT_EM_SELECTION[FCriterion]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%d</TD></TR>',['# clusters -- Start value',FStartClus]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%d</TD></TR>',['# clusters -- End value',FEndClus]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%s</TD></TR>',['Evaluation type',STR_EVAL_TYPE_EM_SELECTION[FEvalType]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%d</TD></TR>',['Folds for CV',FNbFolds]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align="right">%s</TD></TR>',['Random generator',StartSeedDescription[FRandom]]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmClusEMSelection.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FCriterion:= TEnumCriterionEMSelection(prmINI.ReadInteger(prmSection,'criterion',ord(FCriterion)));
 FStartClus:= prmINI.ReadInteger(prmSection,'start_clus',FStartClus);
 FEndClus:= prmINI.ReadInteger(prmSection,'end_clus',FEndClus);
 FEvalType:= prmINI.ReadInteger(prmSection,'eval_type',FEvalType);
 FNbFolds:= prmINI.ReadInteger(prmSection,'folds',FNbFolds);
 FRandom:= TStartSeed(prmINI.ReadInteger(prmSection,'random',ord(FRandom)));
 FUpdateEMClustering:= boolean(prmINI.ReadInteger(prmSection,'update',ord(FUpdateEMClustering)));
end;

procedure TOpPrmClusEMSelection.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FCriterion,sizeof(FCriterion));
 prmStream.ReadBuffer(FStartClus,sizeof(FStartClus));
 prmStream.ReadBuffer(FEndClus,sizeof(FEndClus));
 prmStream.ReadBuffer(FEvalType,sizeof(FEvalType));
 prmStream.ReadBuffer(FNbFolds,sizeof(FNbFolds));
 prmStream.ReadBuffer(FRandom,sizeof(FRandom));
 prmStream.ReadBuffer(FUpdateEMClustering,sizeof(FUpdateEMClustering));
end;

procedure TOpPrmClusEMSelection.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'criterion',ord(FCriterion));
 prmINI.WriteInteger(prmSection,'start_clus',FStartClus);
 prmINI.WriteInteger(prmSection,'end_clus',FEndClus);
 prmINI.WriteInteger(prmSection,'eval_type',FEvalType);
 prmINI.WriteInteger(prmSection,'folds',FNbFolds);
 prmINI.WriteInteger(prmSection,'random',ord(FRandom));
 prmINI.WriteInteger(prmSection,'update',ord(FUpdateEMClustering));
end;

procedure TOpPrmClusEMSelection.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FCriterion,sizeof(FCriterion));
 prmStream.WriteBuffer(FStartClus,sizeof(FStartClus));
 prmStream.WriteBuffer(FEndClus,sizeof(FEndClus));
 prmStream.WriteBuffer(FEvalType,sizeof(FEvalType));
 prmStream.WriteBuffer(FNbFolds,sizeof(FNbFolds));
 prmStream.WriteBuffer(FRandom,sizeof(FRandom));
 prmStream.WriteBuffer(FUpdateEMClustering,sizeof(FUpdateEMClustering));
end;

procedure TOpPrmClusEMSelection.SetDefaultParameters;
begin
 //critre d'valuation
 FCriterion:= emsAIC;
 //val. de dpart et de fin -- 1  10
 FStartClus:= 1;
 FEndClus:= 10;
 //mode d'valuation {0: resubstitution, "1": validation-croise}
 FEvalType:= 1;
 //folds dans le cas de la validation croise -- 5
 FNbFolds:= 5;
 //ala
 FRandom:= seedStandard;
 //mj du composant EM associ
 FUpdateEMClustering:= true;
end;

initialization
 RegisterClass(TGenClusEMSelection);
end.
