(*********************************************************************)
(* UCompSDCatAssocLambda.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Composant association sur variables catgorielles. Lambda de Goodman & Kruskal)
@author(Ricco)
@created(17/05/2007)

Doc : http://v8doc.sas.com/sashtml/stat/chap28/sect20.htm
Formule + Variance asymptotique

}
unit UCompSDCatAssocLambda;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UCompSDDefinition,
        UOperatorDefinition,
        UCalcStatDesCorrelation,
        UCompSDLinearCorrelation;

TYPE
        {Gnrateur de univariate stat cont desc}
        TMLGenCompSDCatLambda = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant}
        TMLCompSDCatLambda = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur associ}
        TOpSDCatLambda = class(TOpSD)
        protected
        procedure   InitializeListStat(); override;
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RebuildStatDes(); override;
        function    CheckAttributes(): boolean; override;
        public
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage d'oprateur tableaux croiss}
        TOpPrmSDCatLambda = class(TOpPrmSD)
        protected
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

uses
  Sysutils, UDatasetDefinition,
  UDatasetImplementation, UDlgOpPrmSDLinearCorrelation, UStringsResources,
  UConstConfiguration, UCalcStatDesCrossTab,
  UDlgOpPrmSDCategoricalAssociation, UDlgOpPrmSDCatAssocLambda;

{ TMLGenCompSDCatLambda }

procedure TMLGenCompSDCatLambda.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDCatLambda.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDCatLambda; 
end;

{ TMLCompSDCatLambda }

function TMLCompSDCatLambda.getClassOperator: TClassOperator;
begin
 result:= TOpSDCatLambda; 
end;

function TMLCompSDCatLambda.GetLogResultDescription: string;
begin
 result:= format('%d Goodman & Kruskal Lambda computed on %d examples',[(Operator as TOpSD).LstStat.Count,(Operator as TOpSD).Workdata.Examples.Size]);
end;

{ TOpSDCatLambda }

function TOpSDCatLambda.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete));
 result:= result and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
end;

function TOpSDCatLambda.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDCatLambda;
end;

const
    VAL_CRIT_LOI_NORMALE_5pc : double = 1.95996108232;

function TOpSDCatLambda.getHTMLResultsSummary: string;
var s: string;
    stat: TCalcSDCrossTab;
    j: integer;
begin
 s:= '<H3>Goodman & Kruskal''s Lambda for nominal attributes</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Y</TH><TH width=80>X</TH><TH width=60>Lambda</TH><TH width=60>sigma</TH><TH width=100>95% C.I.</TH>' + '</TR>';
 for j:= 0 to pred(self.LstStat.Count) do
  begin
   stat:= self.LstStat.Stat(j) as TCalcSDCrossTab;
   //affichage de la ligne
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%s</TD><TD>%s</TD><TD align=right>%7.6f</TD><TD align=right>%.6f</TD><TD align=right>%.4f ; %.4f</TD>',
                  [stat.RowAtt.Name,stat.ColAtt.Name,stat.Lambda,stat.sigmaLambda,stat.Lambda-VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaLambda,stat.Lambda+VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaLambda]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDCatLambda.InitializeListStat;
begin
 FLstStat:= TLstCalcSDCrossTabLambda.Create(nil,nil);
end;

procedure TOpSDCatLambda.RebuildStatDes;
var stat: TCalcSDCrossTab;
    attI,attJ: TAttribute;
    i,j: integer;
begin
 //seul le schma TARGET vs. INPUT est licite !!!
 //pour chaque variable de description
 for i:= 0 to pred(workdata.LstAtts[asTarget].Count) do
  begin
   attI:= workdata.LstAtts[asTarget].Attribute[i];
   //le type est ok ?
   if attI.isCategory(caDiscrete)
    then
     begin
      //les variables  dcrire
      for j:= 0 to pred(WorkData.LstAtts[asInput].Count) do
       begin
        attJ:= WorkData.LstAtts[asInput].Attribute[j];
        //petit test quand mme
        if attJ.isCategory(caDiscrete) and (attI<>attJ)
         then
          begin
           stat:= TCalcSDCrossTab.Create(attI,attJ,nil);
           self.LstStat.AddStat(stat);
          end;
       end;
     end;
  end;
end;


{ TOpPrmSDCatLambda }

function TOpPrmSDCatLambda.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDCatAssocLambda.CreateFromOpPrm(self);
end;

initialization
 RegisterClass(TMLGenCompSDCatLambda);
end.
