(*********************************************************************)
(* UCompSDCatAssocTheilU.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Composant association sur variables catgorielles. U de Theil)
@author(Ricco)
@created(10/05/2007)

U de Theil (Asymmetric Uncertainty Coefficient).
--> http://v8doc.sas.com/sashtml/stat/chap28/sect20.htm

Transformation du CHI-2 pour le test de significativit (cf. Ritschard, 2003).
Variance asymptotique (cf. site SAS ci-dessus).

}
unit UCompSDCatAssocTheilU;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UCompSDDefinition,
        UOperatorDefinition,
        UCalcStatDesCorrelation,
        UCompSDLinearCorrelation;

TYPE
        {Gnrateur de univariate stat cont desc}
        TMLGenCompSDCatTheilU = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant}
        TMLCompSDCatTheilU = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur associ}
        TOpSDCatTheilU = class(TOpSD)
        protected
        procedure   InitializeListStat(); override;
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RebuildStatDes(); override;
        function    CheckAttributes(): boolean; override;
        public
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage d'oprateur tableaux croiss}
        TOpPrmSDCatTheilU = class(TOpPrmSD)
        protected
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

uses
        Sysutils, UDatasetDefinition,
  UDatasetImplementation, UDlgOpPrmSDLinearCorrelation, UStringsResources,
  UConstConfiguration, UCalcStatDesCrossTab,
  UDlgOpPrmSDCategoricalAssociation, UDlgOpPrmSDCatAssocTheil;

{ TMLGenCompSDCatTheilU }

procedure TMLGenCompSDCatTheilU.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDCatTheilU.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDCatTheilU;
end;

{ TOpPrmSDCatTheilU }

function TOpPrmSDCatTheilU.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDCatAssocTheil.CreateFromOpPrm(self);
end;

{ TMLCompSDCatTheilU }

function TMLCompSDCatTheilU.getClassOperator: TClassOperator;
begin
 result:= TOpSDCatTheilU; 
end;

function TMLCompSDCatTheilU.GetLogResultDescription: string;
begin
 result:= format('%d Theil''s U tau computed on %d examples',[(Operator as TOpSD).LstStat.Count,(Operator as TOpSD).Workdata.Examples.Size]);
end;

{ TOpSDCatTheilU }

function TOpSDCatTheilU.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete));
 result:= result and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
end;

function TOpSDCatTheilU.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDCatTheilU; 
end;

const
    VAL_CRIT_LOI_NORMALE_5pc : double = 1.95996108232;

function TOpSDCatTheilU.getHTMLResultsSummary: string;
var s: string;
    stat: TCalcSDCrossTab;
    j: integer;
begin
 s:= '<H3>Theil''s U (Uncertainty Coefficient) for nominal attributes</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Y</TH><TH width=80>X</TH><TH width=60>Theil''s U</TH><TH width=60>Chi</TH><TH width=60>d.f.</TH><TH width=60>p-value</TH><TH width=60>sigma</TH><TH width=100>95% C.I.</TH>' + '</TR>';
 for j:= 0 to pred(self.LstStat.Count) do
  begin
   stat:= self.LstStat.Stat(j) as TCalcSDCrossTab;
   //affichage de la ligne
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%s</TD><TD>%s</TD><TD align=right>%7.6f</TD><TD align=right>%.2f</TD><TD align=right>%d</TD><TD align=right>%.4f</TD><TD align=right>%.6f</TD><TD align=right>%.4f ; %.4f</TD>',
                  [stat.RowAtt.Name,stat.ColAtt.Name,stat.U,stat.chiU,stat.DDL,stat.prU,stat.sigmaU,stat.U-VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaU,stat.U+VAL_CRIT_LOI_NORMALE_5pc*stat.sigmaU]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDCatTheilU.InitializeListStat;
begin
 FLstStat:= TLstCalcSDCrossTabTheil.Create(nil,nil);
end;

procedure TOpSDCatTheilU.RebuildStatDes;
var stat: TCalcSDCrossTab;
    attI,attJ: TAttribute;
    i,j: integer;
begin
 //seul le schma TARGET vs. INPUT est licite !!!
 //pour chaque variable de description
 for i:= 0 to pred(workdata.LstAtts[asTarget].Count) do
  begin
   attI:= workdata.LstAtts[asTarget].Attribute[i];
   //le type est ok ?
   if attI.isCategory(caDiscrete)
    then
     begin
      //les variables  dcrire
      for j:= 0 to pred(WorkData.LstAtts[asInput].Count) do
       begin
        attJ:= WorkData.LstAtts[asInput].Attribute[j];
        //petit test quand mme
        if attJ.isCategory(caDiscrete) and (attI<>attJ)
         then
          begin
           stat:= TCalcSDCrossTab.Create(attI,attJ,nil);
           self.LstStat.AddStat(stat);
          end;
       end;
     end;
  end;
end;

initialization
 RegisterClass(TMLGenCompSDCatTheilU);
end.
