(*****************************************************************************)
(* UCompSDCategoricalCorrelation.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*****************************************************************************)

{
@abstract(Composant corrlations sur variables catgorielles)
@author(Ricco)
@created(04/05/2007)

Calcul des corrlations sur donnes catgorielles multivalues, cf. article S.L. & R.R., PKDD'2000.
+ Transformation de Fisher pour calculer la z-value (qu'on va appeler test-value), calcul de la p-value sur la base de cette transformation. 
}
unit UCompSDCategoricalCorrelation;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UCompSDDefinition,
        UOperatorDefinition,
        UCalcStatDesCorrelation,
        UCompSDLinearCorrelation;

TYPE
        {Gnrateur de univariate stat cont desc}
        TMLGenCompSDCatCorr = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant}
        TMLCompSDCatCorr = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur associ}
        TOpSDCatCorrelation = class(TOpSD)
        protected
        procedure   InitializeListStat(); override;
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RebuildStatDes(); override;
        function    CheckAttributes(): boolean; override;
        public
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage d'oprateur tableaux croiss}
        TOpPrmSDCatCorr = class(TOpPrmSDCorrelation)
        end;

implementation

uses
    SysUtils, UDatasetImplementation, UDatasetDefinition,
  UCalcStatDesCrossTab, UConstConfiguration, FMath;

{ TMLGenCompSDCatCorr }

procedure TMLGenCompSDCatCorr.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDCatCorr.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDCatCorr;
end;

{ TMLCompSDCatCorr }

function TMLCompSDCatCorr.getClassOperator: TClassOperator;
begin
 result:= TOpSDCatCorrelation;
end;

function TMLCompSDCatCorr.GetLogResultDescription: string;
begin
 result:= format('%d correlation(s) computed on %d examples',[(Operator as TOpSD).LstStat.Count,(Operator as TOpSD).Workdata.Examples.Size]);
end;

{ TOpSDCatCorrelation }

function TOpSDCatCorrelation.CheckAttributes: boolean;
var ok: boolean;
begin
 ok:= TRUE;
 case (self.PrmOp as TOpPrmSDCorrelation).InputList of
  //target x input
  0: begin
      ok:= ok and (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete));
      ok:= ok and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
     end;
  //cross-input
  1 : begin
       ok:= ok and (self.WorkData.LstAtts[asInput].Count > 1) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
      end;
 end;
 result:= ok;
end;

function TOpSDCatCorrelation.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDCatCorr; 
end;

function TOpSDCatCorrelation.getHTMLResultsSummary: string;
var s: string;
    stat: TCalcSDCrossTab;
    j: integer;
    r2: double;
begin
 s:= '<H3>Correlation coefficient for categorical attributes</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Y</TH><TH width=80>X</TH><TH width=60>r</TH><TH width=60>r</TH><TH width=60>z-score</TH><TH width=60>p-value</TH>' + '</TR>';
 for j:= 0 to pred(self.LstStat.Count) do
  begin
   stat:= self.LstStat.Stat(j) as TCalcSDCrossTab;
   //r
   r2:= SQR(stat.Correlation);
   //affichage de la ligne
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%s</TD><TD>%s</TD><TD align=right>%5.4f</TD><TD align=right>%5.4f</TD><TD align=right>%.2f</TD><TD align=right>%5.4f</TD>',
                  [stat.RowAtt.Name,stat.ColAtt.Name,stat.Correlation,r2,stat.ZScoreCorrelation,stat.prCorrelation]); 
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDCatCorrelation.InitializeListStat;
begin
 FLstStat:= TLstCalcSDCrossTabCorrelation.Create(nil,nil);
end;

procedure TOpSDCatCorrelation.RebuildStatDes;
var stat: TCalcSDCrossTab;
    attI,attJ: TAttribute;
    i,j: integer;
begin
 self.LstStat.FreeAll();
 //
 case (PrmOp as TOpPrmSDCatCorr).InputList of
  //double liste target-input
  0: begin
       //pour chaque variable de description
       for i:= 0 to pred(workdata.LstAtts[asTarget].Count) do
        begin
         attI:= workdata.LstAtts[asTarget].Attribute[i];
         //le type est ok ?
         if attI.isCategory(caDiscrete)
          then
           begin
            //les variables  dcrire
            for j:= 0 to pred(WorkData.LstAtts[asInput].Count) do
             begin
              attJ:= WorkData.LstAtts[asInput].Attribute[j];
              //petit test quand mme
              if attJ.isCategory(caDiscrete) and (attI<>attJ)
               then
                begin
                 stat:= TCalcSDCrossTab.Create(attI,attJ,nil);
                 self.LstStat.AddStat(stat);
                end;
             end;
           end;
        end;
     end
  else
   //simple liste input  croiser
   begin
    for i:= 0 to WorkData.LstAtts[asInput].Count-2 do
     begin
      attI:= workdata.LstAtts[asInput].Attribute[i];
      if attI.isCategory(caDiscrete)
       then
        begin
          for j:= succ(i) to pred(WorkData.LstAtts[asInput].Count) do
           begin
            attJ:= WorkData.LstAtts[asInput].Attribute[j];
            //le deuxime test renvoie toujours vrai sinon il y a un srieux problme
            if attJ.isCategory(caDiscrete) and (attJ<>attI)
             then
              begin
               stat:= TCalcSDCrossTab.Create(attI,attJ,nil);
               self.LstStat.AddStat(stat);
              end;
           end;
        end;
     end;
   end;
 end;
end;

initialization
 RegisterClass(TMLGenCompSDCatCorr);
end.
