(*****************************************************************************)
(* UCompSDCategoricalAssociation.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*****************************************************************************)

{
@abstract(Composant association sur variables catgorielles)
@author(Ricco)
@created(28/09/2007)

Classes de base pour les associations ordinales.

}

unit UCompSDOrdinalAssociation;

interface

USES
    Forms,Classes,
    IniFiles, Contnrs,
    UCompDefinition,
    UCompSDDefinition,
    UOperatorDefinition,
    UCalcStatDesOrdinalAssociation;

TYPE
    {Gnrateur de univariate stat cont desc}
    TMLGenCompSDOrdinalAssoc = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override; //!\  surcharger
    end;

    {le composant}
    TMLCompSDOrdinalAssoc = class(TMLCompSD)
    protected
    function    getClassOperator: TClassOperator; override; //!\  surcharger
    function    GetLogResultDescription(): string; override;
    end;

    {l'oprateur associ}
    TOpSDOrdinalAssoc = class(TOpSD)
    protected
    //liste des vecteurs de valeurs
    FLstVecInput, FLstVecTarget: TObjectList;
    //supprimer ces listes
    procedure   destroyLstVec();
    //les classiques
    procedure   InitializeListStat(); override;
    procedure   RebuildStatDes(); override;
    function    CheckAttributes(): boolean; override;
    function    getClassParameter: TClassOperatorParameter; override;
    //nouveaux
    function    getClassComputation(): TClassCalcSDOrdinalAssoc; virtual; abstract; //!\  surcharger
    function    getMethodDescription(): string; virtual; abstract; //!\  surcharger
    public
    destructor  destroy(); override;
    function    getHTMLResultsSummary(): string; override;
    end;

    {paramtrage d'oprateur tableaux croiss}
    TOpPrmSDOrdinalAssoc = class(TOpPrmSD)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //montrer le tableau crois des dtails ? - 0, pas de dtails ; 1 - avec le tableau crois (dfaut)
    ShowDetails: integer;
    //type de slection des variables - 0 : croisement target/input (dfaut) ; 1 : croisement input
    InputList: integer;
    //description HTML des paramtres
    function    getHTMLParameters(): string; override;
    {inherited obligatoire chez les hritiers}
    procedure   LoadFromStream(prmStream: TStream); override;
    {inherited obligatoire chez les hritiers}
    procedure   SaveToStream(prmStream: TStream); override;
    {inherited obligatoire chez les hritiers}
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    {inherited obligatoire chez les hritiers}
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

implementation

USES
    Sysutils,
    UDatasetDefinition,
    UDatasetImplementation,
    UConstConfiguration,
    UCalcStructOrdinalAssociation, UDlgOpPrmOrdinalAssociation;

{ TMLGenCompSDOrdinalAssoc }

procedure TMLGenCompSDOrdinalAssoc.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDOrdinalAssoc.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDOrdinalAssoc; 
end;

{ TMLCompSDOrdinalAssoc }

function TMLCompSDOrdinalAssoc.getClassOperator: TClassOperator;
begin
 result:= TOpSDOrdinalAssoc;
end;

function TMLCompSDOrdinalAssoc.GetLogResultDescription: string;
begin
 result:= format('%d ordinal association computed on %d examples',[(Operator as TOpSD).LstStat.Count,(Operator as TOpSD).Workdata.Examples.Size]);
end;

{ TOpPrmSDOrdinalAssoc }

function TOpPrmSDOrdinalAssoc.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDOrdinalAssociation.CreateFromOpPrm(self);
end;

function TOpPrmSDOrdinalAssoc.getHTMLParameters: string;
var s,sPrm: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Cross-tab parameters</TH></TR>';
 if self.SortResult
  then sPrm:= 'yes'
  else sPrm:= 'no';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Sort results</TD><TD align="right">%s</TD></TR>',[sPrm]);
 if self.SortResult
  then
   begin
    case self.SortCriteria of
     0: sPrm:= 'Row attribute name';
     1: sPrm:= 'Column attribute name';
     2: sPrm:= 'Measure';
     3: sPrm:= 'Abs(Measure)';
    end;
    s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Sort criterion</TD><TD align="right">%s</TD></TR>',[sPrm]);
   end;
 if (self.InputList=0)
  then sPrm:= 'Target (Row) and input (Column)'
  else sPrm:= 'Cross-input (Row x Column)';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Input list</TD><TD align="right">%s</TD></TR>',[sPrm]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmSDOrdinalAssoc.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 ShowDetails:= prmINI.ReadInteger(prmSection,'show_details',ShowDetails);
 InputList:= prmINI.ReadInteger(prmSection,'input_list',InputList);
end;

procedure TOpPrmSDOrdinalAssoc.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(ShowDetails,sizeof(ShowDetails));
 prmStream.ReadBuffer(InputList,sizeof(InputList));
end;

procedure TOpPrmSDOrdinalAssoc.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'show_details',ShowDetails);
 prmINI.WriteInteger(prmSection,'input_list',InputList);
end;

procedure TOpPrmSDOrdinalAssoc.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(ShowDetails,sizeof(ShowDetails));
 prmStream.WriteBuffer(InputList,sizeof(InputList));
end;

procedure TOpPrmSDOrdinalAssoc.SetDefaultParameters;
begin
 inherited;
 ShowDetails:= 0;
 InputList:= 0;
end;

{ TOpSDOrdinalAssoc }

function TOpSDOrdinalAssoc.CheckAttributes: boolean;
begin
 if ((self.PrmOp as TOpPrmSDOrdinalAssoc).InputList = 0)//liste target/input ?
  then
   begin
    result:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caContinue));
    result:= result and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue));
   end
  else result:= (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue)); 
end;

destructor TOpSDOrdinalAssoc.destroy;
begin
 self.destroyLstVec();
 inherited;
end;

procedure TOpSDOrdinalAssoc.destroyLstVec;
begin
 if (FLstVecTarget <> nil) then FreeAndNil(FLstVecTarget);
 if (FLstVecInput <> nil) then FreeAndNil(FLstVecInput);
end;

function TOpSDOrdinalAssoc.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDOrdinalAssoc;
end;

function TOpSDOrdinalAssoc.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>' + self.getMethodDescription() + ' for ordinal attributes </H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + self.FLstStat.getHTMLResults((self.PrmOp as TOpPrmSDOrdinalAssoc).ShowDetails);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDOrdinalAssoc.InitializeListStat;
begin
 FLstStat:= TLstSDOrdinalAssoc.Create(nil,nil);
end;

procedure TOpSDOrdinalAssoc.RebuildStatDes;
var vecI, vecJ: TVecDistinctValues;
    i,j: integer;
begin
 //supprimer les listes internes de valeurs uniques
 self.destroyLstVec();
 //reconstruire
 FLstVecInput:= TObjectList.Create(TRUE);
 for j:= 0 to pred(workdata.LstAtts[asInput].Count) do
  FLstVecInput.Add(TVecDistinctValues.create(workdata.LstAtts[asInput].Attribute[j]));
 //pour les targets ?
 if ((self.PrmOp as TOpPrmSDOrdinalAssoc).InputList = 1)
  then FLstVecTarget:= nil
  else
   begin
    FLstVecTarget:= TObjectList.Create(TRUE);
    for j:= 0 to pred(workdata.LstAtts[asTarget].Count) do
     FLstVecTarget.Add(TVecDistinctValues.create(workdata.LstAtts[asTarget].Attribute[j]));
   end;
 //ok, on peut construire les stats  produire maintenant
 if (FLstVecTarget <> nil)
  then
   begin
    for i:= 0 to pred(FLstVecTarget.Count) do
     begin
      vecI:= FLstVecTarget[i] as TVecDistinctValues;
      for j:= 0 to pred(FLstVecInput.Count) do
       begin
        vecJ:= FLstVecInput[j] as TVecDistinctValues;
        if (vecI <> vecJ) then FLstStat.AddStat(getClassComputation().create(vecI,vecJ));
       end;
     end;
  end
 else
  begin
   for i:= 0 to FLstVecInput.Count-2 do
    for j:= succ(i) to FLstVecInput.Count-1 do
     FLstStat.AddStat(getClassComputation().create(FLstVecInput[i] as TVecDistinctValues,FLstVecInput[j] as TVecDistinctValues))
  end;
end;

end.
