(**********************************************************************)
(* UCalcFactRotateFactors.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Rotation des facteurs de l'ACP -- Directement adapt d'OPENSTAT)
@author(Ricco)
@created(08/07/2006)

Rotation des axes factoriels, VARIMAX, QUARTIMAX, etc.
Cette unit est directement adapte des units d'OPENSTAT (MILLER -- http://www.statpages.org/miller/openstat/OpenStat.htm)

}

unit UCalcFactRotateFactors;

interface

USES
    Matrices;

TYPE
//type de rotation
TEnumRotateFactors = (rfVarimax, rfQuartimax);

CONST
STR_FACTOR_ROTATION: array[TEnumRotateFactors] of string = ('VARIMAX','QUARTIMAX');

//********************************************************************************************
//**** procdure de rotation, en entre la matrice de corrlation des variables avec les axes,
//**** le nombre de variables et le nombre de facteurs  modifier
//********************************************************************************************

//rotation VARIMAX
procedure VARIMAX(VAR v: PMatrix; nbVar, nbFactor: integer);

//rotation QUARTIMAX
procedure QUARTIMAX(VAR v: PMatrix; nbVar, nbFactor: integer);

implementation

USES
    Math, ULogFile;

//rotation VARIMAX
procedure VARIMAX(VAR v: PMatrix; nbVar, nbFactor: integer);
label label1;
var
   pi : double;
   A, b, C : PVector;
   i, j, k, M, N, minuscount : integer;
   a1, b1, c1, c2, c3, c4, d1, x1, x2, Y, s1, Q, TotalPercent, t : double;
begin
     pi := 3.14159265358979;
     t := nbVar;
     //SetLength(A,NoVariables);
     DimVector(A,nbVar);
     //SetLength(b,NoVariables);
     DimVector(b,nbVar);
     //SetLength(C,NoVariables);
     DimVector(C,nbVar);

    // calculate proportion of variance accounted for by each factor
    //before rotation
    for j := 1 to nbFactor do
    begin
         A[j] := 0.0;
         for i := 1 to nbVar do  A[j] := A[j] + (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;

    for i := 1 to nbVar do
    begin
         b[i] := 0.0;
         //High_Factor[i] := 0;
    end;

    // Reflect factors 180 degrees if more negative than positive loadings
     for j := 1 to nbFactor do
     begin
          minuscount := 0;
          for i := 1 to nbVar do
          begin
              if ( v[i,j] < 0) then minuscount := minuscount + 1;
          end;
          if ( minuscount > (nbVar / 2)) then
          begin
             for i := 1 to nbVar do  v[i,j] := v[i,j] * -1.0;
          end;
     end;

     // normalize rows of v
     for i := 1 to nbVar do
     begin
         for j := 1 to nbFactor do
         begin
             b[i] := b[i] + (v[i,j] * v[i,j]);
         end;
         b[i] := sqrt(b[i]);
         for j := 1 to nbFactor do v[i,j] := v[i,j] / b[i];
     end;

label1:
     k := 0;
     for M := 1 to nbFactor do
     begin
         for N := M to nbFactor do
         begin
              if ( M <> N) then // compute angle of rotation
              begin
                   for i := 1 to nbVar do
                   begin
                        A[i] := (v[i,M] * v[i,M]) - (v[i,N] * v[i,N]);
                        C[i] := 2.0 * v[i,M] * v[i,N];
                   end;
                   a1 := 0.0;
                   for i := 1 to nbVar do a1 := a1 + A[i];
                   b1 := 0.0;
                   for i := 1 to nbVar do b1 := b1 + C[i];
                   c1 := 0.0;
                   for i :=  1 to nbVar do c1 := c1 + (A[i] * A[i]);
                   c2 := 0.0;
                   for i := 1 to nbVar do c2 := c2 + (C[i] * C[i]);
                   c3 := c1 - c2;
                   d1 := 0.0;
                   for i := 1 to nbVar do d1 := d1 + A[i] * C[i];
                   d1 := 2 * d1;
                   x1 := d1 - 2.0 * a1 * b1 / t;
                   x2 := c3 - ((a1 * a1) - (b1 * b1)) / t;
                   Y := ArcTan(x1 / x2);
                   if ( x2 < 0) then
                   begin
                        if ( x1 >= 0.0) then Y := Y + 2.0 * pi;
                        Y := Y - pi;
                   end;
                   Y := Y / 4.0;
                   //if (fabs(Y) >= 0.0175) // rotate pair of axes
                   if ( abs(Y) >= 0.000001) then
                   begin
                        c4 := cos(Y);
                        s1 := sin(Y);
                        k := 1;
                        for i := 1 to nbVar do
                        begin
                             Q := v[i,M] * c4 + v[i,N] * s1;
                             v[i,N] := v[i,N] * c4 - v[i,M] * s1;
                             v[i,M] := Q;
                        end;
                   end; // if y
              end;  // if m <> n
         end; // next n
    end;  // next m
    if ( k > 0) then goto label1;
    // denormalize rows of v
    for j := 1 to nbFactor do
    begin
         for i := 1 to nbVar do v[i,j] := v[i,j] * b[i];
         A[j] := 0.0;
         for i := 1 to nbVar do A[j] := A[j] + (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;
    for i := 1 to nbVar do b[i] := (b[i] * b[i]) * 100.0;

    // Reflect factors 180 degrees if more negative than positive loadings
    for j := 1 to nbFactor do
    begin
         minuscount := 0;
         for i := 1 to nbVar do
         begin
              if ( v[i,j] < 0) then  minuscount := minuscount + 1;
         end;
         if ( minuscount > (nbVar / 2)) then
         begin
            for i := 1 to nbVar do v[i,j] := v[i,j] * -1.0;
         end;
    end;
    // recalculate proportion of variance accounted for by each factor
    for j := 1 to nbFactor do
    begin
         A[j] := 0.0;
         for i := 1 to nbVar do A[j] := A[j] + (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;

    // clean up heap
    (*
    C := nil;
    b := nil;
    A := nil;
    *)
    DelVector(C,nbVar);
    DelVector(b,nbVar);
    DelVector(A,nbVar);
end; //end of Varimax ;

//rotation QUARTIMAX
procedure QUARTIMAX(VAR v: PMatrix; nbVar, nbFactor: integer);
var
   i, j, M, N, minuscount, NoIters : integer;
   A, b, C : PVector;
   //High_Factor : IntDyneVec;
   c4, s1, Q, NewQ, t : double;
   theta, tan4theta, ssqrp, ssqrj, prodjp, numerator, denominator : double;
   done : boolean;
begin
     (*
     SetLength(A,NoVariables);
     SetLength(b,NoVariables);
     SetLength(C,NoVariables);
     *)
     DimVector(A,nbVar);
     DimVector(b,nbVar);
     DimVector(C,nbVar);
     //SetLength(High_Factor,NoVariables);
     
     NoIters := 0;

    // calculate proportion of variance accounted for by each factor
    //before rotation
    t := nbVar;
    for j := 1 to nbFactor do
    begin
         A[j] := 0.0;
         for i := 1 to nbVar do A[j] := A[j] + (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;

    for i := 1 to nbVar do
    begin
         b[i] := 0.0;
         //High_Factor[i] := 0;
    end;

    // Reflect factors 180 degrees if more negative than positive loadings
     for j := 1 to nbFactor do
     begin
          minuscount := 0;
          for i := 1 to nbVar do
          begin
              if  v[i,j] < 0 then  minuscount := minuscount + 1;
          end;
          if minuscount > (nbVar / 2) then
          begin
             for i := 1 to nbVar do v[i,j] := v[i,j] * -1.0;
          end;
     end;

     t := nbVar;
     // normalize rows of v
     for i := 1 to nbVar do
     begin
         for j := 1 to nbFactor do
         begin
             b[i] := b[i] + (v[i,j] * v[i,j]);
         end;
         b[i] := sqrt(b[i]);
     end;

     done := false;
     Q := 0.0;
     for i := 1 to nbVar do
         for j := 1 to nbFactor do
             Q := Q + Power(v[i,j],4.0);
     while (not done) do
     begin
         for M := 1 to nbFactor-1 do
         begin
             for N := M + 1 to nbFactor do
             begin
                 // compute angle of rotation for this pair of factors
                 numerator := 0.0;
                 denominator := 0.0;
                 for i := 1 to nbVar do
                 begin
                     ssqrp := v[i,M] * v[i,M];
                     ssqrj := v[i,N] * v[i,N];
                     prodjp := 2.0 * v[i,M] * v[i,N];
                     numerator := numerator + prodjp * (ssqrp - ssqrj);
                     denominator := denominator + (Power(ssqrp - ssqrj,2.0) - Power(prodjp,2));
                 end;
                 tan4theta := (2.0 * numerator) / denominator;
                 theta := ArcTan(tan4theta) / 4.0;
                 c4 := cos(theta);
                 s1 := sin(theta);
                 // transform factor loadings
                 for i := 1 to nbVar do
                begin
                    v[i,M] := v[i,M] * c4 + v[i,N] * s1;
                    v[i,N] := v[i,N] * c4 - v[i,M] * s1;
                end;
             end; // next n
         end;  // next m
         NewQ := 0.0;
         for i := 1 to nbVar do
             for j := 1 to nbFactor do
                 NewQ := NewQ + Power(v[i,j],4.0);
         if (abs(Q - NewQ) < 0.00001) then done := true;
         if (nbFactor < 3) then done := true;
         if (not done) then
         begin
            NoIters := NoIters + 1;
            if (NoIters > 25) then
            begin
               (*
               outline := 'Quartimax failed to converge in 25 iterations.';
               OutPutFrm.RichEdit.Lines.Add(outline);
               *)
               TraceLog.WriteToLogFile('[WARNING] !!! Factors Rotation -- QUARTIMAX exceed 25 iterations');
               done := true;
            end;
            Q := NewQ;
         end;
     end; // while not done
{
    // denormalize rows of v
    for ( j := 0; j < nbFactor; j++)
    begin
         for ( i := 0; i < nbVar; i++) v[i,j] *= b[i];
         A[j] := 0.0;
         for ( i := 0; i < nbVar; i++) A[j] += (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;
}
    for i :=  1 to nbVar do b[i] := (b[i] * b[i]) * 100.0;
    (*
    if (FactorFrm.SortBtn.Checked) then
        SORT_LOADINGS(v, nbVar, nbFactor, High_Factor, A, b, RowLabels, order);
    *)
    
    // Reflect factors 180 degrees if more negative than positive loadings
    for j := 1 to nbFactor do
    begin
         minuscount := 0;
         for i := 1 to nbVar do
         begin
              if ( v[i,j] < 0) then minuscount := minuscount  + 1;
         end;
         if ( minuscount > (nbVar / 2)) then
         begin
            for i := 1 to nbVar do v[i,j] := v[i,j] * -1.0;
         end;
    end;
    // recalculate proportion of variance accounted for by each factor
    for j := 1 to nbFactor do
    begin
         A[j] := 0.0;
         for i := 1 to nbVar do A[j] := A[j] + (v[i,j] * v[i,j]);
         A[j] := A[j] / t * 100.0;
    end;

    // clean up heap
     //High_Factor := nil;
     DelVector(C,nbVar);
     DelVector(b,nbVar);
     DelVector(A,nbVar);
     (*
     C := nil;
     b := nil;
     A := nil;
     *)
end; // Quartimax end;
//----------------------------------------------------------------------------


end.
