(********************************************************************)
(* UCompFactAnalysisPCA.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Analyse en Composantes Principales)
@author(Ricco)
@created(12/01/2004)
}
unit UCompFactAnalysisPCA;

interface

USES
        Forms,Classes,
        IniFiles,
        Matrices,
        UCalcStatDes,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UDatasetDefinition;

TYPE
        {gnrateur de composant ACP}
        TMLGenCompFactPCA = class(TMLGenComp)
                            protected
                            procedure   GenCompInitializations(); override;
                            public
                            function    GetClassMLComponent: TClassMLComponent; override;
                            end;

        {composant ACP}
        TMLCompFactPCA = class(TMLCompFactAnalysis)
                         protected
                         function    getClassOperator: TClassOperator; override;
                         end;

        {oprateur ACP, on travaille sur les INPUT}
        TOpFactPCA = class(TOpFactAnalysis)
                     private
                     {Trace de la matrice  diagonaliser}
                     FTrace: double;
                     {Nombre d'axes  traiter}
                     FNbAxis: integer;
                     {StatDes}
                     FLstStats: TLstCalcStatDesContinuous;
                     {dimension des matrices}
                     FDimAnalysis: integer;
                     {matrice de corrlation}
                     FCorrMat: PMatrix;
                     {matrice, corrlation des variables avec les axes}
                     FLoadings: PMatrix;
                     {vecteur des valeurs propres}
                     FEgVal: PVector;
                     {matrice des vecteurs propres}
                     FEgVec: PMatrix;
                     {calculer les stats descriptives}
                     procedure   CalcStats();
                     {construire la matrice de corrlation}
                     procedure   BuildVCV();
                     {calculer les valeurs et vecteurs propres}
                     procedure   BuildEigen();
                     {calculer les corrlations des variables avec les axes}
                     procedure   CalcLoadings();
                     protected
                     {dtruire les matrices et vecteurs}
                     procedure   DelPrivateMatrix(); override;
                     function    getClassParameter: TClassOperatorParameter; override;
                     function    CheckAttributes(): boolean; override;
                     function    CoreExecute(): boolean; override;
                     public
                     constructor Create(AOwner: TObject); override;
                     destructor  Destroy; override;
                     function    getHTMLResultsSummary(): string; override;
                     function    getHTMLLoadings(loadings: PMatrix; nbFactors: integer): string;
                     function    getHTMLRegCoef(eigVec: PMatrix; nbFactors: integer): string;
                     {effectuer les projections sur les axes factoriels}
                     procedure   SetProjections(atts: TLstAttributes; coefs: PMatrix; nbFactors: integer; prefix: string = 'PCA'; number: integer = 1);
                     //new -- 04/02/2007 -- calculer les cos2 et ctr
                     procedure   setCos2Ctr(atts: TLstAttributes; nbFactors: integer; prefix: string = 'PCA'; number : integer = 1);
                     property    Trace: double read FTrace;
                     property    Loadings: PMatrix read FLoadings;
                     property    EgVec: PMatrix read FEgVec;
                     property    DimAnalysis: integer read FDimAnalysis;
                     property    NbAxis: integer read FNbAxis;
                     property    MatCorr: PMatrix read FCorrMat;
                     end;

        {paramtre oprateur ACP}
        TOpPrmFactPCA = class(TOpPrmFactAnalysis)
                        private
                        {nombre d'axes  crer, si -1 on les gnre tous}
                        FNbAxisCreated: integer;
                        //new -- 04/02/2007 -- projeter les COS et CTR par individu, df. false
                        FCos2Ctr: boolean;
                        protected
                        function    CreateDlgParameters(): TForm; override;
                        procedure   SetDefaultParameters(); override;
                        public
                        function    getHTMLParameters(): string; override;
                        procedure   LoadFromStream(prmStream: TStream); override;
                        procedure   SaveToStream(prmStream: TStream); override;
                        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                        property    NbAxisCreated: integer read FNbAxisCreated write FNbAxisCreated;
                        property    Cos2Ctr: boolean read FCos2Ctr write FCos2Ctr;
                        end;

implementation

USES
        SysUtils,Math,
        Contnrs,
        Windows,
        UDatasetImplementation, UCalcMatrixToAttributes,
        UCalcMatrixAdditionalFunctions, UConstConfiguration,
        UDlgOpPrmPCA, UStringsResources, UStringAddBuffered, ULogFile,
        UCalcFactRotateFactors;

{ TMLGenCompFactPCA }

procedure TMLGenCompFactPCA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
 //FMLNumIcon:= 10;
 //FMLCompName:= str_comp_name_fact_pca;
 //FMLBitmapFileName:= 'MLFactorialPCA.bmp';
end;

function TMLGenCompFactPCA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactPCA;
end;

{ TMLCompFactPCA }

function TMLCompFactPCA.getClassOperator: TClassOperator;
begin
 result:= TOpFactPCA;
end;

{ TOpFactPCA }

procedure TOpFactPCA.BuildEigen;
begin
 GetEigensFromSymetricMatrix(FCorrMat,FDimAnalysis,FEgVal,FEgVec);
end;

procedure TOpFactPCA.SetProjections(atts: TLstAttributes; coefs: PMatrix; nbFactors: integer; prefix: string = 'PCA'; number: integer = 1);
var i,j,k: integer;
    newAtt: TAttContinue;
    //att: TAttribute;
    //stat: TCalcStatDesContinuous;
    attSize: integer;
    s: TTypeContinue;
    //new -- 18/03/2006 -- acclrer le temps de traitement
    //en crant des tableaux temporaires
    avg,std: array of double;
begin
 atts.Clear;
 attSize:= WorkData.LstAtts[asAll].Size;
 //new -- 18/03/2006 -- crer les tableaux temporaires, une fois pour toutes
 setlength(avg,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  avg[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).Average;
 setlength(std,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA)
   then std[j]:= 1.0
   //else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDev;
   //new -- 06/07/2006 -- on utilise l'cart-type de la population pour les calculs (cf. matrice de corrlation)
   else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop;
 //>> suite...
 //ok on peut lancer les calculs
 for k:= 1 to nbFactors do
  begin
   newAtt:= TAttContinue.Create(format('%s_%d_Axis_%d',[prefix,number,k]),attSize);
   //pour chaque individu i
   for i:= 1 to attSize do
    begin
     s:= 0.0;
     for j:= 1 to FDimAnalysis do
      begin
       //att:= WorkData.LstAtts[asInput].Attribute[pred(j)];
       //stat:= TCalcStatDesContinuous(FLstStats.Stat(pred(j)));
       //new -- pas de division par zro please
       (*
       if (abs(stat.StdDev) < EPSILON_STD_DEV_FOR_PCA)
        then s:= s + FEgVec^[j]^[k]*(att.cValue[i]-stat.Average)/1.0
        else s:= s + FEgVec^[j]^[k]*(att.cValue[i]-stat.Average)/stat.StdDev;
       *)
       //new -- 18/03/2006 -- a devrait amliorer (oui !!! division par 8 du temps de calcul...)
       s:= s + coefs^[j]^[k]*(WorkData.LstAtts[asInput].Attribute[pred(j)].cValue[i] - avg[j]) / std[j];
      end;
     newAtt.cValue[i]:= s;
    end;
   atts.Add(newAtt);
  end;
 //>>
 finalize(avg);
 finalize(std);
end;

procedure TOpFactPCA.BuildVCV;
var i: integer;
begin
 FCorrMat:= BuildMatVCV(workdata.Examples,Workdata.lstAtts[asInput],vcvNormNormalized,FLstStats);
 FTrace:= 0.0;
 for i:= 1 to self.FDimAnalysis do
  FTrace:= FTrace + FCorrMat^[i]^[i];
end;

function TOpFactPCA.CoreExecute: boolean;
var tps: cardinal;
begin
 result:= TRUE;
 TRY
 //dtruire les matrices internes
 self.DelPrivateMatrix();
 //calculs
 tps:= GetTickCount();
 self.CalcStats();
 TraceLog.WriteToLogFile(format('[PCA] ok calc stat = %d ms.',[GetTickCount()-tps]));
 tps:= GetTickCount();
 self.BuildVCV();
 TraceLog.WriteToLogFile(format('[PCA] ok build vcv = %d ms.',[GetTickCount()-tps]));
 tps:= GetTickCount();
 self.BuildEigen();
 TraceLog.WriteToLogFile(format('[PCA] ok compute eigens = %d ms.',[GetTickCount()-tps]));
 //nombre d'axes cr = min de demand ou nombre de variables
 FNbAxis:= MIN((self.PrmOp as TOpPrmFactPCA).NbAxisCreated,FDimAnalysis);
 tps:= GetTickCount();
 self.CalcLoadings();
 TraceLog.WriteToLogFile(format('[PCA] ok compute loadings = %d ms.',[GetTickCount()-tps]));
 tps:= GetTickCount();
 self.SetProjections(self.GenAtts,FEgVec,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 TraceLog.WriteToLogFile(format('[PCA] ok projection = %d ms.',[GetTickCount()-tps]));
 //new -- 04/02/2007 -- le cas cheant, calculer les contributions et les cos2  partir des projections
 if ((self.PrmOp as TOpPrmFactPCA).Cos2Ctr = true)
  then self.setCos2Ctr(self.GenAtts,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 EXCEPT
 self.DelPrivateMatrix();
 result:= FALSE;
 END;
end;

procedure TOpFactPCA.DelPrivateMatrix;
begin
 TRY
 //dtruire les matrices internes
 if assigned(FCorrMat) then delMatrix(FCorrMat,FDimAnalysis,FDimAnalysis);
 if assigned(FEgVec) then delMatrix(FEgVec,FDimAnalysis,FDimAnalysis);
 if assigned(FEgVal) then delVector(FEgVal,FDimAnalysis);
 if assigned(FLoadings) then delMatrix(FLoadings,FDimAnalysis,FNbAxis);
 FLstStats.FreeAll;
 EXCEPT
 END;
end;

function TOpFactPCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactPCA;
end;

function TOpFactPCA.getHTMLResultsSummary: string;
var s: string;
    sum: double;
    tot: double;
    i: integer;
    buf: TBufString;
begin
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //#ToDo1 - affichage des rsultats cf. SPAD et STATISTICA
 //afficher les stats des.
 //afficher la matrice de corrlation
 //afficher les valeurs propres
 //la trace de la matrice
 s:= '<P><H3>Eigen values</H3>';
 s:= s + format('<H5>Matrix trace = %.2f</H5>',[FTrace]);
 s:= s+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH>Eigen value</TH><TH>% explained</TH><TH width="200">Histogram</TH><TH>% cumulated</TH></TR>';
 buf.AddStr(s);
 tot:= 1.0*FDimAnalysis;
 sum:= 0.0;
 for i:= 1 to FDimAnalysis do
  begin
   sum:= sum+FEgVal^[i];
   if (tot>0)
    then
     begin
      s:= HTML_TABLE_COLOR_DATA_GRAY+
         format('<TH>%d</TH><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD><TD align=right>%.2f%s</TD><TD>%s</TD><TD align=right>%.2f%s</TD></TR>',
                [i,FEgVal^[i],100.0*FEgVal^[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*FEgVal^[i]/tot)),100.0*sum/tot,'%'])+'</TR>';
      buf.AddStr(s);
     end;
  end;
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>Tot.</TH><TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>',[sum]));
 buf.AddStr('</table>');

 //**
 //afficher la corrlation des attributs avec les axes
 //**
 s:= '<P><H3>Factor Loadings [Communality Estimates]</H3>';
 buf.AddStr(s+self.getHTMLLoadings(FLoadings,FNbAxis));

 //**
 //new -- 13/02/2005 -- afficher les vecteurs propres
 //**
 s:= '<P><H3>Eigen vectors -- Factor Scores</H3>';
 buf.AddStr(s+self.getHTMLRegCoef(FEgVec,FNbAxis));

 //envoyer la sauce
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

constructor TOpFactPCA.Create(AOwner: TObject);
begin
 inherited Create(AOwner);
 FLstStats:= TLstCalcStatDesContinuous.Create(NIL,NIL);
end;

procedure TOpFactPCA.CalcStats;
begin
 FDimAnalysis:= WorkData.LstAtts[asInput].Count;
 FLstStats.RebuildStatDes(WorkData.LstAtts[asInput],Workdata.Examples);
end;

destructor TOpFactPCA.Destroy;
begin
 inherited Destroy;
 FLstStats.Free;
end;

function TOpFactPCA.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asInput].Count>0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue));
end;

procedure TOpFactPCA.CalcLoadings;
var i,j: integer;
    value: double;
begin
 dimMatrix(FLoadings,FDimAnalysis,FNbAxis);
 for i:= 1 to FDimAnalysis do
  begin
   for j:= 1 to FNbAxis do
    begin
      value:= sqrt(FEgVal^[j])*FEgVec^[i]^[j];
      FLoadings[i,j]:= value;
    end;
  end;
end;

function TOpFactPCA.getHTMLLoadings(loadings: PMatrix; nbFactors: integer): string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    vExpl: array of double;
    strColor: string;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 setLength(vExpl,succ(nbFactors));
 for j:= 1 to nbFactors do
  begin
   buf.AddStr(format('<TH align=center colspan=2>Axis_%d</TH>',[j]));
   //en profiter pour calculer la variance explique
   vExpl[j]:= 0.0;
   for i:= 1 to FDimAnalysis do
    vExpl[j]:= vExpl[j] + loadings[i,j] * loadings[i,j];
  end;
 buf.AddStr('</TR>');
 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>-</TH>');
 for j:= 1 to nbFactors do
  buf.AddStr('<TD align=center width=60>Corr.</TD><TD align=center width=60>% (Tot. %)</TD>');
 //les valeurs
 for i:= 1 to FDimAnalysis do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[WorkData.LstAtts[asInput].Attribute[pred(i)].Name]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to nbFactors do
    begin
      TRY
      value:= loadings[i,j];
      sum:= sum + value * value;
      //mettre en valeur si % de la variable > % de l'axe
      if (abs(value) > 0.7)
       then strColor:= HTML_BGCOLOR_DATA_RED
       else strColor:= HTML_BGCOLOR_DATA_GRAY; 
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60 %s>%.4f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[strColor,value,100.0*value*value,'%',100.0*sum,'%']));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Var. Expl.']));
 sum:= 0.0;
 for j:= 1 to nbFactors do
  begin
   sum:= sum + vExpl[j];
   buf.AddStr(format('<TD align=right width=60>%.4f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[vExpl[j],100.0*vExpl[j]/FTrace,'%',100.0*sum/FTrace,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
 Finalize(vExpl);
end;

function TOpFactPCA.getHTMLRegCoef(eigVec: PMatrix;
  nbFactors: integer): string;
var  buf: TBufString;
     s: string;
     i,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //en-tte pour chaque axe factoriel demand
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>Attribute</TH>');
 //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
 s:= format('<TH width=60 align=right %s>Mean</TH><TH width=60 align=right %s>Std-dev</TH>',[HTML_BGCOLOR_HEADER_GREEN,HTML_BGCOLOR_HEADER_GREEN]);
 for j:= 1 to nbFactors do
  s:= s+format('<TH width=60 align=right>Axis_%d</TH>',[j]);
 s:= s+'</TR>';
 //le contenu
 for i:= 1 to FDimAnalysis do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[WorkData.LstAtts[asInput].Attribute[pred(i)].Name]);
   //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
   s:= s + format('<TD align=right %s>%.6f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).Average]);
   s:= s + format('<TD align=right %s>%.6f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop]);
   //suite...
   for j:= 1 to nbFactors do
    begin
      TRY
      s:= s+format('<TD align=right>%.6f</TD>',[eigVec^[i]^[j]]);
      EXCEPT
      END;
    end;
   s:= s+'</TR>';
  end;
 s:= s+'</table>';
 //dans le buffer...
 buf.AddStr(s);
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

procedure TOpFactPCA.setCos2Ctr(atts: TLstAttributes; nbFactors: integer;
  prefix: string; number: integer);
var lst: TObjectList;
    newAtt: TAttribute;
    factor: TAttribute;
    attSize: integer;
    k,i,j: integer;
    poids,d: single;
    avg, std: array of single;
    dist: array of single;
begin
 attSize:= WorkData.LstAtts[asAll].Size;
 poids:= 100.0/(1.0*attSize);//poids en %
 lst:= TObjectList.Create(false);//non propritaire
 //calcul des CTR et COS  partir des atts... (coordonnes sur les axes factoriels)
 //**************
 //calcul des CTR
 //**************
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_CTR_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   //calcul Saporta --> pp.175, CTR en % puisque poids en %
   if (FEgVal[k] > 0.0)
    then
     begin
      for i:= 1 to attSize do newAtt.cValue[i]:= poids * MATH.Power(factor.cValue[i],2.0) / FEgVal[k];
     end
    else for i:= 1 to attSize do newAtt.cValue[i]:= 0.0;
  end;
 //***************
 //calcul des COS
 //***************
 //crer les tableaux temporaires, une fois pour toutes
 setlength(avg,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do avg[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).Average;
 setlength(std,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA)
   then std[j]:= 1.0
   else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop;//new -- 06/07/2006 -- on utilise l'cart-type de la population pour les calculs (cf. matrice de corrlation)
 //distance au centre de gravit des individus
 setLength(dist,succ(attSize));
 for i:= 1 to attSize do
  begin
   d:= 0.0;
   for j:= 1 to FDimAnalysis do
    d:= d + MATH.POWER((WorkData.LstAtts[asInput].Attribute[pred(j)].cValue[i] - avg[j]) / std[j], 2.0);
   dist[i]:= d;
  end;
 //fin des tableaux temporaires avg, std
 finalize(avg);
 finalize(std);
 //calculer le COS2
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_COS2_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   for i:= 1 to attSize do
    begin
     if (dist[i] > 0.0)
      then newAtt.cValue[i]:= MATH.Power(factor.cValue[i],2.0)/dist[i]
      else newAtt.cValue[i]:= 0.0;
    end;
  end;
 finalize(dist);
 //flush...
 for j:= 0 to pred(lst.Count) do
  atts.Add(lst[j] as TAttribute);
 //>>
 lst.Free();
end;

{ TOpPrmFactPCA }

function TOpPrmFactPCA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmPCA.CreateFromOpPrm(self);
end;

function TOpPrmFactPCA.getHTMLParameters: string;
var s: string;
begin
 s:= '<P><B>Generated axis : </B>';
 if (FNbAxisCreated<0)
  then s:= s+'all'
  else s:= s+IntToStr(FNbAxisCreated);
 //new -- 04/02/2007
 s:= s + '<P><B>Compute COS2 and CTR : </B>' + IntToStr(ord(FCos2Ctr));
 result:= s;
end;

procedure TOpPrmFactPCA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbAxisCreated:= prmINI.ReadInteger(prmSection,'nb_axis',FNbAxisCreated);
 FCos2Ctr:= prmINI.ReadBool(prmSection,'cos2_ctr',FCos2Ctr);
end;

procedure TOpPrmFactPCA.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNBAxisCreated,sizeof(FNBAxisCreated));
 prmStream.ReadBuffer(FCos2Ctr,sizeof(FCos2Ctr));
end;

procedure TOpPrmFactPCA.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',FNbAxisCreated);
 prmINI.WriteBool(prmSection,'cos2_ctr',FCos2Ctr);
end;

procedure TOpPrmFactPCA.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNBAxisCreated,sizeof(FNBAxisCreated));
 prmStream.WriteBuffer(FCos2Ctr,sizeof(FCos2Ctr));
end;

procedure TOpPrmFactPCA.SetDefaultParameters;
begin
 FNbAxisCreated:= 10;
 FCos2Ctr:= false;
end;

initialization
 CLASSES.RegisterClass(TMLGenCompFactPCA);
end.
