(**********************************************************************)
(* UCompFactRotateFactors.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Rotation des facteurs d'une analyse en composantes principales)
@author(Ricco)
@created(08/07/2006)
}

unit UCompFactRotateFactors;

interface

USES
    UCompDefinition,
    UCompManageDataset,
    UOperatorDefinition,
    UCompFactAnalysisPCA,
    UCalcFactRotateFactors,
    Classes, Forms, Inifiles,
    Matrices;

TYPE
    {gnrateur de composant}
    TMLGenCompFactRotation = class(TMLGenComp)
      protected
      procedure   GenCompInitializations(); override;
      public
      function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant}
    TMLCompFactRotation = class(TMLCompLocalData)
      protected
      function    getClassOperator: TClassOperator; override;
      public
      //ne pourra tre connect que sous une analyse en composantes principales et ses drivs
      class function isConnectableTo(destination: TMLComponent): boolean; override;
    end;

    {oprateur}
    TOpFactRotation = class(TOpLocalData)
      private
      {oprateur d'ACP  traiter}
      FOpACP: TOpFactPCA;
      {nouveau loadings}
      FNewLoadings: PMatrix;
      {nouveaux vecteurs propres}
      FNewEgVec: PMatrix;
      {nombre de facteurs  traiter}
      FNbFactors: integer;
      protected
      function    getClassParameter: TClassOperatorParameter; override;
      function    CoreExecute(): boolean; override;
      //recalculer les nouveaux coefficients de rgression (ou les nouveaux vecteurs propres... c'est la mme chose)
      function    computeNewRegCoefficients(): boolean;
      //raliser la projection avec les nouveaux coefs.
      function    setProjections(): boolean;
      public
      //en profiter pour rcuprer l'oprateur d'ACP 
      constructor Create(AOwner: TObject); override;
      destructor  Destroy(); override;
      function    getHTMLResultsSummary(): string; override;
    end;

    {paramtrage}
    TOpFactRotationParameter = class(TOperatorParameter)
      private
      {nombre de facteurs  traiter}
      FNbFactors: integer;
      {type de rotation  mettre en place}
      FRotation: TEnumRotateFactors;
      protected
      function    CreateDlgParameters(): TForm; override;
      procedure   SetDefaultParameters(); override;
      public
      function    getHTMLParameters(): string; override;
      procedure   LoadFromStream(prmStream: TStream); override;
      procedure   SaveToStream(prmStream: TStream); override;
      procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
      procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
      property    NbFactors: integer read FNbFactors write FNbFactors;
      property    Rotation: TEnumRotateFactors read FRotation write FRotation;
    end;

implementation

uses
    UCalcMatrixAdditionalFunctions, UConstConfiguration,
    Sysutils, UDlgOpPrmRotateFactors;

{ TMLGenCompFactRotation }

procedure TMLGenCompFactRotation.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactRotation.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactRotation; 
end;

{ TMLCompFactRotation }

function TMLCompFactRotation.getClassOperator: TClassOperator;
begin
 result:= TOpFactRotation; 
end;

class function TMLCompFactRotation.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 result:= destination.inheritsFrom(TMLCompFactPCA);
end;

{ TOpFactRotation }

function TOpFactRotation.computeNewRegCoefficients(): boolean;
var ok: boolean;
    invCorr: PMatrix;
    i,j: integer;
    sum: double;
begin
 //inverser la matrice de corrlation
 Matrices.DimMatrix(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
 //ok:= (Matrices.InvMat(FOpACP.MatCorr,1,FOpACP.DimAnalysis,invCorr) = Matrices.MAT_OK);
 Matrices.CopyMatrix(invCorr,FOpACP.MatCorr,1,1,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
 ok:= UCalcMatrixAdditionalFunctions.SVDInverse(invCorr,FOpACP.DimAnalysis);
 if ok
  then
   begin
    try
    //produit matriciel inverse de la matrice de corrlation avec les nouveaux loadings
    UCalcMatrixAdditionalFunctions.MatrixProd(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis,FNewLoadings,FOpACP.DimAnalysis,FNbFactors,FNewEgVec);

    //normer les vecteurs ???  la diffrence des logiciels U.S.
    //calculer la norme
    for j:= 1 to FNbFactors do
     begin
      sum:= 0.0;
      for i:= 1 to FOpACP.DimAnalysis do
       sum:= sum + FNewEgVec[i,j] * FNewEgVec[i,j];
      //passer  la racine carre
      sum:= SQRT(sum);
      //normer maintenant
      if (sum > 0)
       then
        begin
         for i:= 1 to FOpACP.DimAnalysis do
          FNewEgVec[i,j]:= FNewEgVec[i,j] / sum;
        end;
     end;
    //
    
    except
    ok:= false;
    end;
   end;
 Matrices.delMatrix(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
 result:= ok;
end;

function TOpFactRotation.CoreExecute: boolean;
var ok: boolean;
begin
 //si c'est matrices sont assignes, on doit bien sr utiliser l'ancienne valeur de NbFactors pour les librer
 //si NbFactors = 0, ces matrices n'ont pas pu tre assignes, donc pas de soucis potentiels...
 if assigned(FNewLoadings) then DelMatrix(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 if assigned(FNewEgVec) then DelMatrix(FNewEgVec,FOpACP.DimAnalysis,FNbFactors);
 TRY
 //rcuprer les nouvelles valeurs (ventuellement) de NbFactors
 FNbFactors:= (self.PrmOp as TOpFactRotationParameter).FNbFactors;
 //copier la matrice
 Matrices.DimMatrix(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 Matrices.CopyMatrix(FNewLoadings,FOpACP.Loadings,1,1,FOpACP.DimAnalysis,FNbFactors);
 //effectuer la rotation selon le paramtrage demand
 case (self.PrmOp as TOpFactRotationParameter).FRotation of
 rfQuartimax: QUARTIMAX(FNewLoadings,FOpACP.DimAnalysis,FNbFactors)
 else VARIMAX(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 end;
 //recalculer les nouveaux coefficients de projection
 ok:= self.computeNewRegCoefficients();
 //la projection -- si c'est pas ok dans la ligne du dessus, la projection n'est pas tente
 ok:= ok and self.setProjections(); 
 //
 result:= ok;
 EXCEPT
 result:= false;
 END;
end;

constructor TOpFactRotation.Create(AOwner: TObject);
begin
 inherited create(AOwner);
 FOpACP:= (MLOwner as TMLComponent).Predecessor.Operator as TOpFactPCA;
end;

destructor TOpFactRotation.Destroy;
begin
 if assigned(FNewLoadings) then DelMatrix(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 if assigned(FNewEgVec) then DelMatrix(FNewEgVec,FOpACP.DimAnalysis,FNbFactors);
 inherited destroy();
end;

function TOpFactRotation.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpFactRotationParameter;
end;

function TOpFactRotation.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>Rotated Factor Loadings</H3>';
 s:= s + FOpACP.getHTMLLoadings(FNewLoadings,FNbFactors);
 s:= s + '<H4>vs. Unrotated Factor Loadings</H4>';
 s:= s + FOpACP.getHTMLLoadings(FOpACP.Loadings,FNbFactors);
 s:= s + '<H3>Eigen vectors -- Factor Scores</H3>';
 s:= s + FOpACP.getHTMLRegCoef(FNewEgVec,FNbFactors);
 //
 result:= s;
end;

function TOpFactRotation.setProjections(): boolean;
begin
 try
 FOpACP.SetProjections(GenAtts,FNewEgVec,FNbFactors,'FR',(MLOwner as TMLComponent).Number);
 result:= true;
 except
 result:= false;
 end;
end;

{ TOpFactRotationParameter }

function TOpFactRotationParameter.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmFactRotateFactors.CreateFromOpPrm(self);
end;

function TOpFactRotationParameter.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Factors rotation</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD>Method</TD>'+ format('<TD>%s</TD>',[STR_FACTOR_ROTATION[FRotation]]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD># factors</TD>'+ format('<TD>%d</TD>',[FNbFactors]) + '</TR>';
 s:= s + '</table>';
 result:= s;
end;

procedure TOpFactRotationParameter.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbFactors:= prmINI.ReadInteger(prmSection,'nb_factors',FNbFactors);
 FRotation:= TEnumRotateFactors(prmINI.ReadInteger(prmSection,'method',ord(FRotation)));
end;

procedure TOpFactRotationParameter.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNbFactors,sizeof(FNbFactors));
 prmStream.ReadBuffer(FRotation,sizeof(FRotation));
end;

procedure TOpFactRotationParameter.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_factors',FNbFactors);
 prmINI.WriteInteger(prmSection,'method',ord(FRotation));
end;

procedure TOpFactRotationParameter.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNbFactors,sizeof(FNbFactors));
 prmStream.WriteBuffer(FRotation,sizeof(FRotation));
end;

procedure TOpFactRotationParameter.SetDefaultParameters;
begin
 FNbFactors:= 2;
 FRotation:= rfVarimax;
end;

initialization
 RegisterClass(TMLGenCompFactRotation);
end.
