(********************************************************************)
(* UCompFactAnalysisAFC.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Analyse Factorielles des Correspondances)
@author(Ricco)
@created(09/02/2007)

Calculs Lebart et al., pp. 67  107
}
unit UCompFactAnalysisAFC;

interface

USES
  Forms,Classes,Contnrs,
  IniFiles,
  Matrices,
  UCalcStatDes,
  UCompDefinition,
  UCompFADefinition,
  UOperatorDefinition,
  UDatasetDefinition;

TYPE
  //gnrateur
  TMLGenCompFactAFC = class(TMLGenComp)
  protected
  procedure   GenCompInitializations(); override;
  public
  function    GetClassMLComponent: TClassMLComponent; override;
  end;

  //composant
  TMLCompFactAFC = class(TMLCompFactAnalysis)
  protected
  function    getClassOperator: TClassOperator; override;
  end;

  //oprateur
  TOpFactAFC = class(TOpFactAnalysis)
  private
  //nb de valeurs propres  analyser
  FEgSize: integer;
  //matrices des donnes, des frquences relatives
  MatD, MatF: PMatrix;
  //la matrice  diaginaliser
  MatA: PMatrix;
  //vecteur des valeurs propres
  FEgVal: PVector;
  //matrice des vecteurs propres
  FEgVec: PMatrix;
  //La trace -- inertie totale et le CHI-2
  FTrace, FChi2: double;
  //matrice des coordonnes factorielles, CTR et cos2 pour les lignes
  FLigCtr, FLigCos2: PMatrix;
  //matrice des coordonnes factorielles, CTR et cos2 pour les colonnes
  FColCtr, FColCos2: PMatrix;
  //liste des contributions
  FLstContrib: TObjectList;
  //initialiser les matrices
  procedure   initMatrix();
  //diagonalisation
  procedure   diagonalize();
  //calculer les projections lignes
  procedure   computeLigProj();
  //calculer les projetcions colonnes,  partir des projections lignes
  procedure   computeColProj();
  //calculer les contributions
  procedure   computeContributions();
  //le test d'independance
  function    getHTMLChi2(): string;
  //les valeurs propres en HTML
  function    getHTMLEgValues(): string;
  //les projections des lignes
  function    getHTMLLignes(): string;
  //les projections colonnes
  function    getHTMLColonnes(): string;
  //les contributions
  function    getHTMLContributions(): string;
  protected
  procedure   DelPrivateMatrix(); override;
  function    getClassParameter: TClassOperatorParameter; override;
  function    CheckAttributes(): boolean; override;
  function    CoreExecute(): boolean; override;
  function    getNewForm(): TForm; override;
  public
  //nombre d'axes rellement calcul
  FNbAxis: integer;
  //nombre de lignes et de colonnes
  FNbLig, FNbCol: integer;
  //coordonnes lignes et colonnes
  FLigCoord, FColCoord: PMatrix;
  function    getHTMLResultsSummary(): string; override;
  end;

  //paramtrage
  TOpPrmFactAFC = class(TOpPrmFactAnalysis)
  protected
  function    CreateDlgParameters(): TForm; override;
  procedure   SetDefaultParameters(); override;
  public
  //nombre d'axes -- df. 5
  AskNbAxis: integer;
  //calculer et afficher les contributions au CHI-2 -- df. false
  IsComputeContrib: boolean;
  function    getHTMLParameters(): string; override;
  procedure   LoadFromStream(prmStream: TStream); override;
  procedure   SaveToStream(prmStream: TStream); override;
  procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
  procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
  end;

implementation

USES
    UDatasetImplementation,
    UCalcMatrixAdditionalFunctions, UStringAddBuffered, UConstConfiguration,
    Math, FMath, Sysutils, UFrmOpViewFactAnalysisAFC,
  UDlgOpPrmFactAnalysisAFC;

TYPE
  //structure pour calcul des contributions
  TContribCHI2 = class
  //identifiant de ligne et de colonne
  sLig, sCol: string;
  //valeur et valeur espre pour la case
  value,expected: double;
  //contribution
  contrib: double;
  //en pourcentage
  percent: double;
  end;


{ TMLGenCompFactAFC }

procedure TMLGenCompFactAFC.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactAFC.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactAFC; 
end;

{ TMLCompFactAFC }

function TMLCompFactAFC.getClassOperator: TClassOperator;
begin
 result:= TOpFactAFC;
end;

{ TOpFactAFC }

function TOpFactAFC.CheckAttributes: boolean;
var ok: boolean;
begin
 //un target discret pour indicateur de lignes du tableau de contingence
 ok:= (self.WorkData.LstAtts[asTarget].Count = 1) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete) = true);
 //puis plusieurs input, tous continus
 ok:= ok and ((self.WorkData.LstAtts[asInput].Count > 1) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue) = true));
 //and then...
 result:= ok;
end;

procedure TOpFactAFC.computeColProj;
var a,j,i: integer;
    sum: double;
    dist: array of single;
begin
 //les coordonnes, Lebart et al., p.85 -- relations de transitions
 dimMatrix(FColCoord,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  begin
   for j:= 1 to FNbCol do
    begin
     sum:= 0.0;
     for i:= 1 to FNbLig do
      sum:= sum + MatF^[i]^[j]/MatF^[0]^[j]*FLigCoord^[i]^[a];
     FColCoord^[j]^[a]:= sum/SQRT(FEgVal^[a+1]);
    end;
  end;
 //contributions
 dimMatrix(FColCtr,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  for j:= 1 to FNbCol do
   FColCtr^[j]^[a]:= 100.0 * MatF^[0]^[j] * SQR(FColCoord^[j]^[a]) / FEgVal^[a+1];
 //distance
 setLength(dist,succ(FNbCol));
 for j:= 1 to FNbCol do
  for i:= 1 to FNbLig do
   dist[j]:= dist[j] + 1.0/MatF^[i]^[0]*SQR(MatF^[i]^[j]/matF^[0]^[j]-MatF^[i]^[0]);
 //cos
 dimMatrix(FColCos2,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  for j:= 1 to FNbCol do
   FColCos2^[j]^[a]:= SQR(FColCoord^[j]^[a])/dist[j];
 //
 finalize(dist);
end;

//critre de tri, descendant
function triContrib(Item1, Item2: Pointer): integer;
var c1, c2: TContribChi2;
begin
 c1:= TContribCHI2(Item1);
 c2:= TContribCHI2(Item2);
 if (c1.contrib < c2.contrib)
  then result:= +1
  else if (c1.contrib > c2.contrib)
    then result:= -1
    else result:= 0;
end;

procedure TOpFactAFC.computeContributions;
var i,j: integer;
    sI,sJ: string;
    c: TContribCHI2;
begin
 FLstContrib:= TObjectList.Create(TRUE);//propritaire
 //pour chaque case du tableau de contingence
 for i:= 1 to FNbLig do
  begin
   sI:= self.WorkData.LstAtts[asTarget].Attribute[0].sValue[self.WorkData.Examples.Number[i]];
   for j:= 1 to FNbCol do
    begin
     sJ:= self.WorkData.LstAtts[asInput].Attribute[pred(j)].Name;
     //
     c:= TContribCHI2.Create();
     c.sLig:= sI;
     c.sCol:= sJ;
     c.value:= MatD^[i]^[j];
     c.expected:= MatD^[i]^[0]*MatD^[0]^[j]/MatD^[0]^[0];
     c.contrib:= SQR(c.value - c.expected)/c.expected;
     c.percent:= 100.0*c.contrib/FChi2;
     //
     FLstContrib.Add(c);
    end;
  end;
 //trier par ordre d'importance dcroissante
 FLstContrib.Sort(triContrib);
end;

procedure TOpFactAFC.computeLigProj;
var i,j,a: integer;
    sum: double;
    dist: array of single;
begin
 //coordonnes
 dimMatrix(FLigCoord,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  begin
   for i:= 1 to FNbLig do
    begin
     sum:= 0.0;
     for j:= 1 to FNbCol do
      sum:= sum + MatF^[i]^[j] / (MatF^[i]^[0] * MatF^[0]^[j]) * FEgVec^[j]^[a+1];
     FLigCoord^[i]^[a]:= sum;
    end;
  end;
 //contributions
 dimMatrix(FLigCtr,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  for i:= 1 to FNbLig do
   FLigCtr^[i]^[a]:= 100.0 * MatF^[i]^[0] * SQR(FLigCoord^[i]^[a]) / FEgVal^[a+1];
 //calculer la distance du point au centre de gravit
 setLength(dist,succ(FNbLig));
 for i:= 1 to FNbLig do
  for j:= 1 to FNbCol do
   dist[i]:= dist[i] + 1.0/MatF^[0]^[j]*SQR(MatF^[i]^[j]/MatF^[i]^[0]-MatF^[0]^[j]);
 //cos
 dimMatrix(FLigCos2,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  for i:= 1 to FNbLig do
   FLigCos2^[i]^[a]:= SQR(FLigCoord^[i]^[a])/dist[i];
 //finalize
 Finalize(dist);
end;

function TOpFactAFC.CoreExecute: boolean;
begin
 result:= true;
 TRY
 //supprimer les matrices prcdentes
 self.DelPrivateMatrix();
 //nouvelle dimension des matrices
 FNbLig:= self.WorkData.Examples.Size;
 FNbCol:= self.WorkData.LstAtts[asInput].Count;
 //calculer le nombre d'axes  traiter
 FEgSize:= MATH.Min(FNbLig,FNbCol);//puisqu'on travaille dans l'espace des colonnes toujours
 FNbAxis:= MATH.Min(MATH.Min(FNbLig,FNbCol)-1,(self.PrmOp as TOpPrmFactAFC).AskNbAxis);
 //copier les matrices de donnes et la matrice  diagonaliser
 self.initMatrix();
 //calcul des valeurs et vecteurs propres
 self.diagonalize();
 //calculer les projections lignes
 self.computeLigProj();
 //et les projections colonnes
 self.computeColProj();
 //ventuellement, calcul des contributions
 if ((self.PrmOp as TOpPrmFactAFC).IsComputeContrib = true)
  then self.computeContributions();
 EXCEPT
 result:= false;
 END;
end;

procedure TOpFactAFC.DelPrivateMatrix;
begin
 if assigned(MatD) then delMatrix(MatD,FNbLig,FNbCol);
 if assigned(MatF) then delMatrix(MatF,FNbLig,FNbCol);
 if assigned(MatA) then delMatrix(MatA,FNbCol,FNbCol);
 if assigned(FEgVec) then delMatrix(FEgVec,FNbCol,FNbCol);
 if assigned(FEgVal) then delVector(FEgVal,FNbCol);
 //pour les lignes
 if assigned(FLigCoord) then delMatrix(FLigCoord,FNbLig,FNbAxis);
 if assigned(FLigCtr) then delMatrix(FLigCtr,FNbLig,FNbAxis);
 if assigned(FLigCos2) then delMatrix(FLigCos2,FNbLig,FNbAxis);
 //pour les colonnes
 if assigned(FColCoord) then delMatrix(FColCoord,FNbCol,FNbAxis);
 if assigned(FColCtr) then delMatrix(FColCtr,FNbCol,FNbAxis);
 if assigned(FColCos2) then delMatrix(FColCos2,FNbCol,FNbAxis);
 //pour la liste des contributions
 if assigned(FLstContrib) then FreeAndNil(FLstContrib);
end;

procedure TOpFactAFC.diagonalize;
var j,i: integer;
begin
 GetEigensFromSymetricMatrix(MatA,FNbCol,FEgVal,FEgVec);
 //calcul de la trace
 FTrace:= 0.0;
 for j:= 2 to FEgSize do FTrace:= FTrace + FEgVal^[j];
 //et du chi-2 : trace * effectif total
 FChi2:= FTrace * MatD^[0]^[0];
 //correction des vecteurs propres. Lebart et al., pp.102
 for j:= 1 to FNbCol do //cf. la formule u = Dp^(-1/2) * w (le signe semble invers ? c'est +1/2 ???) ou c'est moi qui interprte mal ???
   for i:= 1 to FNbCol do    //hum, il semble que ce soit une multiplication par la racine carre de la marge, et non une division ???
    FEgVec^[i]^[j]:= FEgVec^[i]^[j] * SQRT(MatF^[0]^[i]); //en tous cas, avec cette criture du programme, le rsultat est correct, c'est l'essentiel
end;

function TOpFactAFC.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactAFC;
end;

function TOpFactAFC.getHTMLChi2: string;
var p: double;
    ddl: integer;
    s: string;
begin
 ddl:= (FNbLig-1) * (FNbCol-1);
 p:= PKhi2(ddl,FChi2);
 s:= '<P><H3>CHI-SQUARE statistic</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>Trace</TD><TD>%.4f</TD></TR>',[FTrace]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>Chi</TD><TD>%.2f</TD></TR>',[FChi2]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>d.f</TD><TD>%d</TD></TR>',[ddl]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>p-value</TD><TD>%.4f</TD></TR>',[p]);
 s:= s + '</table>';
 result:= s; 
end;

function TOpFactAFC.getHTMLColonnes: string;
var s: string;
    j,a: integer;
begin
 s:= '<H3>Columns analysis</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=2>%s</TH>',['-']) +
         format('<TH colspan=%d>Coord.</TH>',[FNbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[FNbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[FNbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH><TH width="50">Weight</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">coord %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">ctr %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 //valeurs
 for j:= 1 to FNbCol do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD align="left" %s>%s</TD><TD %s align="right">%.2f</TD>',[HTML_BGCOLOR_DATA_BLUE,workdata.lstAtts[asInput].Attribute[pred(j)].Name,HTML_BGCOLOR_HEADER_GRAY,100.0*MatF^[0]^[j]]);
   for a:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FColCoord^[j]^[a]]);
   for a:= 1 to FNbAxis do s:= s + format('<TD align="right">%.1f</TD>',[FColCtr^[j]^[a]]);
   for a:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FColCos2^[j]^[a]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLContributions: string;
var seuil: double;
    i: integer;
    bf: TBufString;
    c: TContribCHI2;
    color: string;
begin
 if assigned(FLstContrib) and (FLstContrib.Count > 0)
  then
   begin
    //le seuil pour ne pas tout afficher
    seuil:= FChi2/(1.0*FNbLig*FNbCol);
    //zoo..
    bf:= TBufString.Create();
    bf.BeginUpdate();
    //
    bf.AddStr('<H3>CHI-2 contributions</H3>');
    bf.AddStr(HTML_HEADER_TABLE_RESULT);
    bf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Row</TH><TH>Column</TH><TH>Value</TH><TH>Expected</TH><TH>Contrib.</TH><TH>%</TH></TR>');
    i:= 0;
    c:= FLstContrib.Items[i] as TContribCHI2;
    while (c<>nil) and (c.contrib > seuil) do
     begin
      //affichage
      bf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%s</TD><TD>%s</TD><TD>%.0f</TD><TD>%.1f</TD>',[c.sLig,c.sCol,c.value,c.expected]));
      if (c.value < c.expected)
       then color:= HTML_BGCOLOR_DATA_RED
       else color:= HTML_BGCOLOR_DATA_GREEN;
      bf.AddStr(format('<TD %s>%.2f</TD><TD>%.2f</TD>',[color,c.contrib,c.percent]));
      bf.AddStr('</TR>');
      //next...
      inc(i);
      if (i < FLstContrib.Count)
       then c:= FLstContrib.Items[i] as TContribCHI2
       else c:= nil;
     end;
    bf.AddStr('</table>');
    //
    bf.EndUpdate();
    result:= bf.BufS;
    bf.Free();
   end
  else result:= '';
end;

function TOpFactAFC.getHTMLEgValues: string;
var tot, sum: double;
    s: string;
    i: integer;
begin
 s:= '<P><H3>Eigen values</H3>';
 s:= s + format('<H5>Matrix trace = %.4f</H5>',[FTrace]);
 s:= s+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH>Eigen value</TH><TH>% explained</TH><TH width="200">Histogram</TH><TH>% cumulated</TH></TR>';
 tot:= FTrace;
 sum:= 0.0;
 for i:= 2 to FEgSize do
  begin
   sum:= sum + FEgVal^[i];
   if (tot>0)
    then
     s:= s + HTML_TABLE_COLOR_DATA_GRAY+
         format('<TH>%d</TH><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD><TD align=right>%.2f%s</TD><TD>%s</TD><TD align=right>%.2f%s</TD></TR>',
                [i-1,FEgVal^[i],100.0*FEgVal^[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*FEgVal^[i]/tot)),100.0*sum/tot,'%'])+'</TR>';
  end;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>Tot.</TH><TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>',[sum]);
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLLignes: string;
var s: string;
    i,j: integer;
begin
 s:= '<H3>Rows analysis</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=2>%s</TH>',['-']) +
         format('<TH colspan=%d>Coord.</TH>',[FNbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[FNbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[FNbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH><TH width="50">Weight</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">coord %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">ctr %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 //valeurs
 for i:= 1 to FNbLig do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD align="left" %s>%s</TD><TD %s align="right">%.2f</TD>',[HTML_BGCOLOR_DATA_GREEN,workdata.lstAtts[asTarget].Attribute[0].sValue[workdata.examples.Number[i]],HTML_BGCOLOR_HEADER_GRAY,100.0*MatF^[i]^[0]]);
   for j:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FLigCoord^[i]^[j]]);
   for j:= 1 to FNbAxis do s:= s + format('<TD align="right">%.1f</TD>',[FLigCtr^[i]^[j]]);
   for j:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FLigCos2^[i]^[j]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLResultsSummary: string;
var buf: TBufString;
begin
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //test d'indpendance
 buf.AddStr(self.getHTMLChi2());
 //afficher les valeurs propres et la trace de la matrice
 buf.AddStr(self.getHTMLEgValues());
 //les lignes
 buf.AddStr(self.getHTMLLignes());
 //les colonnes
 buf.AddStr(self.getHTMLColonnes());
 //les contributions ?
 if ((self.PrmOp as TOpPrmFactAFC).IsComputeContrib = true)
  then buf.AddStr(self.getHTMLContributions());
 //and then...
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFC.getNewForm: TForm;
begin
 result:= TFrmOpViewFactAnalysisAFC.CreateFromOperator(self);
end;

procedure TOpFactAFC.initMatrix;
var sum: single;
    i,j,jprim,example: integer;
begin
 //matrice des donnes -- les marges sont mises  zro automatiquement, ouf...
 dimMatrix(MatD,FNbLig,FNbCol);
 for i:= 1 to FNbLig do
  begin
   example:= self.WorkData.Examples.Number[i];
   for j:= 1 to FNbCol do
    begin
     MatD^[i]^[j]:= self.WorkData.LstAtts[asInput].Attribute[pred(j)].cValue[example];
     //les marges
     matD^[i]^[0]:= matD^[i]^[0] + MatD^[i]^[j];
     matD^[0]^[j]:= matD^[0]^[j] + MatD^[i]^[j];
     matD^[0]^[0]:= matD^[0]^[0] + MatD^[i]^[j];
    end;
  end;
 //matrice des frquences, y compris les frquences marginales
 dimMatrix(MatF,FNbLig,FNbCol);
 for i:= 0 to FNbLig do
  for j:= 0 to FNbCol do
   MatF^[i]^[j]:= MatD^[i]^[j] / matD^[0]^[0];
 //matrice de calcul -- de taille col x col -  diagonaliser
 dimMatrix(matA,FNbCol,FNbCol);
 for j:= 1 to FNbCol do
  for jprim:= 1 to FNbCol do
   begin
    sum:= 0.0;
    for i:= 1 to FNbLig do
     sum:= sum + (MatF^[i]^[j] * MatF^[i]^[jprim]) / (MatF^[i]^[0] * SQRT(MatF^[0]^[j] * MatF^[0]^[jprim]));
    MatA^[j]^[jprim]:= sum;
   end;
end;

{ TOpPrmFactAFC }

function TOpPrmFactAFC.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFactAnalysisAFC.CreateFromOpPrm(self);
end;

function TOpPrmFactAFC.getHTMLParameters: string;
var s: string;
begin
 s:= format('<b>Nb axis : %d</b><br>',[AskNbAxis]);
 s:= s + format('<b>Contrib. : %d</b><br>',[ord(IsComputeContrib)]);
 result:= s;
end;

procedure TOpPrmFactAFC.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 AskNbAxis:= prmINI.ReadInteger(prmSection,'nb_axis',AskNbAxis);
 IsComputeContrib:= prmINI.ReadBool(prmSection,'contrib',IsComputeContrib);
end;

procedure TOpPrmFactAFC.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(AskNbAxis,sizeof(AskNbAxis));
 prmStream.ReadBuffer(IsComputeContrib,sizeof(IsComputeContrib));
end;

procedure TOpPrmFactAFC.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',AskNbAxis);
 prmINI.WriteBool(prmSection,'contrib',IsComputeContrib);
end;

procedure TOpPrmFactAFC.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(AskNbAxis,sizeof(AskNbAxis));
 prmStream.WriteBuffer(IsComputeContrib,sizeof(IsComputeContrib));
end;

procedure TOpPrmFactAFC.SetDefaultParameters;
begin
 //nombre d'axes -- df. 5
 AskNbAxis:= 5;
 //calculer et afficher les contributions au CHI-2 -- df. false
 IsComputeContrib:= false;
end;

initialization
 RegisterClass(TMLGenCompFactAFC);
end.
