(*****************************************************************)
(* UCompFCDiscValues.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(Continuous to discrete)
@author(Ricco)
@created(01/10/2007)

Dnombre les valeurs d'une variable continue, cre une variable discrte
correspondante en affectant une modalit  chaque valeur.

Cela peut tre intressant si la variable originelle est ordinale, code 1, 2, 3...

}

unit UCompFCDiscValues;

interface

USES
        Forms, Classes, IniFiles, Contnrs,
        UCompDefinition,
        UCompFCDefinition,
        UOperatorDefinition,
        UCalcStatDes, UDatasetDefinition, UDatasetImplementation,
        UCalcStructOrdinalAssociation;

TYPE
        {gnrateur de standardisation}
        TGenFCC2D = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant standardisation}
        TMLCompFC2D = class(TMLCompFC)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {oprateur de standaridsation}
        TOpFCC2D = class(TOperatorFC)
        private
        //liste des vecteurs de valeurs des variables continues
        FLstVec: TObjectList;
        //tester si la liste actuelle est valable, dans ce cas pas de calcul
        function    isNeedNewComputation(): boolean;
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        {au dpart, il n'y a que des inputs, et ils sont tous continus}
        function    CheckAttributes(): boolean; override;
        {calculer les stats sur les bons individus et variables, puis normaliser}
        function    CoreExecute(): boolean; override;
        public
        {prparer la liste des stat-des pour la normalisation}
        constructor Create(AOwner: TObject); override;
        {dtruire la liste des stat-des}
        destructor  Destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtre de l'oprateur}
        TPrmOpFC2D = class(TOperatorPrmFC)
        protected
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        end;

implementation

USES
    Sysutils,
    UConstConfiguration, ULogFile;


{ TGenFCC2D }

procedure TGenFCC2D.GenCompInitializations;
begin
 FMLComp:= mlcFeatureConstruction;
end;

function TGenFCC2D.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFC2D; 
end;

{ TMLCompFC2D }

function TMLCompFC2D.getClassOperator: TClassOperator;
begin
 result:= TOpFCC2D; 
end;

function TMLCompFC2D.GetLogResultDescription: string;
begin
 result:= format('%d discrete attributes generated',[LocalDataset.Count]);
end;

{ TOpFCC2D }

function TOpFCC2D.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue) = true);
end;

function TOpFCC2D.CoreExecute: boolean;
var j,i,k: integer;
    att,newAtt: TAttribute;
    vec: TVecDistinctValues;
    value: TTypeDiscrete;
begin
 result:= true;
 //a-t-on besoin de recalculer ?
 if (self.isNeedNewComputation() = true)
  then
   begin
    //vider les listes intermdiaires
    FLstVec.Clear;
    GenAtts.Clear();
    //gnrer la nouvelle liste
    TRY
    for j:= 0 to pred(self.WorkData.LstAtts[asInput].Count) do
     begin
      att:= self.WorkData.LstAtts[asInput].Attribute[j];
      vec:= TVecDistinctValues.create(att);
      newAtt:= TAttDiscrete.Create('c2d_'+att.Name+'_'+IntToStr((MLOwner as TMLComponent).Number),att.Size);
      TraceLog.WriteToLogFile(format('[CONT_2_DISC] WARNING !!! exceeding number of values for %s',[att.Name]));
      //si trop large, on cre une constante...
      if (vec.nbValues > MAX_NB_VALUES_CATEGORICAL)
       then
        begin
         value:= newAtt.LstValues.getValue('_NA_');
         for i:= 1 to newAtt.Size do newAtt.dValue[i]:= value;
        end
       else
        begin
         //crer la liste des modalits
         for k:= 1 to vec.nbValues do
          newAtt.LstValues.getValue('_'+inttostr(k)+'_'+FormatFloat('0.00',vec.getValue(k)));
         //il faut coder pour chaque valeur
         for i:= 1 to newAtt.Size do
          newAtt.dValue[i]:= vec.getIndex(att.cValue[i]);
        end;
      //rajouter
      GenAtts.Add(newAtt);
      //dtruire le vecteur des valeurs
      if assigned(vec) then vec.Free();
     end;
    EXCEPT
    result:= false;
    END;
   end;
end;

constructor TOpFCC2D.Create(AOwner: TObject);
begin
 inherited;
 FLstvec:= TObjectList.Create(TRUE);
end;

destructor TOpFCC2D.Destroy;
begin
 FLstVec.Free();
 inherited;
end;

function TOpFCC2D.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpFC2D; 
end;

function TOpFCC2D.getHTMLResultsSummary: string;
var j: integer;
    s: string;
begin
 //lister les variables gnres et le nombre de modalits associes
 s:= '<H3>Continuous to discrete transformation</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>New attribute name</TH><TH>Values</TH></TR>';
 for j:= 0 to pred(GenAtts.Count) do
  s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="left">%s</TD><TD align="center">%d</TD></TR>',[GenAtts.Attribute[j].Name,GenAtts.Attribute[j].NbValues]);
 s:= s + '</table>';
 result:= s;
end;

function TOpFCC2D.isNeedNewComputation: boolean;
var ok: boolean;
    j: integer;
begin
 //vrifier si la liste des stats et la liste des variables demande est la mme
 if (FLstVec.Count = self.WorkData.LstAtts[asInput].Count) and (FLstVec.Count = self.GenAtts.Count)
  then
   begin
    //passer en revue
    ok:= true;
    for j:= 0 to pred(FLstVec.Count) do
     ok:= ok and ((FLstVec[j] as TVecDistinctValues).Attribute.Name = self.WorkData.LstAtts[asInput].Attribute[j].Name);
    //and then
    result:= not(ok);
   end
  else result:= true;
end;

{ TPrmOpFC2D }

function TPrmOpFC2D.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

procedure TPrmOpFC2D.SetDefaultParameters;
begin
 //none
end;

initialization
 RegisterClass(TGenFCC2D);
end.
