(***************************************************************)
(* UCalcFSStepDisc.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(***************************************************************)

{

@abstract(Structures de calcul pour la slection de variables base sur la rgression logistique)
@author(Ricco)
@created(05/12/2007)

Structures utilises pour la slection de variables base sur la rgression logistique.
Il s'agit d'un vrai plagiat de UCALCFSSTEPDISC mais a vite de modifier du code dj valid
pour laborer une systme d'hritage compliqu.

}
unit UCalcFSStepLogisticReg;

interface

USES
    Contnrs,
    UDatasetDefinition;

TYPE
    ///un candidat pour la recherche
    TCandidat = class
    ///numro de variable chez les input
    Numero: integer;
    ///variable
    Variable: TAttribute;
    ///constructeur
    constructor create(num: integer; prmVar : TAttribute);
    ///destructeur
    destructor  destroy(); override;
    end;

    ///liste de candidats -- elle peut tre propritaire ou pas
    TLstCandidat = class
    ///liste interne
    Lst : TObjectList;
    ///constructeur
    constructor create(owner: boolean);
    ///destructeur
    destructor  destroy(); override;
    ///assigner
    procedure   assign(source: TLstCandidat);
    ///rcuprer
    function    getCandidat(index: integer): TCandidat;
    ///rcuprer l'index d'un candidat
    function    indexOf(candidat: TCandidat): integer;
    ///retirer un candidat
    procedure   remove(index: integer);
    ///ajouter un candidat
    procedure   add(candidat: TCandidat);
    ///nombre
    function    getCount(): integer;
    ///vider
    procedure   clear();
    end;

    ///un rsultat de calcul
    TResultat = class
    ///candidat
    Candidat : TCandidat;
    ///CHI-2 du SCORE (Forward) ou de WALD (Backward) selon le cas
    Chi2: double;
    ///p-value
    PValue: double;
    ///constructeur
    constructor create(prmCandidat: TCandidat);
    ///destructeur
    destructor  destroy(); override;
    ///description HTML avec le code couleur
    function    getHTML(colorCode: string): string;
    end;

    ///liste de rsultats pour un niveau
    TLevelResultat = class
    ///Best Resultat -- la premire dans la liste
    Best: TResultat;
    ///AIC du modle trait
    AIC: double;
    ///Chi2 du modle trait
    CHi2: double;
    ///ddl du niveau de recherche
    ddl: integer;
    ///p-value du modle trait
    pValue: double;
    ///liste des rsultats associs  ce niveau de recherche -- liste propritaire
    LstResultat: TObjectList;
    ///constructeur
    constructor create();
    ///destructeur
    destructor  destroy(); override;
    ///trier sur le CHI2 (les ddl sont les mme pour un niveau), de manire croissante (backward)
    procedure   sortAscendant();
    ///trier sur le lambda -- pour forward
    procedure   sortDescendant();
    ///nombre de rsultats dispo
    function    getCountResults(): integer;
    ///ajouter un rsultat
    procedure   addresult(resultat: TResultat);
    ///rcuprer un rsultat
    function    getResult(index: integer): TResultat;
    ///description HTML, en paramtre le nombre de rsultats  afficher
    function    getHTML(number: integer): string;
    end;

    ///liste globale des rsultats -- pour chaque niveau
    TGlobalResultat = class
    ///liste des niveaux -- liste propritaire
    LstLevel: TObjectList;
    ///constructeur
    constructor create();
    ///destructeur
    destructor  destroy(); override;
    ///liste des niveaux dispo
    function    getCountLevel(): integer;
    ///ajouter un niveau de rsultat
    procedure   addLevel(level: TLevelResultat);
    ///description HTML, en paramtre le nombre de colonnes  afficher
    function    getHTML(numCol: integer): string;
    ///rcupration d'un niveau
    function    getLevel(index: integer): TLevelResultat;
    end;

implementation

USES
    Sysutils, UConstConfiguration;

{ TStepCandidat }

constructor TCandidat.create(num: integer; prmVar: TAttribute);
begin
 inherited create();
 Numero:= num;
 Variable:= prmVar;
end;

{ TLstCandidat }

procedure TLstCandidat.add(candidat: TCandidat);
begin
 Lst.Add(candidat);
end;

procedure TLstCandidat.assign(source: TLstCandidat);
begin
 Lst.Assign(source.Lst);
end;

procedure TLstCandidat.clear;
begin
 Lst.Clear();
end;

constructor TLstCandidat.create(owner: boolean);
begin
 inherited create();
 Lst:= TObjectList.Create(owner);
end;

destructor TLstCandidat.destroy;
begin
 Lst.Free();//selon qu'il est owner ou pas, les objets seront dtruits...
 inherited;
end;

function TLstCandidat.getCandidat(index: integer): TCandidat;
begin
 result:= Lst[index] as TCandidat;
end;

function TLstCandidat.getCount: integer;
begin
 result:= Lst.Count;
end;

function TLstCandidat.indexOf(candidat: TCandidat): integer;
begin
 result:= Lst.IndexOf(candidat);
end;

procedure TLstCandidat.remove(index: integer);
begin
 //attention, si la liste est propritaire, l'objet est dtruit
 Lst.Delete(index);
end;

{ TLevelResultat }

procedure TLevelResultat.addresult(resultat: TResultat);
begin
 LstResultat.Add(resultat);
end;

constructor TLevelResultat.create;
begin
 inherited create();
 LstResultat:= TObjectList.Create(true);
end;

destructor TLevelResultat.destroy;
begin
 LstResultat.Free();
 inherited;
end;

function TLevelResultat.getCountResults: integer;
begin
 result:= LstResultat.Count;
end;

function TLevelResultat.getHTML(number: integer): string;
var s: string;
    i: integer;
    res: TResultat;
begin
 //la meilleure solution
 if (self.Best = nil)
  then s:= format('<TD %s>-</TD>',[HTML_BGCOLOR_DATA_RED])
  else s:= best.getHTML(HTML_BGCOLOR_DATA_RED);
 //les autres solutions
 for i:= 1 to number do
  begin
   if (i > self.getCountResults())
    then s:= s + format('<TD %s>-</TD>',[HTML_BGCOLOR_DATA_GRAY])
    else
     begin
      res:= self.getResult(i-1);
      s:= s + res.getHTML(HTML_BGCOLOR_DATA_GRAY);
     end;
  end;
 //and then...
 result:= s;
end;

function TLevelResultat.getResult(index: integer): TResultat;
begin
 result:= LstResultat[index] as TResultat;
end;

function compareAscendant(Item1, Item2: Pointer): integer;
var res1, res2: TResultat;
begin
 res1:= TResultat(Item1);
 res2:= TResultat(Item2);
 if (res1.CHI2 < res2.CHI2)
  then result:= -1
  else if (res1.CHI2 > res2.CHI2)
        then result:= +1
        else result:= 0;
end;

procedure TLevelResultat.sortAscendant;
begin
 self.LstResultat.Sort(compareAscendant);
end;

function compareDescendant(Item1, Item2: Pointer): integer;
begin
 result:= -1 * compareAscendant(Item1,Item2);
end;

procedure TLevelResultat.sortDescendant;
begin
 self.LstResultat.Sort(compareDescendant);
end;

{ TGlobalResultat }

procedure TGlobalResultat.addLevel(level: TLevelResultat);
begin
 LstLevel.Add(level);
end;

constructor TGlobalResultat.create;
begin
 inherited create();
 LstLevel:= TObjectList.Create(true);
end;

destructor TGlobalResultat.destroy;
begin
 LstLevel.Free();
 inherited;
end;

function TGlobalResultat.getCountLevel: integer;
begin
 result:= LstLevel.Count;
end;

function TGlobalResultat.getHTML(numCol: integer): string;
var i: integer;
    s: string;
    level: TLevelResultat;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 //titre
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>N</TH><TH>Current Reg.</TH><TH>Moved</TH>';
 for i:= 1 to numCol do
  s:= s + format('<TH>Sol.%d</TH>',[i]);
 s:= s + '</TR>';
 //pour chaque resultat
 for i:= 0 to pred(self.getCountLevel()) do
  begin
   level:= self.getLevel(i);
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%d</TD>',[succ(i)]);
   //mettre les infos de la rgression courante
   s:= s + format('<TD>AIC : %.2f<br>CHI-2 : %.2f<br>d.f. : %d <br>p-value : %.4f</TD>',[level.AIC,level.CHI2,level.ddl,level.pValue]);
   //dtail de la recherche
   s:= s + level.getHTML(numCol);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 //
 result:= s;
end;

function TGlobalResultat.getLevel(index: integer): TLevelResultat;
begin
 result:= LstLevel[index] as TLevelResultat;
end;

{ TResultat }

constructor TResultat.create(prmCandidat: TCandidat);
begin
 inherited create();
 Candidat:= prmCandidat;
end;

destructor TResultat.destroy;
begin
 Candidat:= nil;//surtout ne pas librer l'objet, faire rfrencer  NIL simplement, c'est  LstCandidat(Owner) de le faire
 inherited;
end;

function TResultat.getHTML(colorCode: string): string;
begin
 result:= format('<TD ALIGN=LEFT %s><b>%s</b><br>Chi-2 : %.3f<br>p : %.4f</TD>',
     [colorCode,Candidat.Variable.Name,CHI2,PValue]);
end;

destructor TCandidat.destroy;
begin
 Variable:= NIL;
 inherited;
end;

end.
