(**************************************************************************)
(* UCompFSLogisticRegBackward.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(**************************************************************************)

{

@abstract(Slection de variables pour la rgression logistique binaire - Mthode Backward base sur la statistique de Wald)
@author(Ricco)
@created(05/12/2007)

Trs simple, on effectue la rgression, on regarde la variable qui a le CHI-2 le plus bas. Si non significatif,
on la supprime, et on r-effectue la rgression, etc....

Paramtre de calcul : la p-value seuil...
Autres paramtres pour le rapport : affichage dtaill + nombre de colonnes pour l'affichage

}

unit UCompFSLogisticRegBackward;

interface

USES
    Classes, IniFiles, Forms,
    UCompDefinition,
    UCompFSDefinition,
    UCompFSInputSelection,
    UOperatorDefinition,
    UCalcFSStepLogisticReg,
    UDatasetDefinition;

TYPE
    ///classe de base de paramtrage pour le rapport et le seuil
    TOpPrmFSRegLogBase = class(TOpPrmFSInputSelection)
    protected
    procedure   SetDefaultParameters(); override;
    public
    PValue: double;//p-value seuil, df. = 0.01
    IsShowDetails: boolean;///montrer le dtail des calculs dans un tableau
    NbSoluceShowed: integer;//nombre de colonnes dans le tableau de dtail - df. = 5, si = 0 cd. on montrer toutes les colonnes
    function    getHTMLParameters(): string; override;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

    ///gnrateur de backawrd selection pour rgression logistique
    TMLGenFSBackwardLogReg = class(TMLGenFS)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    ///composant backward selection pour rgression logistique
    TMLCompFSBackwardLogReg = class(TMLCompFSInputSelection)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    ///oprateur de calcul
    TOpFSBackwardLogReg = class(TOpFSInputSelSpvInputContinuous)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    ///lancer les calculs --  surcharger plus tard pour la forward
    procedure   RunSearch(); virtual;
    ///lister les variables slectionnes
    function    getHTMLSelectedAtts(): string;
    ///attribut en entre doit tre binaire
    function    CheckAttributes(): boolean; override;
    public
    ///la solution globale
    Soluce: TGlobalResultat;
    ///Candidats  l'tape t
    Candidats: TLstCandidat;
    ///attribut target
    Target: TAttribute;
    ///input
    Inputs: TLstAttributes;
    ///prparer les objets pour les calculs
    procedure   prepareObjects();
    ///lancer les calculs
    function    CoreExecute(): boolean; override;
    ///dtruire les objets restants
    destructor  Destroy(); override;
    ///rapport
    function    getHTMLResultsSummary(): string; override;
    end;

    ///paramtrage
    TOpPrmFSRegLogBackward = class(TOpPrmFSRegLogBase)
    protected
    function    CreateDlgParameters(): TForm; override;
    end;
    

implementation

uses
    Sysutils,
    UDatasetImplementation, FMath, ULogFile, UConstConfiguration,
    UCompSpvLogisticRegression, UCompManageDataset, UDlgOpPrmFSLogRegBase;

{ TOpPrmFSRegLogBase }

function TOpPrmFSRegLogBase.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan="2">Parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Probability cut</TD><TD align="right">%.4f</TD></TR>',[PValue]);
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TD colspan=2>Report</TD>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Show details</TD><TD align="right">%d</TD></TR>',[ord(IsShowDetails)]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD># columns</TD><TD align="right">%d</TD></TR>',[NbSoluceShowed]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmFSRegLogBase.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 PValue:= prmINI.ReadFloat(prmSection,'p_value',PValue);
 IsShowDetails:= prmINI.ReadBool(prmSection,'details',IsShowDetails);
 NbSoluceShowed:= prmINI.ReadInteger(prmSection,'nb_soluce',NbSoluceShowed);
end;

procedure TOpPrmFSRegLogBase.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(PValue,sizeof(PValue));
 prmStream.ReadBuffer(IsShowDetails,sizeof(IsShowDetails));
 prmStream.ReadBuffer(NbSoluceShowed,sizeof(NbSoluceShowed));
end;

procedure TOpPrmFSRegLogBase.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'p_value',PValue);
 prmINI.WriteBool(prmSection,'details',IsShowDetails);
 prmINI.WriteInteger(prmSection,'nb_soluce',NbSoluceShowed);
end;

procedure TOpPrmFSRegLogBase.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(PValue,sizeof(PValue));
 prmStream.WriteBuffer(IsShowDetails,sizeof(IsShowDetails));
 prmStream.WriteBuffer(NbSoluceShowed,sizeof(NbSoluceShowed));
end;

procedure TOpPrmFSRegLogBase.SetDefaultParameters;
begin
 PValue:= 0.01;
 IsShowDetails:= true;
 NbSoluceShowed:= 5;
end;

{ TMLGenFSBackwardLogReg }

function TMLGenFSBackwardLogReg.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFSBackwardLogReg; 
end;

{ TMLCompFSBackwardLogReg }

function TMLCompFSBackwardLogReg.getClassOperator: TClassOperator;
begin
 result:= TOpFSBackwardLogReg; 
end;

{ TOpFSBackwardLogReg }

function TOpFSBackwardLogReg.CheckAttributes: boolean;
begin
 result:= inherited CheckAttributes() AND (self.WorkData.LstAtts[asTarget].Attribute[0].nbValues = 2);
end;

function TOpFSBackwardLogReg.CoreExecute: boolean;
begin
 result:= true;
 try
 //target and inputs
 Target:= self.WorkData.LstAtts[asTarget].Attribute[0];
 Inputs:= self.WorkData.LstAtts[asInput];
 //prparer les objets
 self.prepareObjects();
 //lancer la recherche
 self.RunSearch();
 except
 result:= false;
 end;
end;

destructor TOpFSBackwardLogReg.Destroy;
begin
 if assigned(Soluce) then Soluce.Free();
 if assigned(Candidats) then Candidats.Free();//aprs SOLUCE !!! car c'est le proprio des objets candidats !!!
 inherited;
end;

function TOpFSBackwardLogReg.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFSRegLogBackward;
end;

function TOpFSBackwardLogReg.getHTMLResultsSummary: string;
var s: string;
    prm: TOpPrmFSRegLogBackward;
begin
 s:= '<H3>Selection results</H3>';
 s:= s + Format('<b>[%d]</b> selected attributes on <b>[%d]</b>',[self.OutputData.LstAtts[asInput].Count,Inputs.Count]);
 s:= s + self.getHTMLSelectedAtts();
 prm:= self.PrmOp as TOpPrmFSRegLogBackward;
 if (prm.IsShowDetails = true)
  then
   begin
    s:= s + '<H3>Detailed results</H3>';
    if (prm.NbSoluceShowed < 1)
     then s:= s + Soluce.getHTML(Inputs.Count)
     else s:= s + Soluce.getHTML(prm.NbSoluceShowed);
   end;
 result:= s;
end;

function TOpFSBackwardLogReg.getHTMLSelectedAtts: string;
var s: string;
    j: integer;
begin
 s:= '<H3>Selected attributes'' subset</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>N</TH><TH>Selected atts</TH></TR>';
 for j:= 0 to pred(self.OutputData.LstAtts[asInput].Count) do
  s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>%d</TD><TD>%s</TD></TR>',[succ(j),self.OutputData.LstAtts[asInput].Attribute[j].Name]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpFSBackwardLogReg.prepareObjects;
var j: integer;
begin
 if assigned(Soluce) then Soluce.Free();
 Soluce:= TGlobalResultat.create();
 //les candidats -- trait aprs soluce...
 if assigned(Candidats) then Candidats.Free();
 Candidats:= TLstCandidat.create(true);//propritaire de la liste
 for j:= 0 to pred(Inputs.Count) do
  Candidats.add(TCandidat.create(succ(j),Inputs.Attribute[j])); 
end;

procedure TOpFSBackwardLogReg.RunSearch;
var nbInputs: integer;
    inputList: TLstCandidat;
    model: TCalcLogisticReg;
    curInputs: TLstAttributes;
    levelRes: TLevelResultat;//rsultat pour un niveau
    res: TResultat;//un rsultat
    ok: boolean;
    j: integer;
    pSeuil: double;
begin
 pSeuil:= (self.PrmOp as TOpPrmFSRegLogBackward).PValue;
 inputList:= TLstCandidat.create(false); inputList.assign(Candidats); //liste locale des candidats
 curInputs:= TLstAttributes.Create(false,self.WorkData.Examples.Size);
 REPEAT
  nbInputs:= inputList.getCount();
  //rcuprer les variables  tester
  curInputs.Clear(); for j:= 0 to pred(inputList.getCount()) do curInputs.Add(inputList.getCandidat(j).Variable);
  //crer le modle avec les variables candidates courantes
  model:= TCalcLogisticReg.create(NIL,self.Target,curInputs,self.Inputs);
  ok:= model.learning(self.WorkData.Examples);
  //pour le niveau tudi
  if (ok = true)
   then
    begin
     levelRes:= TLevelResultat.create();//rsultat pour un niveau
     levelRes.AIC:= model.AIC;
     levelRes.CHi2:= model.CHI2;
     levelRes.ddl:= model.ddlCHI2;
     levelres.pValue:= model.pCHI2;
     //valuer chaque variable
     for j:= 0 to pred(inputList.getCount()) do
      begin
       res:= TResultat.create(inputList.getCandidat(j));
       res.Chi2:= model.Wald[succ(j)];
       res.PValue:= model.pWald[succ(j)];
       levelRes.addresult(res);
      end;
     levelRes.sortAscendant();
     //ajouter ce niveau
     Soluce.addLevel(levelRes);
     //tester maintenant si le retrait est souhaitable
     if (levelRes.getResult(0).PValue > pSeuil)
      then
       begin
        //retrait de la variable
        inputList.remove(inputList.indexOf(levelRes.getResult(0).Candidat));
        //et la dsigner comme  enlever
        levelRes.Best:= levelRes.getResult(0);
       end;
    end;
  model.Free();
  //dtecter la variable qui minimise le CHI-2 de Wald
  //retirer la variable si p-value infrieure au p-value seuil
 UNTIl not(ok) OR (inputList.getCount() = 0) OR (inputList.getCount() = nbInputs);
 //recopier la solution
 //rajouter maintenant dans la sortie les variables slectionnes --> ceux qui sont rests dans la inputList sont les slectionns
 for j:= 0 to pred(inputList.getCount()) do
  self.OutputData.LstAtts[asInput].Add(inputList.getCandidat(j).Variable);
 //***************************************************************
 inputList.Free();
 curInputs.Free();
end;

{ TOpPrmFSRegLogBackward }

function TOpPrmFSRegLogBackward.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFSLogRegBase.CreateFromOpPrm(self);
end;

initialization
 RegisterClass(TMLGenFSBackwardLogReg);
end.
