(****************************************************************)
(* UCompISRuleBased.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Slectionner un sous-ensemble d'individus  partir d'une ou plusieurs rgles imbriques)
@author(Ricco)
@created(11/08/2006)

La liste de rgles est reprsente par une liste de chanes :
-> chaque chane est une rgle
-> chaque chane doit tre parse pour lire ses propositions

}

unit UCompISRuleBased;

interface

USES
        Classes, IniFiles, Forms, Contnrs,
        UCompDefinition, UOperatorDefinition,
        UCompISDefinition,
        UCompManageDataset,
        UDatasetDefinition;

TYPE
        {gnrateur}
        TMLGenCompISRuleBased = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant}
        TMLCompISRuleBased = class(TMLCompIS)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur}
        TOpISRuleBased = class(TOpLocalData)
        private
        //liste des premisses
        FLstPremisses: TObjectList;
        //tableau des individus couverts
        FTabCovered: array of integer;
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    CoreExecute(): boolean; override;
        public
        constructor Create(AOwner: TObject); override;
        destructor  Destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        end;

        //mode de combinaison des rgles pour la slection
        //Or : au moins une des rgles dclenches
        //And : toutes les rgles doivent tre dclenches
        TTypeISRuleCombination = (rcbOr, rcbAnd);

CONST
        STR_RULE_COMBINATION: array[TTypeISRuleCombination] of string = ('Or','And');

TYPE
        {paramtrage}
        TOpPrmISRuleBased = class(TOperatorParameter)
        private
        //liste des rgles
        FLstRules: TStringList;
        //mode de combinaison : par dfaut "And"
        FRuleComb: TTypeISRuleCombination;
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        destructor  destroy(); override;
        function    getHTMLParameters(): string; override;
        //I-O
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        //
        property    LstRules: TStringList read FLstRules write FLstRules;
        property    RuleCombination: TTypeISRuleCombination read FRuleComb write FRuleComb;
        end;


implementation

USES
    Sysutils, UDlgOpPrmISRuleBased, UCalcRulesImplementation,
  UDatasetExamples, UDatasetImplementation, UStringAddBuffered,
  UConstConfiguration, ULogFile;

{ TMLGenCompISRuleBased }

procedure TMLGenCompISRuleBased.GenCompInitializations;
begin
 FMLComp:= mlcInstanceSelection;
end;

function TMLGenCompISRuleBased.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompISRuleBased;
end;

{ TMLCompISRuleBased }

function TMLCompISRuleBased.getClassOperator: TClassOperator;
begin
 result:= TOpISRuleBased;
end;

{ TOpISRuleBased }

function TOpISRuleBased.CoreExecute: boolean;
var premisse: TRulePremisse;
    k, i, example, nbPremisses: integer;
    exSource, exOut: TExamples;
    hit: integer;
    ok: boolean;
begin
 TRY
 //vider les rgles prcdentes
 FLstPremisses.Clear();
 //reconstruire la base de rgles filtrantes
 for k:= 0 to pred((self.PrmOp as TOpPrmISRuleBased).LstRules.Count) do
  begin
   premisse:= TRulePremisse.decomposeReglePremisse((self.PrmOp as TOpPrmISRuleBased).LstRules[k],self.WorkData.LstAtts[asAll]);
   if (premisse <> nil)
    then
     begin
      FLstPremisses.Add(premisse);
      //TraceLog.WriteToLogFile(format('ADD RULE-BASED IS >> %s',[premisse.getHTMLDescription()]));
     end;
  end;
 //brancher les pointeur sur les listes d'individus
 exSource:= self.WorkData.Examples;
 exOut:= self.OutputData.Examples;
 nbPremisses:= FLstPremisses.Count;
 //filtrer
 if (nbPremisses > 0)
  then
   begin
    setLength(FTabCovered,FLstPremisses.Count);
    for k:= 0 to pred(length(FTabCovered)) do FTabCovered[k]:= 0;
    //ajouter maintenant
    exOut.BeginAdd();
    for i:= 1 to exSource.Size do
     begin
      example:= exSource.Number[i];
      hit:= 0;
      for k:= 0 to pred(nbPremisses) do
       begin
        premisse:= FLstPremisses.Items[k] as TRulePremisse;
        if premisse.testExample(example)
         then
          begin
           //au moins une fois
           inc(hit);
           //recension
           inc(FTabCovered[k]);
          end;
       end;
      //selon le mode de combinaison
      case (self.PrmOp as TOpPrmISRuleBased).RuleCombination of
       rcbOr : ok:= (hit > 0);
       //par dfaut, c'est AND
       else ok:= (hit = nbPremisses);
      end;
      //alors ?
      if ok
       then exOut.AddExample(example);
     end;
    exOut.EndAdd();
   end
  else exOut.Copy(exSource);
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

constructor TOpISRuleBased.Create(AOwner: TObject);
begin
 inherited;
 FLstPremisses:= TObjectList.Create(TRUE);
end;

destructor TOpISRuleBased.Destroy;
begin
 Finalize(FTabCovered);
 FLstPremisses.Free();
 inherited;
end;

function TOpISRuleBased.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmISRuleBased;
end;

function TOpISRuleBased.getHTMLResultsSummary: string;
var k: integer;
    s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Rule</TH><TH>Examples</TH></TR>';
 for k:= 0 to pred(length(FTabCovered)) do
  s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Rule %d</TD><TD align="right">%d</TD></TR>',[succ(k),FTabCovered[k]]);
 s:= s + '</TABLE>';
 s:= s + format('<BR><BR>Total : <b>%d</b> examples selected from <b>%d</b>',[self.OutputData.Examples.Size,self.workdata.examples.Size]);
 result:= s;
end;

{ TOpPrmISRuleBased }

function TOpPrmISRuleBased.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmISRuleBased.CreateFromOpPrm(self);
end;

destructor TOpPrmISRuleBased.destroy;
begin
 FLstRules.Free();
 inherited;
end;

function TOpPrmISRuleBased.getHTMLParameters: string;
var s: string;
    k: integer;
begin
 s:= Format('Rule combination : <b>%s</b><br><br><u><b>Rules</b></u><br>',[STR_RULE_COMBINATION[FRuleComb]]);
 for k:= 0 to pred(self.FLstRules.Count) do
  s:= s + format('Rule <b>%d</b> : %s <br>',[succ(k),self.FLstRules[k]]);
 result:= s;
end;

procedure TOpPrmISRuleBased.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
var k,nb: integer;
    chaine: string;
begin
 //mode de combinaison
 FRuleComb:= TTypeISRuleCombination(prmINI.ReadInteger(prmSection,'rule_combination',ord(FRuleComb)));
 //les rgles
 FlstRules.Clear();
 nb:= prmINI.ReadInteger(prmSection,'rules',0);
 for k:= 1 to nb do
  begin
   chaine:= prmINI.ReadString(prmSection,'rule_'+IntToStr(k),'');
   if (chaine <> '') then FLstRules.Add(chaine);
  end;
end;

procedure TOpPrmISRuleBased.LoadFromStream(prmStream: TStream);
var k,nb: integer;
    chaine: string;
begin
 //mode de combinaison
 prmStream.ReadBuffer(FRuleComb,sizeof(FRuleComb));
 //les rgles
 FLstRules.Clear();
 prmStream.ReadBuffer(nb,sizeof(nb));
 for k:= 0 to pred(nb) do
  begin
   loadStringFromStream(chaine,prmStream);
   FLstRules.Add(chaine);
  end;
end;

procedure TOpPrmISRuleBased.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
var k: integer;
begin
 //mode de combinaison
 prmINI.WriteInteger(prmSection,'rule_combination',ord(FRuleComb));
 //les rgles
 prmINI.WriteInteger(prmSection,'rules',FLstRules.Count);
 for k:= 1 to FLstRules.Count do
  prmINI.WriteString(prmSection,'rule_'+IntToStr(k),FlstRules[pred(k)]);
end;

procedure TOpPrmISRuleBased.SaveToStream(prmStream: TStream);
var k,nb: integer;
    chaine: string;
begin
 //mode de combinaison
 prmStream.WriteBuffer(FRuleComb,sizeof(FRuleComb));
 //les rgles
 nb:= FLstRules.Count;
 prmStream.WriteBuffer(nb,sizeof(nb));
 for k:= 0 to pred(FLstRules.Count) do
  begin
   chaine:= FLstRules[k];
   saveStringToStream(chaine,prmStream);
  end;
end;

procedure TOpPrmISRuleBased.SetDefaultParameters;
begin
 FLstRules:= TStringList.Create();
 FRuleComb:= rcbAnd;
end;

initialization
 RegisterClass(TMLGenCompISRuleBased);
end.
