(*****************************************************************)
(* UCompPLSConfidenceInterval.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(PLS -- Calcul des intervalles de confiance des coefficients de la rgression par bootstrap)
@author(Ricco)
@created(30/10/2006)

Utilisation du bootstrap pour calculer les coefficients de la rgression PLS.
Mthode par les percentiles.

}

unit UCompPLSConfidenceInterval;

interface

USES
    Classes, Forms,
    IniFiles, Contnrs,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    JFGCompPLS;

TYPE
    //gnrateur
    TMLGenCompPLSConfInterval = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompPLSConfInterval = class(TMLCompLocalData)
    protected
    function    getClassOperator: TClassOperator; override;
    public
    //n'est connectable que sur un composant PLS regression
    class function isConnectableTo(destination: TMLComponent): boolean; override;
    end;

    //structure pour les coefficients
    TDataCoefRegPLS = array of single;

    //rsultats pour une variable
    TDataResultCoefOneVar = class
    private
    procedure   sort(var ra: TDataCoefRegPLS);
    public
    //coefficients standardiss
    stdCoef: TDataCoefRegPLS;
    //coefficients d-standardiss
    Coef: TDataCoefRegPLS;
    //constructeur
    constructor create(NbReplication: integer);
    //destructeur
    destructor  destroy(); override;
    //placer une valeur -- coefficients sur donnes standardises et normales
    procedure   setValue(replication: integer; Bstd, Bc: single);
    //trier les valeurs collectes
    procedure   sorting();
    end;

    //rsultats pour l'ensemble des variables
    TInputsResultsBootstrap = class
    public
    FInputs: TLstAttributes;
    FLstResult: TObjectList;
    constructor create(inputs: TLstAttributes; nbReplication: integer);
    destructor  destroy(); override;
    function    getInputAttributeResult(j: integer): TDataResultCoefOneVar;
    procedure   sorting();
    end;

    //rsultats pour chaque variable target
    TTargetsResultsBootstrap = class
    public
    FTargets: TLstAttributes;
    FLst: TObjectList;
    constructor create(targets, inputs: TLstAttributes; nbReplication: integer);
    destructor  destroy(); override;
    function    getTargetResult(i: integer): TInputsResultsBootstrap;
    procedure   sorting(); 
    end;

    //oprateur
    TOpPLSConfInterval = class(TOpLocalData)
    private
    FOpPLS: TOpPLS;
    FResult: TTargetsResultsBootstrap; 
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    CoreExecute(): boolean; override;
    public
    //rcuprer au passage l'oprateur PLS
    constructor Create(AOwner: TObject); override;
    destructor  destroy(); override;
    function    getHTMLResultsSummary(): string; override;    
    end;

    //paramtrage
    TOpPrmPLSConfInterval = class(TOperatorParameter)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //risque de 1re espce (df: 0.05)
    Alpha: double;
    //replications (20  2000, df: 100)
    NbReplication: integer;
    //montrer les rsultats pour les variables standardises
    ShowStdResult: boolean;
    //mthodes
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;
        

implementation

uses
    JFGCalcPLS, UDatasetExamples, UDatasetImplementation,
    UCalcRndGenerator, UStringAddBuffered, UConstConfiguration,
    SysUtils, UDlgOpPrmPlsConfidenceInterval, UDlgBaseOperatorParameter,
  ULogFile;

{ TMLGenCompPLSConfInterval }

procedure TMLGenCompPLSConfInterval.GenCompInitializations;
begin
 FMLComp:= mlcPLS;
end;

function TMLGenCompPLSConfInterval.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompPLSConfInterval;
end;

{ TMLCompPLSConfInterval }

function TMLCompPLSConfInterval.getClassOperator: TClassOperator;
begin
 result:= TOpPLSConfInterval; 
end;

class function TMLCompPLSConfInterval.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 result:= (destination is TMLCompPLS);
end;

{ TOpPLSConfInterval }

function TOpPLSConfInterval.CoreExecute: boolean;
var calc: TCalcPLS;
    inputs, targets: TLstAttributes;
    sample, examples: TExamples;
    k,j,i: integer;
    rnd: TRndGenerator;
begin
  //rcupration des infos. pour le calcul
  inputs:= self.WorkData.LstAtts[asInput];
  targets:= self.WorkData.LstAtts[asTarget];
  examples:= self.WorkData.Examples;
  //rsultats
  if assigned(FResult) then FResult.Free();
  FResult:= TTargetsResultsBootstrap.create(targets,inputs,(self.PrmOp as TOpPrmPLSConfInterval).NbReplication);
  //prparation
  TRY
    sample:= TExamples.Create(examples.Size);
    rnd:= TRndGenerator.Create(seedStandard,DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
    TRY
    //pour chaque replication
    for k:= 1 to (self.PrmOp as TOpPrmPLSConfInterval).NbReplication do
     begin
      //tirage alatoire avec remise
      examples.SampleReplicate(sample,rnd);
      //calcul
      calc:= TCalcPLS.create(inputs,targets,sample,true,(FOpPLS.PrmOp as TOpPrmPLS).NbAxisCreated,false);
      calc.runAnalysis(sample);
      //rcupration des rsultats
      for i:= 0 to pred(targets.count) do
       begin
        for j:= 0 to pred(inputs.Count) do
         FResult.getTargetResult(i).getInputAttributeResult(j).setValue(k,calc.stdCoefReg(i,j),calc.coefReg(i,j));
        //la constante n'a de sens que pour les coef. non standardiss
        FResult.getTargetResult(i).getInputAttributeResult(inputs.Count).setValue(k,0.0,calc.coefReg(i,inputs.Count));
       end;
      //dtruire l'oprateur de calcul
      calc.Free();
     end;
    //trier le tout
    FResult.sorting();
    FINALLY
    sample.Free();
    rnd.Free();
    END;
  result:= true;
  EXCEPT
  result:= false;
  END;
end;

constructor TOpPLSConfInterval.Create(AOwner: TObject);
begin
 inherited;
 FOpPLS:= (MLOwner as TMLComponent).Predecessor.Operator as TOpPLS;
end;

destructor TOpPLSConfInterval.destroy;
begin
 if assigned(FResult) then FResult.Free();
 FOpPLS:= nil;//surtout pas de suppression, un simple drfrencement
 inherited;
end;

function TOpPLSConfInterval.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmPLSConfInterval;
end;

function TOpPLSConfInterval.getHTMLResultsSummary: string;
var buf: TBufString;
    s: string;
    i,j,FCountX,FCountY: integer;
    calc: TCalcPLS;
    inf,sup: integer;
    demiAlpha: double;
    replication: integer;
    bgColor: string;
    bInf,bSup: double;
begin

  //prparatoire
  FCountX:= self.WorkData.LstAtts[asInput].Count;
  FCountY:= self.WorkData.LstAtts[asTarget].Count;
  calc:= FOpPLS.getCalc();

  //calcul des bornes pour affichage des percentiles
  replication:= (self.PrmOp as TOpPrmPLSConfInterval).NbReplication; 
  demiAlpha:= 0.5 * (self.PrmOp as TOpPrmPLSConfInterval).Alpha;
  inf:= trunc(replication*demiAlpha);
  sup:= replication - inf;

  if (inf < 1) then inf := 1;
  if (sup > replication) then sup:= replication;

  TraceLog.WriteToLogFile(format('[PLS CONFIDENCE INTERVAL] == INDEX INF : %d -- INDEX SUP : %d',[inf,sup]));

  //former la chane de caractre dans le buffer...
  buf:= TBufString.Create();
  buf.BeginUpdate();

  //*************************************************************
  //******** coefficients d-standardiss ***********************
  //*************************************************************

  s:= '<H3>Regression coefficients and confidence intervals</H3>';
  s:= s + HTML_HEADER_TABLE_RESULT;
  
  //en-tte les variables Y
  s:= s + HTML_TABLE_COLOR_HEADER_GREEN+'<TH>X/Y</TH>';
  for i:= 0 to pred(FCountY) do
   s:= s + format('<TH colspan="3">%s</TH>',[self.WorkData.LstAtts[asTarget].Attribute[i].Name]);
  s:= s + '</TR>';
  buf.AddStr(s);

  s:= HTML_TABLE_COLOR_HEADER_GRAY+'<TH>-</TH>';
  for i:= 0 to pred(FCountY) do
   s:= s + '<TH width="60">Value</TH><TH width="60">L.Bound</TH><TH width="60">U.Bound</TH>';
  s:= s + '</TR>';
  buf.AddStr(s);

  //les coeffs. pour chaque X
  for j:= 0 to FCountX do
   begin
    if (j < FCountX)
     then s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[self.WorkData.LstAtts[asInput].Attribute[j].Name])
     else s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',['constant']);//la constante
    for i:= 0 to pred(FCountY) do
     begin
      bInf:= FResult.getTargetResult(i).getInputAttributeResult(j).Coef[inf];
      bSup:= FResult.getTargetResult(i).getInputAttributeResult(j).Coef[sup];
      //significativement diffrent de zro ?
      if (bInf * bSup > 0)
       then bgColor:= HTML_BGCOLOR_DATA_RED
       else bgColor:= HTML_BGCOLOR_DATA_GRAY;   
      s:= s + format('<TD align="right" %s>%.6f</TD><TD align="right" %s>%.6f</TD><TD align="right" %s>%.6f</TD>',[bgColor,calc.coefReg(i,j),bgColor,bInf,bgColor,bSup]);
     end;
    s:= s + '</TR>';
    buf.AddStr(s);
   end;

  buf.AddStr('</table>');

  //*********************************************************************
  //****** coefficients sur variables standardises *********************
  //*********************************************************************

  if (self.PrmOp as TOpPrmPLSConfInterval).ShowStdResult
   then
    begin

    // misre le copier-coller....
    s:= '<H3>Regression coefficients and confidence intervals on standardized attributes</H3>';
    s:= s + HTML_HEADER_TABLE_RESULT;

    //en-tte les variables Y
    s:= s + HTML_TABLE_COLOR_HEADER_GREEN+'<TH>X/Y</TH>';
    for i:= 0 to pred(FCountY) do
     s:= s + format('<TH colspan="3">%s</TH>',[self.WorkData.LstAtts[asTarget].Attribute[i].Name]);
    s:= s + '</TR>';
    buf.AddStr(s);

    s:= HTML_TABLE_COLOR_HEADER_GRAY+'<TH>-</TH>';
    for i:= 0 to pred(FCountY) do
     s:= s + '<TH width="60">Value</TH><TH width="60">L.Bound</TH><TH width="60">U.Bound</TH>';
    s:= s + '</TR>';
    buf.AddStr(s);

    //les coeffs. pour chaque X
    for j:= 0 to pred(FCountX) do
     begin
      s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[self.WorkData.LstAtts[asInput].Attribute[j].Name]);
      for i:= 0 to pred(FCountY) do
       begin
        bInf:= FResult.getTargetResult(i).getInputAttributeResult(j).stdCoef[inf];
        bSup:= FResult.getTargetResult(i).getInputAttributeResult(j).stdCoef[sup];
        //significativement diffrent de zro ?
        if (bInf * bSup > 0)
         then bgColor:= HTML_BGCOLOR_DATA_RED
         else bgColor:= HTML_BGCOLOR_DATA_GRAY;   
        s:= s + format('<TD align="right" %s>%.6f</TD><TD align="right" %s>%.6f</TD><TD align="right" %s>%.6f</TD>',[bgColor,calc.stdCoefReg(i,j),bgColor,bInf,bgColor,bSup]);
       end;
      s:= s + '</TR>';
      buf.AddStr(s);
     end;

    buf.AddStr('</table>');
    end;

  buf.EndUpdate();
  result:= buf.BufS;
  buf.Free();

end;

{ TDataResultCoefOneVar }

constructor TDataResultCoefOneVar.create(NbReplication: integer);
begin
 inherited create();
 //commencer les indices  1
 setLength(stdCoef,NbReplication+1);
 setLength(coef,NbReplication+1);
end;

destructor TDataResultCoefOneVar.destroy;
begin
 Finalize(coef);
 Finalize(stdCoef);
 inherited;
end;

procedure TDataResultCoefOneVar.setValue(replication: integer; Bstd,
  Bc: single);
begin
 stdCoef[replication]:= Bstd;
 coef[replication]:= Bc;
end;

procedure TDataResultCoefOneVar.sort(var ra: TDataCoefRegPLS);
LABEL 99;
VAR
   l,j,ir,i: integer;
   rra: single;
   n: integer;
BEGIN
   //nombre de replication (1  n)
   n:= length(ra)-1;
   //
   l := (n DIV 2)+1;
   ir := n;
   WHILE true DO BEGIN
      IF (l > 1) THEN BEGIN
         l := l-1;
         rra := ra[l]
      END ELSE BEGIN
         rra := ra[ir];
         ra[ir] := ra[1];
         ir := ir-1;
         IF (ir = 1) THEN BEGIN
            ra[1] := rra;
            GOTO 99
         END
      END;
      i := l;
      j := l+l;
      WHILE (j <= ir) DO BEGIN
         IF (j < ir) THEN
            IF (ra[j] < ra[j+1]) THEN j := j+1;
         IF (rra < ra[j]) THEN BEGIN
            ra[i] := ra[j];
            i := j;
            j := j+j
         END ELSE
            j := ir+1
      END;
      ra[i] := rra
   END;
99:   END;


procedure TDataResultCoefOneVar.sorting;
begin
 self.sort(stdCoef);
 self.sort(coef);
end;

{ TInputsResultsBootstrap }

constructor TInputsResultsBootstrap.create(inputs: TLstAttributes;
  nbReplication: integer);
var j: integer;
begin
 inherited create();
 FInputs:= inputs;
 FLstResult:= TObjectList.Create(TRUE);
 for j:= 0 to inputs.Count do //ne pas oublier la constante
  FLstResult.Add(TDataResultCoefOneVar.create(nbReplication));
end;

destructor TInputsResultsBootstrap.destroy;
begin
 FLstResult.Free();
 inherited;
end;

function TInputsResultsBootstrap.getInputAttributeResult(
  j: integer): TDataResultCoefOneVar;
begin
 result:= FLstResult.Items[j] as TDataResultCoefOneVar;
end;

procedure TInputsResultsBootstrap.sorting;
var i: integer;
begin
 for i:= 0 to pred(FLstResult.Count) do
  self.getInputAttributeResult(i).sorting();
end;

{ TTargetsResultsBootstrap }

constructor TTargetsResultsBootstrap.create(targets,
  inputs: TLstAttributes; nbReplication: integer);
var i: integer;
begin
 inherited create();
 FTargets:= targets;
 FLst:= TObjectList.Create(true);
 for i:= 0 to pred(targets.count) do
  FLst.Add(TInputsResultsBootstrap.create(inputs,nbReplication));
end;

destructor TTargetsResultsBootstrap.destroy;
begin
 FLst.Free();
 FTargets:= nil;
 inherited;
end;

function TTargetsResultsBootstrap.getTargetResult(
  i: integer): TInputsResultsBootstrap;
begin
 result:= FLst.Items[i] as TInputsResultsBootstrap;
end;

procedure TTargetsResultsBootstrap.sorting;
var i: integer;
begin
 for i:= 0 to pred(FLst.Count) do
  self.getTargetResult(i).sorting();
end;

{ TOpPrmPLSConfInterval }

function TOpPrmPLSConfInterval.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmPlsConfInterval.CreateFromOpPrm(self);
end;

function TOpPrmPLSConfInterval.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Parameter</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD># replication</TD><TD>%d</TD></TR>',[NbReplication]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Aplha</TD><TD>%.4f</TD></TR>',[Alpha]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Show std coef.</TD><TD>%d</TD></TR>',[ord(ShowStdResult)]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmPLSConfInterval.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 Alpha:= prmINI.ReadFloat(prmSection,'alpha',Alpha);
 NbReplication:= prmINI.ReadInteger(prmSection,'replication',NbReplication);
 ShowStdResult:= prmINI.ReadBool(prmSection,'std_result',ShowStdResult);
end;

procedure TOpPrmPLSConfInterval.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(Alpha,sizeof(Alpha));
 prmStream.ReadBuffer(NbReplication,sizeof(NbReplication));
 prmStream.ReadBuffer(ShowStdResult,sizeof(ShowStdResult));
end;

procedure TOpPrmPLSConfInterval.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'alpha',Alpha);
 prmINI.WriteInteger(prmSection,'replication',NbReplication);
 prmINI.WriteBool(prmSection,'std_result',ShowStdResult);
end;

procedure TOpPrmPLSConfInterval.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(Alpha,sizeof(Alpha));
 prmStream.WriteBuffer(NbReplication,sizeof(NbReplication));
 prmStream.WriteBuffer(ShowStdResult,sizeof(ShowStdResult));
end;

procedure TOpPrmPLSConfInterval.SetDefaultParameters;
begin
 Alpha:= 0.05;
 NbReplication:= 100;
 ShowStdResult:= false;
end;

initialization
 RegisterClass(TMLGenCompPLSConfInterval);
end.
