(******************************************************************)
(* UCompRegForwardSel.pas - Copyright (c) 2005 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Regression multiple avec limination BACKWARD des variables)
@author(Ricco)
@created(03/05/2007)

Solution simple :
(1) rgression sur toutes les variables
(2) limination de la variable la moins significative
(3) rgression sur les variables restantes
(4) limination, etc. jusqu' ce qu'aucune limination ne soit possible

}
unit UCompRegBackwardElimination;

interface

USES
    Forms,Classes,IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompRegDefinition,
    UDatasetDefinition,
    UCalcStatDesContinuousPlus,
    UCalcRegLinMultiple,
    UCompRegMultiple,
    MatrixMM,
    UCompRegForwardSel;

TYPE
    //gnrateur
    TMLGenRegBackward = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompRegBackward = class(TCompRegMultiple)
    protected
    function    getClassOperator: TClassOperator; override;
    function    getGenericAttName(): string; override;
    end;

    //information sur chaque corrlation partielle
    TInfoTStudent = record
    //t
    t: single;
    //p-value
    p_value: single;
    end; 

    //oprateur -- la rgression est toujours une rgression avec constante
    TOpRegBackward = class(TOpRegMultiple)
    private
    //nombre d'observations ayant servi  la rgression
    FNbExamples: integer;
    //nombre de variables slectionnes
    FNbSelVar: integer;
    //liste originale des variables exognes
    FLstOrigExogenous: TLstAttributes;
    //tableau des informations pour chaque tape de calcul
    FArrayStep: array of array of TInfoTStudent;
    //tableau de numro de la var. slectionne  l'tape j
    FArrayRemovedVar: array of integer;
    //tableau des R
    FArrayR2: array of double;
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    public
    {dfinir une liste d'exognes vide}
    constructor Create(AOwner: TObject); override;
    {dtruire la liste d'exogne}
    destructor  Destroy; override;
    {excution effective}
    function    CoreExecute(): boolean; override;
    {rsum des rsultats}
    function    getHTMLResultsSummary(): string; override;
    end;

    //paramtrage, idem que slection FORWARD
    TPrmOpRegBackward = class(TPrmOpRegForward)
    end;

implementation

uses
    Sysutils, Math, FMath,
    UDlgOpPrmRegForward, UConstConfiguration, UDatasetImplementation;

{ TMLGenRegBackward }

procedure TMLGenRegBackward.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenRegBackward.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompRegBackward; 
end;

{ TMLCompRegBackward }

function TMLCompRegBackward.getClassOperator: TClassOperator;
begin
 result:= TOpRegBackward;
end;

function TMLCompRegBackward.getGenericAttName: string;
begin
 result:= 'bwReg';
end;

{ TOpRegBackward }

function TOpRegBackward.CoreExecute: boolean;
var
    sigLevel: double;
    //nombre de variables de dpart
    nbAllVar: integer;
    //calcul courant
    calc: TCalcRegMultiple;
    //id de variables
    i,j,jMin: integer;
    //valeur absolue de t
    t, tMin: double;
    //ecart-type des coefs.
    sigma: double;
    //p-value corresp.
    pvalue: double;
    //ddlCourant
    ddl: integer;
    //suite OK
    okRemoved: boolean;
    //test erreur
    okProcess: boolean;
    //attribut
    att: TAttribute;
    //numro d'tape
    numStep: integer;
begin
 okProcess:= true;
 TRY
 //rcuprer le niveau de signification demand
 sigLevel:= (self.PrmOp as TPrmOpRegBackward).SigLevel; 
 //n, p, ddl initial, etc.
 FNbExamples:= self.WorkData.Examples.Size;
 nbAllVar:= self.LstExogenous.Count;
 ddl:= FNbExamples - nbAllVar - 1;
 //prparer les tableaux, garder tout pour pouvoir tout afficher...
 SetLength(FArrayStep,nbAllVar,nbAllVar);
 for i:= 0 to pred(nbAllVar) do
  for j:= 0 to pred(nbAllVar) do
   FArrayStep[i][j].p_value:= 2.0;//pour invalider l'affichage par la suite
 //tableau de numro de variable limine  l'tape j
 SetLength(FArrayRemovedVar,nbAllVar);
 FillChar(FArrayRemovedVar[0],nbAllVar*sizeof(integer),0);//r-initialiser  zro pour viter les persistences !!!
 //tableau des R des rgressions
 SetLength(FArrayR2,nbAllVar);
 //numroter les variables en interne pour une recherche plus rapide
 FLstOrigExogenous.Assign(self.LstExogenous);
 for j:= 0 to pred(FLstOrigExogenous.Count) do FLstOrigExogenous.Attribute[j].Tag:= j+1;
 //zoo...
 okRemoved:= true; numStep:= 0;
 while (okRemoved = true) and (ddl > 0) do
  begin
   //rgression
   calc:= TCalcRegMultiple.create(Endogenous,LstExogenous,TRUE);
   calc.execute(self.WorkData.Examples);
   FArrayR2[numStep]:= calc.R2;
   //dtecter la variable la moins significative
   tMin:= Math.MaxDouble; jMin:= -1;
   for j:= 0 to pred(self.LstExogenous.Count) do
    begin
     //variable  considrer
     att:= self.LstExogenous.Attribute[j];
     //calculer les indicateurs
     sigma:= calc.SCA.Prendre(j+1,j+1);
     sigma:= sqrt(sigma);
     if (sigma > 0.0)
      then
       begin
        t:= calc.A.Prendre(j+1,1)/sigma;
        pvalue:= PStudent(ddl,abs(t));//test bilatral
       end
      else
       begin
        t:= 0.0;
        pvalue:= 1.0;
       end;
     //affecter
     FArrayStep[att.Tag-1][numStep].t:= t;
     FArrayStep[att.Tag-1][numStep].p_value:= pvalue;
     //dtecter le moins significatif
     if (abs(t) < tMin)
      then
       begin
        tMin:= abs(t);
        jMin:= j;
       end;
    end;
   //est-ce que la moins significative est rellement non-significative ?
   pvalue:= PStudent(ddl,abs(tMin));
   if (pvalue > sigLevel)
    then
     begin
      att:= self.LstExogenous.Attribute[jMin];
      self.LstExogenous.delete(att);//supprimer de la liste
      FArrayRemovedVar[numStep]:= att.Tag;//rfrencer comme supprime
     end
    else okRemoved:= false;
   //tape suivante
   inc(numStep);
   dec(ddl);
   calc.Free(); //!\ ne pas oublier...
  end;
 EXCEPT
 okProcess:= false;
 END;
 //si tout s'est bien pass, lancer la rgression finale, ha ! ha! si okProcess le deuximne test n'est pas effectu, donc pas de calcul...
 okProcess:= okProcess and (inherited CoreExecute());
 //qq bilan
 FNbSelVar:= self.LstExogenous.Count;
 result:= okProcess;
end;

constructor TOpRegBackward.Create(AOwner: TObject);
begin
 inherited;
 FLstOrigExogenous:= TLstAttributes.Create(FALSE,Workdata.LstAtts[asAll].Size);
end;

destructor TOpRegBackward.Destroy;
begin
 FLstOrigExogenous.Free();
 inherited;
end;

function TOpRegBackward.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpRegBackward;
end;

function TOpRegBackward.getHTMLResultsSummary: string;
var s: string;
    j,i: integer;
    nbAllVar,columns: integer;
begin
 nbAllVar:= FLstOrigExogenous.Count;
 columns := MIN(nbAllVar-FNbSelVar+1,nbAllVar);  
 //appeler l'affichage des coeffs de la rgression finale
 s:= self.getCoefsRegressionHTML();
 //afficher le tableau du processus de slection
 s:= s+'<H3>Backward Elimination Process</H3>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 //les tapes (en-tte)
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TD align="right">t <br>(p-value)</TD>';
 for j:= 0 to pred(columns) do
  s:= s+format('<TD align="center">Step %d</TD>',[j+1]);
 s:= s+'</TR>';
 //les ddl.
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>#var (d.f.)</TD>';
 for j:= 0 to pred(columns) do
  s:= s+format('<TD align="center">%d (%d)</TD>',[nbAllVar-j,FNbExamples-nbAllVar-1-j]);
 s:= s+'</TR>';
 //les R successifs
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>Adj.R  (R)</TD>';
 for j:= 0 to pred(columns) do
  s:= s+format('<TD align="center">%.3f  (%.3f)</TD>',[1.0-(-1.0+FNbExamples)/(-1.0+FNbExamples-(nbAllVar-j))*(1.0-FArrayR2[j]),FArrayR2[j]]);
 s:= s+'</TR>';
 //liste des variables supprimes  chaque tape
 s:= s+HTML_TABLE_COLOR_DATA_GREEN+'<TD>Removed</TD>';
 for j:= 0 to pred(columns) do
  begin
   if (FArrayRemovedVar[j] > 0)
    then s:= s + format('<td align="center">%s</td>',[FLstOrigExogenous.Attribute[FArrayRemovedVar[j]-1].Name])
    else s:= s + '<td align="center">-</td>';
  end;
 //dtail des calculs pour chaque variable  chaque tape
 for j:= 0 to pred(nbAllVar) do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<td %s>%s</td>',[HTML_BGCOLOR_HEADER_GRAY,FLstOrigExogenous.Attribute[j].Name]);
   for i:= 0 to pred(columns) do
    begin
     if (FArrayStep[j][i].p_value > 1.0)
      then s:= s + '<td align="center">-</td>'
      else
       begin
        if (j = pred(FArrayRemovedVar[i]))
         then s:= s + format('<td align="right" %s>%.2f <br>(%.4f)',[HTML_BGCOLOR_DATA_GREEN,FArrayStep[j][i].t,FArrayStep[j][i].p_value])
         else s:= s + format('<td align="right">%.2f <br>(%.4f)',[FArrayStep[j][i].t,FArrayStep[j][i].p_value]);
       end;
    end;
  end;
 s:= s + '</table>';
 //et l'analyse des rsidus
 //les rsidus
 s:= s+self.getResidualsHTML();
 //and then...
 result:= s;
end;

initialization
 registerclass(TMLGenRegBackward);
end.
