(**********************************************************************)
(* UCalcSpvTreeRandomTree.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Structure arbre de dcision pour les frets alatoires (Breiman, 2001))
@author(Ricco)
@created(22/07/2006)

C'est du C4.5 mais sans le post-lagage...

}
unit UCalcSpvTreeRandomTree;

interface

USES
    UCalcTreeStructureDefinition,
    UCalcSpvTreeDefinition,
    UOperatorDefinition,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcRndGenerator,
    UCalcSpvTreeC45,
    EzdslBAr;

TYPE
    //feuille
    TSplitLeafSpvRnd = class(TSplitLeafSpvC45)
    end;

    //split
    TSplitAttributSpvRnd = class(TSplitAttributSpvC45)
    protected
    function    getClassSplitLeaf(): TClassSplitLeaf; override;
    //test avec le gain standard...
    //function    ComputeGoodness(): double; override;
    end;

    //liste de splits
    TLstSplitAttSpvRnd = class(TLstSplitAttSpvC45)
    protected
    function  getClassSplitAttribut(): TClassSplitAttribut; override;
    public
    //c'est ici que le choix alatoire va se jouer !!!
    procedure   buildCandidateSplits(prmInputAtts: TLstAttributes); override;
    end;

    //noeud de l'arbre
    TMLTreeNodeSpvRnd = class(TMLTreeNodeSpvC45)
    private
    protected
    function  getClassLstSplitAttributes(): TClassLstSplitAttributes; override;
    procedure AssignConclusion(); override;
    public
    end;

    //structure d'arbre
    TMLTreeStructureSpvRnd = class(TMLTreeStructureSpvC45)
    private
    //gnrateur de nombre alatoire pour le tirage des attributs
    FRnd: TRndGenerator;
    //tableau des attributs  traiter sur un sommet
    FTabInputs: TBooleanArray;
    //nombre d'attributs  considrer lors du SPLIT
    FNbSplitAtt: integer; 
    protected
    function   getClassMLTreeNode(): TClassMLTreeNode; override;
    public
    constructor create(paramMethod: TOperatorParameter; prmTarget,prmInput: TLstAttributes; prmExamples: TExamples); override;
    destructor  destroy; override;
    procedure   PostPruning(); override;
    end;

    
implementation

uses
    Math,
    Sysutils,
    UCompSpvTreeRandomTree, ULogFile;

{ TSplitAttributSpvRnd }

(*
function TSplitAttributSpvRnd.ComputeGoodness: double;
var localNode: TMLTreeNodeSpv;
    total,partial,v: double;
    i: integer;
    leaf: TSplitLeafSpv;
begin
 localNode:= self.Node as TMLTreeNodeSpv;
 total:= localNode.DistClass.Shannon();
 partial:= 0.0;
 for i:= 0 to pred(self.Count) do
  begin
   leaf:= self.SplitLeaf[i] as TSplitLeafSpv;
   v:= 1.0*leaf.Dist.TabFreq.Value[0]/(1.0*localNode.DistClass.TabFreq.Value[0]);
   partial:= partial + v * leaf.Dist.Shannon();
  end;
 //le gain d'entropie
 //comme la construction de l'arbre n'est pas ordonn, il est inutile de pondrer par le poids du neoud
 //il en serait tout autrement s'il s'agissait d'un graphe... ou de la mthode Catlett limit en splits...
 result:= -1.0*(total-partial);
end;
*)

function TSplitAttributSpvRnd.getClassSplitLeaf: TClassSplitLeaf;
begin
 result:= TSplitLeafSpvRnd;
end;

{ TLstSplitAttSpvRnd }

procedure TLstSplitAttSpvRnd.buildCandidateSplits(
  prmInputAtts: TLstAttributes);
var j,jMax: integer;
    att: TAttribute;
    split: TSplitAttribut;
    goodMax: double;
    tree: TMLTreeStructureSpvRnd;
begin
 tree:= self.Node.TreeStructure as TMLTreeStructureSpvRnd;
 
 //inialiser le tableau des attributs
 if (tree.FNbSplitAtt >= prmInputAtts.Count)
  then tree.FTabInputs.SetAllTrue()
  else
   begin
    tree.FTabInputs.SetAllFalse();
    //le remplir
    for j:= 1 to tree.FNbSplitAtt do
     tree.FTabInputs[tree.FRnd.IRanMarRange(prmInputAtts.Count)]:= TRUE;
    tree.FTabInputs.RefreshCount();
    //vrifier que l'on a le bon nombre -- un peu bourrin mais a ne peut que marcher
    while (tree.FTabInputs.Count < tree.FNbSplitAtt) do
     begin
      tree.FTabInputs[tree.FRnd.IRanMarRange(prmInputAtts.Count)]:= TRUE;
      tree.FTabInputs.RefreshCount();
     end;
   end;

 //TraceLog.WriteToLogFile(format('***** NEW SESSION FOR SPLITTING ***** >> [%d] attributes used -- [%d] examples',[tree.FTabInputs.Count,self.Node.Examples.Size]));

 //retour au calcul standard
 jMax:= -1;
 goodMax:= -1.0e308;
 //pour chaque attribut candidat
 for j:= 0 to pred(tree.FTabInputs.Capacity) do
  begin
   //vrifier qu'il a t inclu dans la liste
   if (tree.FTabInputs[j] = FALSE)
    then FLstSplitAttributes.Add(NIL)
    else
     begin
      att:= prmInputAtts.Attribute[j];
      split:= self.getClassSplitAttribut.Create(Node.PrmMethod,att,Node);
      //afficher les infos
      //TraceLog.WriteToLogFile(format('%s>>%.4f,%d',[split.Attribute.Name,split.GoodnessSplit,ord(split.AcceptSplit)]));
      //tester si max
      if split.AcceptSplit and (split.GoodnessSplit>goodMax)
       then
        begin
         jMax:= j;
         //ce n'est pas une fonction, on peut l'appeler plusieurs fois
         goodMax:= split.GoodnessSplit;
        end;
      //l'ajouter dans la liste
      FLstSplitAttributes.Add(split);
     end;
  end;

 //TraceLog.WriteToLogFile(format('%.4f value is selected',[goodMax]));

 //ok donc, on dispose des bonnes rfrences
 FNumberSplit:= jMax;
 FGoodnessSplit:= goodMax;
end;

function TLstSplitAttSpvRnd.getClassSplitAttribut: TClassSplitAttribut;
begin
 result:= TSplitAttributSpvRnd;
end;

{ TMLTreeNodeSpvRnd }

procedure TMLTreeNodeSpvRnd.AssignConclusion;
var iMax: TTypeDiscrete;
begin
 iMax:= DistClass.TabFreq.getIndexMaxValue();
 //si pas de conclusion alors conclusion alatoire
 if (iMax = 0)
  then iMax:= succ(random(DistClass.Attribute.nbValues));
 //affecter
 FNodeConclusion:= iMax;
end;

function TMLTreeNodeSpvRnd.getClassLstSplitAttributes: TClassLstSplitAttributes;
begin
 result:= TLstSplitAttSpvRnd;
end;

{ TMLTreeStructureSpvRnd }

constructor TMLTreeStructureSpvRnd.create(paramMethod: TOperatorParameter;
  prmTarget, prmInput: TLstAttributes; prmExamples: TExamples);
begin
 //initialisations spcifiques
 FRnd:= TRndGenerator.Create(seedStandard,DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
 FTabInputs:= TBooleanArray.Create(prmInput.Count);
 FNbSplitAtt:= (paramMethod as TOpPrmSpvRndTree).NbSelAtt;
 //si on doit le fixer nous mme
 if (FNbSplitAtt <= 0)
  then
   begin
    //formule de derrire les fagots
    FNbSplitAtt:= TRUNC(Math.Log2(1.0 * prmInput.Count) + 1);
    //et dans le cas o il est rellement rfractaire
    if (FNbSplitAtt <= 0) then FNbSplitAtt:= 1;
   end;
 //vrifier que l'on ne dborde pas
 if (FNbSplitAtt > prmInput.Count) then FNbSplitAtt:= prmInput.Count;

 //suite normale...
 inherited;
end;

destructor TMLTreeStructureSpvRnd.destroy;
begin
 FTabInputs.Free();
 FRnd.Free();
 inherited;
end;

function TMLTreeStructureSpvRnd.getClassMLTreeNode: TClassMLTreeNode;
begin
 result:= TMLTreeNodeSpvRnd;
end;

procedure TMLTreeStructureSpvRnd.PostPruning;
begin
 //rien de nouveau...
end;

end.
