(**********************************************************************)
(* UCompSpvTreeRandomTree.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(dfinition du composant de calcul d'un arbre de dcision "alatoire" - Essentiellement pour les "Random Forest")
@author(Ricco)
@created(22/07/2006)

On drive simplement une srie de classes de C4.5, avec 2 nouveauts

(1) les paramtres sont fixs en interne de manire  avoir des arbres profonds
(2) un nouveau paramtres -- nombre de variables  slectionner sur les noeuds -- est introduit

}

unit UCompSpvTreeRandomTree;

interface

USES
    Classes, IniFiles, Forms,
    UCompDefinition, UOperatorDefinition, UCompSpvLDefinition,
    UCalcTreeStructureDefinition, 
    UCompSpvTreeC45;

TYPE

        {le gnrateur de composant Supervised}
        TMLGCompSpvRndTree = class(TMLGCompSpvTreeC45)
        protected
         procedure   GenCompInitializations(); override;
        public
         function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant SpvTree}
        TMLCompSpvRndTree = class(TMLCompSpvTreeC45)
        protected
         function    getClassOperator: TClassOperator; override;
        end;

        {l'oprateur}
        TOpSpvRndTree = class(TOpSpvTreeC45)
        protected
         function    getClassParameter: TClassOperatorParameter; override;
         function    getClassSpvLearning(): TClassCalcSpvLearning; override;
        end;

        {paramtre de l'algo}
        TOpPrmSpvRndTree = class(TOpPrmSpvTreeC45)
        private
         //nombre d'attributs slectionns pour la segmentation
         FNbSelAtt: integer;
        protected
         procedure   SetDefaultParameters(); override;
         function    CreateDlgParameters(): TForm; override;
        public
         function    getHTMLParameters(): string; override;
         {I/O FLUX}
         procedure   LoadFromStream(prmStream: TStream); override;
         procedure   SaveToStream(prmStream: TStream); override;
         {I/O Fichier INI}
         procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
         procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
         property    NbSelAtt: integer read FNbSelAtt write FNbSelAtt;
        end;

        {algo d'arbre de dcision supervise}
        TCalcSpvRndTree = class(TCalcSpvTreeC45)
        protected
         function    getClassTreeStructureSpv(): TClassMLTreeStructure; override;
        end;

implementation

USES
    Sysutils,
    UCalcSpvTreeRandomTree, UDlgOpPrmSpvTreeRandomTree;


{ TMLGCompSpvRndTree }

procedure TMLGCompSpvRndTree.GenCompInitializations;
begin
 FMLComp:= mlcSpvLearning;
end;

function TMLGCompSpvRndTree.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvRndTree; 
end;

{ TMLCompSpvRndTree }

function TMLCompSpvRndTree.getClassOperator: TClassOperator;
begin
 result:= TOpSpvRndTree;
end;

{ TOpSpvRndTree }

function TOpSpvRndTree.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvRndTree;
end;

function TOpSpvRndTree.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvRndTree;
end;

{ TOpPrmSpvRndTree }

function TOpPrmSpvRndTree.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmSpvTreeRandomTree.CreateFromOpPrm(self);
end;

function TOpPrmSpvRndTree.getHTMLParameters: string;
begin
 result:= format('<b>Nb att for split = %d',[FNbSelAtt]);
end;

procedure TOpPrmSpvRndTree.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbSelAtt:= prmINI.ReadInteger(prmSection,'nb_att_for_split',FNbSelAtt);
end;

procedure TOpPrmSpvRndTree.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNbSelAtt,sizeof(FNbSelAtt));
end;

procedure TOpPrmSpvRndTree.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_att_for_split',FNbSelAtt);
end;

procedure TOpPrmSpvRndTree.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNbSelAtt,sizeof(FNbSelAtt));
end;

procedure TOpPrmSpvRndTree.SetDefaultParameters;
begin
 self.PLevel:= 0.25;//ne sert  rien en fait puisque pas de pruning mais pour assurer la cohrence...
 self.SizeMinLeaf:= 1;
 //par dfaut -1, cela veut dire qu'il prendra automatiquement la valeur "optimale" INT(log2(J)+1)
 FNbSelAtt:= -1;
end;

{ TCalcSpvRndTree }

function TCalcSpvRndTree.getClassTreeStructureSpv: TClassMLTreeStructure;
begin
 result:= TMLTreeStructureSpvRnd;
end;

initialization
 RegisterClass(TMLGCompSpvRndTree);
end.
