(**************************************************************************)
(* UCompSpvLogisticRegression.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(**************************************************************************)

{
@abstract(Rgression logistique, attention l'attribut classe ne peut prendre que deux modalits !!!)
@author(Ricco)
@created(12/01/2004)

Premire bauche, se documenter plus en dtail pour les statistiques indiquant la qualit de la rgression
(globalement, test du chi-2, les diffrents R2, etc.), et les aides  l'interprtation (odds-ratios).

L'algo d'optimisation utilis ici est Levenberg-Marquardt, qui a l'avantage de proposer directement
la matrice hessienne  l'optima, et donc une estimation adquate de la variance.

new -- 15/04/2005 -- intgration de la version Juillet 2004 de la biblio de J. DEBORD -- biblio "d6_xxxxx.pas" -- seuls les parties utiles  l'optimisation ont t intgres
new -- 03/05/2006 -- calcul direct du vecteur gradient et de la matrice hessienne (cf. Nakache & Confais, 2005 -- pp. 82 et annexes + doc. rcupre sur internet V. Jain Fv.2006)
new -- 03/05/2006 -- autre astuce norme, on centre et rduit automatiquement les donnes avant les calculs, a vite les dbordements sur l'exponentielle (!!!)
new -- 19/08/2007 -- calcul des pseudo-R^2
new -- 01/12/2007 -- rorganisation du code, sans modification des calculs (compatibilit avec la tentative de Reg.Log.Multinomiale)
new -- 02/12/2007 -- tentative d'intgration du code de OpenStat mais la procdure finale est trs instable... mme quand on y introduit la rgularisation
new -- 03/12/2007 -- les donnes centres rduites sont charges dans un tableau temporaire, indispensable pour acclrer les calculs mais trs gourmand en mmoire (tableau de single quand mme)
new -- 03/12/2007 -- nouvelle initialisation avec le modle initial compos uniquement de la constante
new -- 07/12/2007 -- ajouter un terme de rgularisation dans la diagonale de la matrice hessienne

}
unit UCompSpvLogisticRegression;

interface

USES
        Forms, Contnrs,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcDistribution,
        Stat,
        //new test with biblio ver. 07/2004
        d6_matrices, d6_fmath, d6_optim,
        UCalcSpvStructScore,
        UCalcStatDes;

TYPE
        {le gnrateur de composant Supervised}
        TMLGCompLogisticReg = class(TMLGenCompSpvLearning)
                              protected
                              procedure   GenCompInitializations(); override;
                              public
                              function    GetClassMLComponent: TClassMLComponent; override;
                              end;

        {le composant Spv}
        TMLCompLogisticReg = class(TMLCompSpvLearning)
                             protected
                             function    getClassOperator: TClassOperator; override;
                             function    GetLogResultDescription(): string; override;
                             end;

        {l'oprateur}
        TOpLogisticReg = class(TOpSpvLearningContinuous)
                          protected
                          function    ConnectClassAtt(prmData: TMLDataset): boolean; override; 
                          function    getClassParameter: TClassOperatorParameter; override;
                          function    getClassSpvLearning(): TClassCalcSpvLearning; override;
                          end;

        {paramtrage de l'oprateur}
        TOpPrmLogisticReg = class(TOpPrmSpvLearning)
                            protected
                            {pas de paramtres pour l'instant, ajouter les options plus tard}
                            function    CreateDlgParameters(): TForm; override;
                            end;

        TTabXCorrected = array of array of single;

        {le calculateur}
        TCalcLogisticReg = class(TCalcSpvLearning)
                           protected
                           NbClasses: integer;//nombre de modalits de la variables  prdire
                           NbVar: integer;//nombre de variables
                           NbParam: integer;//nombre total de paramtres  grer --> (NbClasses-1)*(NbVar+1)
                           {Code d'erreur renvoye par la procdure d'optimisation}
                           FErrCodeOptimization: integer;
                           {la valeur optimise}
                           FMinLogLikelihood: float;
                           {La valeur du modle uniquement compos de la constante}
                           FNullLogLikelihood: float;
                           {les coefficients de l'quation}
                           FCoef: TVector;
                           {l'inverse de la matrice hessienne}
                           FHInv: TMatrix;
                           {le vecteurt des cart-types}
                           FStdDev: TVector;
                           {Les vecteurs Wald et p-value de Wald}
                           FWald, FPValueWald: TVector;
                           {les individus  manipuler}
                           FCurExamples: TExamples;
                           {Chi-2 et proba associe}
                           FChi2_ddl: integer;
                           FChi2, FPrChi2: double;
                           {Statistiques descriptives sur les descripteurs}
                           FStatDes: TLstCalcStatDesContinuous;
                           {tableaux internes pour centrage-rduction}
                           FCentrage,FReduction: array of float;
                           {tableaux des probas individus}
                           //FTabPk: array of float;
                           {Pseudo-R^2 -- Mc Fadden, Cox & Snell et Nagelkerke}
                           FR2M, FR2CS, FR2N: double;
                           {AIC et BIC}
                           AICModel, AICIntercept, BICModel, BICIntercept: double;
                           {new -- R.R. -- 03/12/2007 -- tableau temporaire pour les donnes}
                           FXData: TTabXCorrected;//<<a bouffe de la mmoire certes, mais bon vu les ordinateurs actuels ???
                           {calcul des paramtres de centrage et rduction}
                           procedure   computePrmCentrageReduction(); virtual;
                           {rcuprer la valeur corrige d'un individu (example) sur une variable (j)}
                           function    getCorrectedValue(j: integer; example: integer): float;
                           {initialiser les coefficients}
                           procedure   initCoefs(); virtual;
                           {calculer les statistiques, ex. wald etc.}
                           procedure   CalcStats(); virtual;
                           {calculer la log-vraisemblance de l'intercept seule}
                           function    computeDevianceInterceptOnly(): double; virtual;
                           {calculer les statistqiues standards}
                           procedure   computeStdStats(); virtual;
                           {calculer les statistiques spcifiques}
                           procedure   computeSpecificStats(); virtual;
                           {la fonction  optimiser}
                           function    Log_Likelihood(X: TVector): float; virtual;
                           {new -- 03/05/2006 -- calcul de la matrice hessienne et du vecteur gradient}
                           procedure   directHessGrad(Func : TFuncNVar; X: TVector; Lbound, Ubound: Integer; G: TVector; H: TMatrix); virtual;
                           {new -- 02/12/2007 -- calcul du vecteur gradient uniquement}
                           procedure   directGrad(Func : TFuncNVar; X: TVector; Lbound, Ubound: Integer; G: TVector);
                           {new -- 02/12/2007 -- fixed hessian optimizer -- ne calculer la matrice hessienne qu'une fois -- cf. http://research.microsoft.com/~minka/papers/logreg/minka-logreg.pdf}
                           function   optimizationFixedHessian_plus_Marquardt(Func: TFuncNVar; Grad: TGradientObject; HessGrad : THessGradObject; X: TVector; Lbound, Ubound : Integer; MaxIter : Integer; Tol: Float; var F_min : Float; H_inv : TMatrix) : Integer;
                           {new -- 02/12/2007 -- fixed hessian suivi de Newton Raphson}
                           function   optimizationFHNewtonRaphson(Func: TFuncNVar; Grad: TGradientObject; HessGrad : THessGradObject; X: TVector; Lbound, Ubound : Integer; MaxIter : Integer; Tol: Float; var F_min : Float; H_inv : TMatrix) : Integer;
                           {new -- 03/12/2007 -- Newton-Raphson tout bte}
                           function   optimizationNewtonRaphson(Func: TFuncNVar; Grad: TGradientObject; HessGrad : THessGradObject; X: TVector; Lbound, Ubound : Integer; MaxIter : Integer; Tol: Float; var F_min : Float; H_inv : TMatrix) : Integer;
                           {new -- 03/12/2007 -- BFGS + Newton-Raphson}
                           function   optimizationBFGSNewtonRaphson(Func: TFuncNVar; Grad: TGradientObject; HessGrad : THessGradObject; X: TVector; Lbound, Ubound : Integer; MaxIter : Integer; Tol: Float; var F_min : Float; H_inv : TMatrix) : Integer;
                           {new -- 02/12/2007 -- adapte de "Logistic Regression Adapted from John C. Pezzullo -- Java program at http://members.aol.com/johnp71/logistic.html -- rcupr dans OpenStat}
                           function   optimizationFast(X: TVector; MaxIter: integer; Tol: float; var F_Min: float; H_inv : TMatrix): integer;
                           {lancer l'optimisation}
                           procedure   RunOptimization(); virtual;
                           {corriger les coefficients calculs}
                           procedure   CorrectComputedCoefs(); virtual;
                           {crer les structures de calcul}
                           procedure   createStructures(); override;
                           {vider les structures}
                           procedure   destroyStructures(); override;
                           {apprentissage proprement dit}
                           function    coreLearning(examples: TExamples): boolean; override;
                           {Crer le tableau temporaire des probabilits, etc.}
                           procedure   createTempTabProba(examples: TExamples); virtual;
                           {dtruire les tableaux temporaires des probabilits, etc.}
                           procedure   destroyTempTabProba(); virtual;
                           {Description de l'en-tte des rsultats globaux}
                           function    getHTMLRefValue(): string; virtual;
                           {Description des rsultats globaux}
                           function    getHTMLGlobalResult(): string;
                           {Description des coefficients}
                           function    getHTMLCoefficients(): string; virtual;
                           {Autres descriptions, ex. les odds-ratios}
                           function    getHTMLPostResults(): string; virtual;
                           public
                           procedure   getScore(example: integer; var postProba: TTabScore); override;
                           function    getHTMLResults(): string; override;
                           //proprits
                           property    Wald: TVector read FWald;
                           property    pWald: TVector read FPValueWald;
                           property    AIC: double read AICModel;
                           property    CHI2: double read FChi2;
                           property    ddlCHI2: integer read FChi2_ddl;
                           property    pCHI2: double read FPrChi2;
                           end;


implementation

USES
        Windows, Sysutils, Classes, Math,
        fmath,
        UConstConfiguration, UStringsResources, ULogFile;

CONST
        MAX_ITER_OPTMIZATION_LOGISTIC = 50;
        TOLERANCE_OPTIMIZATION_LOGISTIC = 1.0e-10;//1.0e-3

        TOLERANCE_DECREASE_LOGISTIC = 1.0e-10;

        MIN_VALUE_FOR_STDDEV = 1.0e-6;

        LAMBDA_FOR_REGULARIZATION = 0.0001;


{ TMLGCompLogisticReg }

procedure TMLGCompLogisticReg.GenCompInitializations;
begin
 FMLComp:= mlcSpvLearning;
end;

function TMLGCompLogisticReg.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompLogisticReg;
end;

{ TMLCompLogisticReg }

function TMLCompLogisticReg.getClassOperator: TClassOperator;
begin
 result:= TOpLogisticReg;
end;

function TMLCompLogisticReg.GetLogResultDescription: string;
begin
 result:= format('Logistic regression result generated [%s]',[self.Description]);
end;

{ TOpLogisticReg }

function TOpLogisticReg.ConnectClassAtt(prmData: TMLDataset): boolean;
var ok: boolean;
begin
 //discret ?
 ok:= inherited ConnectClassAtt(prmData);
 //binaire ?
 {$B-}
 ok:= ok and (ClassAttribute.nbValues = 2);
 //retour
 result:= ok;
end;

function TOpLogisticReg.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmLogisticReg;
end;

function TOpLogisticReg.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcLogisticReg;
end;

{ TOpPrmLogisticReg }

function TOpPrmLogisticReg.CreateDlgParameters: TForm;
begin
 result:= NIL;
end;

{ TCalcLogisticReg }

procedure TCalcLogisticReg.CalcStats;
begin
 FNullLogLikeliHood:= self.computeDevianceInterceptOnly();//modle sous H0
 self.computeSpecificStats();//spcifique au binaire ou multinomial
 self.computeStdStats();//globale aux rg. 
end;

procedure TCalcLogisticReg.getScore(example: integer;
  var postProba: TTabScore);
var v,proba: double;
    j: integer;
begin
 //new -- 03/05/2006 -- calcul de la proba P(Y=1/X), premire modalit de Y
 //>>on a chang les signes des coefs. pour cela
 //suite...
 //combinaison linaire
 v:= FCoef[0];
 for j:= 1 to descriptors.Count do
  v:= v+FCoef[j]*descriptors.Attribute[pred(j)].cValue[example];
 //la proba
 proba:= 1.0/(1.0+exp(-v));
 //affecation
 postProba[0]:= 1.0;
 postProba[1]:= proba;//<<  la place de "1.0-proba"
 postProba[2]:= 1.0-proba;//<<  la place de "proba"
end;

function TCalcLogisticReg.coreLearning(examples: TExamples): boolean;
begin
 result:= true;
 try
 FCurExamples:= examples;
 //new -- 03/05/2006 -- calcul des stat. descriptives pour centrage et rduction
 if assigned(FStatDes) then FStatDes.Free;
 FStatDes:= TLstCalcStatDesContinuous.Create(self.Descriptors,examples);
 //initialisation des coefficients
 self.initCoefs();
 //calcul des paramtres de centrage et rduction
 //!\ attention, les coefs. initiaux peuvent tre mis  zro si on dsactive une variable qui s'avre tre une constante
 self.computePrmCentrageReduction();
 //crer le tableau des probas temporaires
 self.createTempTabProba(examples);
 //optimimisation....
 self.RunOptimization();
 //corriger les coefficients suite au centrage-rduction - ne jamais oublier...
 self.CorrectComputedCoefs();
 //calcul des statistiques, wald, chi-2, etc.
 self.CalcStats();
 //dtruire les tableaux de proba temporaire
 self.destroyTempTabProba(); 
 except
 result:= false;
 end;
end;

procedure TCalcLogisticReg.createStructures;
begin
 NbClasses:= self.ClassAttribute.nbValues;
 NbVar:= self.Descriptors.Count;
 NbParam:= (NbClasses-1)*(NbVar+1);//+1  cause de la constante
 dimVector(FCoef,Descriptors.Count);
 dimMatrix(FHInv,Descriptors.Count,Descriptors.Count);
 dimVector(FStdDev,Descriptors.Count);
 dimVector(FWald,Descriptors.Count);
 dimvector(FPvalueWald,Descriptors.Count);
end;

procedure TCalcLogisticReg.destroyStructures;
begin
 //>
 //delVector(FCoef,Descriptors.Count);
 FCoef:= nil;
 //delMatrix(FHInv,Descriptors.Count,Descriptors.Count);
 FHInv:= nil;
 //delVector(FStdDev,Descriptors.Count);
 FStdDev:= nil;
 FWald:= nil;
 FPValueWald:= nil;
 //new tableaux pour centrage-rduction -- 03/05/2006
 Finalize(FCentrage);
 Finalize(FReduction);
 if assigned(FStatDes) then FStatDes.Free;
end;

function TCalcLogisticReg.getHTMLResults: string;
var s: string;
begin
 s:= '';
 if (FErrCodeOptimization = OPT_NON_CONV)
  then s:= Format('<P><B>Warning -- No convergence after %d iterations </B></P>',[MAX_ITER_OPTMIZATION_LOGISTIC]);
 //valuation globale
 s:= s + self.getHTMLGlobalResult();
 //les coefs.
 s:= s + self.getHTMLCoefficients();
 //autres affichage (odds-ratio etc.)
 s:= s + self.getHTMLPostResults();
 //renvoyer le tout
 result:= s;
end;

procedure TCalcLogisticReg.initCoefs;
var i: integer;
    v: double;
begin
 //constante
 v:= StatClassAtt.TabFreq.Frequence[2];
 FCoef[0]:= ln(v/(1.0-v));//initial model 
 for i:= 0 to descriptors.Count do
  FCoef[i]:= 0.0;
end;

function TCalcLogisticReg.Log_Likelihood(X: TVector): d6_fmath.float;
var i,j: integer;
    //s,v,tmpY: double;
    s,v,tmpY: extended;//pour prvenir les dbordements de capacit...
    example: integer;
begin
 TRY
 s:= 0.0;
 //pour chaque observation
 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   //**** produit scalaire ****
   //constante
   v:= X[0];
   for j:= 1 to descriptors.Count do
    //v:= v + X[j] * self.getCorrectedValue(pred(j),example);//
    v:= v + X[j] * FXData[i,j];//<<
   //hypothse implicite, Y est cod y1=1 et y2=2, c'est ce qui utilis dans la structure LSTVALUES
   //proba P(Y=2/X)
   v:= 1.0 / (1.0 + exp(-1.0 * v));
   if (v > 0.0) and (v < 1.0)
    then
     begin
      //log vraisemblance
      //1.0-1.0=0.0 et 2.0-1.0=1.0 au cas o on en douterait
      //en tous les cas, c'est nettement plus rapide qu'un IF
      tmpY:= ClassAttribute.cValue[example]-1.0;
      v:= tmpY*ln(v)+(1.0-tmpY)*ln(1.0-v);
      //somme de la vraisemblance
      s:= s+v;
     end;
  end;
 //pour la minimisation, on passe au ngatif
 result:= -2.0*s;
 EXCEPT
   on e: exception do
   begin
    TraceLog.WriteToLogFile(Format('[LOGREG] error LL function == %s',[e.Message]));
    result:= 0.0001*MATH.MaxDouble;
   end;
 END;
end;

procedure TCalcLogisticReg.RunOptimization;
begin
 //optimisation sur les variables centres
 //new -- R.R. -- 21/11/2007 -- mener dans le voisinage avant d'optimiser --> BFGS est trs rapide de toute manire
 //FErrCodeOptimization:= d6_optim.BFGS_MARQUARDT_OBJECT_PLUS(Log_Likelihood,directHessGrad,FCoef,0,descriptors.Count,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //new -- R.R. -- 02/12/2007 -- tir de OpenStat (LogRegUnit.cpp)
 //FErrCodeOptimization:= self.optimizationFast(FCoef,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //new -- R.R. -- 03/12/2007 -- matrice hessienne fixe pour initialiser la recherche
 //FErrCodeOptimization:= self.optimizationFixedHessian_plus_Marquardt(Log_Likelihood,directGrad,directHessGrad,FCoef,0,descriptors.Count,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //new -- R.R. -- 03/12/2007 -- retour sur le couple BFGS et MARQUARDT mais on donne en plus le gradient maintenant
 //FErrCodeOptimization:= d6_optim.BFGS_MARQUARDT_OBJECT_PLUS_PLUS(Log_Likelihood,directGrad,directHessGrad,FCoef,0,descriptors.Count,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //new -- R.R. -- 03/12/2007 -- BFGS + NEWTON-RAPHSON (au vu des perfs. de BFGS, on peut se contenter d'une moindre exigeance aprs)
 FErrCodeOptimization:= self.optimizationBFGSNewtonRaphson(Log_Likelihood,directGrad,directHessGrad,FCoef,0,descriptors.Count,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
end;

procedure TCalcLogisticReg.directHessGrad(Func: TFuncNVar; X: TVector;
  Lbound, Ubound: Integer; G: TVector; H: TMatrix);
var j,j1,j2,i: integer;
    rapport,value: extended;
    example: integer;
    y,z,z1,z2: extended;
begin

  for j1:= LBound to UBound do G[j1]:= 0.0;
  for j1:= LBound to UBound do for j2:= LBound to UBound do H[j1,j2]:= 0.0;

 //**************************
 //remplissage du tableau des probabilits individuelles Pk
 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   y:= ClassAttribute.cValue[example] - 1.0;
   //produit scalaire trs simple
   value:= X[0];
   for j:= 1 to self.Descriptors.Count do
    //value:= value + X[j] * self.getCorrectedValue(pred(j),example);
    value:= value + X[j] * FXData[i,j];//<<

   //fonction logistique
   value:= 1.0 / (1.0 + exp(-1.0 * value));

   //**************************
   //vecteur gradient -- eq. (4)
   for j:= LBound to UBound do
    begin
     //somme des produits
     //if (j > 0) then z:= self.getCorrectedValue(pred(j),example) else z:= 1.0;//
     z:= FXData[i,j];//<<
     G[j]:= G[j] + (y - value) * z;
    end;

   //*********************************
   //matrice hessienne -- eq. suivante
   for j1:= LBound to UBound do
    for j2:= j1 to UBound do
     begin
      //if (j1 > 0) then z1:= self.getCorrectedValue(pred(j1),example) else z1:= 1.0;//
      z1:= FXData[i,j1];//<<
      //if (j2 > 0) then z2:= self.getCorrectedValue(pred(j2),example) else z2:= 1.0;//
      z2:= FXData[i,j2];//<<
      rapport:= z1 * z2 * value * (1.0 - value);
      H[j1,j2]:= H[j1,j2] + rapport;
     end;

  end;//de for i

  //remplir le triangle infrieur de la matrice hessienne
  for j1:= LBound to pred(UBound) do
   for j2:= succ(j1) to UBound do
    H[j2,j1]:= H[j1,j2];

  //mettre le terme de rgularisation dans la diagonale
  for j:= LBound to UBound do
   H[j,j]:= H[j,j] - LAMBDA_FOR_REGULARIZATION; 

  (*
  //affichage de contrle
  TraceLog.WriteToLogFile('>>');
  for j1:= LBound to UBound do
   begin
    s:= 'HESS = [';
    for j2:= LBound to UBound do
     s:= s + format('%.4f,',[H[j1,j2]]);
    s:= copy(s,1,length(s)-1)+']';
    TraceLog.WriteToLogFile(s);
   end;
  TraceLog.WriteToLogFile('<<');
  *)
   
end;

procedure TCalcLogisticReg.computePrmCentrageReduction;
var j: integer;
begin
 setlength(FCentrage,FStatDes.Count+1);
 setLength(FReduction,FStatDes.Count+1);
 //pour la constante
 FCentrage[0]:= 0;
 FReduction[0]:= 1;
 for j:= 0 to pred(FStatDes.Count) do
  begin
   FCentrage[succ(j)]:= (FStatDes.Stat(j) as TCalcStatDesContinuous).Average;
   FReduction[succ(j)]:= (FStatDes.Stat(j) as TCalcStatDesContinuous).StdDev;
   //!\ si la varible est quasi-constante, le coef. associ est corrig
   if (FReduction[succ(j)] < MIN_VALUE_FOR_STDDEV)
    then FCoef[succ(j)]:= 0.0;
  end;
end;

function TCalcLogisticReg.getCorrectedValue(j, example: integer): float;
begin
 if (FReduction[succ(j)] < MIN_VALUE_FOR_STDDEV)
  then result:= 0
  else result:= (self.Descriptors.Attribute[j].cValue[example] - FCentrage[succ(j)]) / FReduction[succ(j)];
end;

procedure TCalcLogisticReg.createTempTabProba(examples: TExamples);
var i,j: integer;
begin
 //tableau des probas individuelles, on s'en sert comme un tampon pour viter les re-calculs
 //setLength(FTabPk,succ(examples.Size));
 setlength(FXData,succ(examples.Size),succ(Descriptors.Count));
 //tableau temporaire pour les donnes
 for i:= 1 to examples.Size do FXData[i,0]:= 1.0;
 for j:= 1 to descriptors.Count do
  for i:= 1 to examples.Size do
   FXData[i,j]:= self.getCorrectedValue(pred(j),examples.Number[i]);
end;

procedure TCalcLogisticReg.CorrectComputedCoefs;
var j: integer;
begin
 //correction des coefficients
 for j:= 1 to self.Descriptors.Count do
  begin
   //variable considre comme constante ?
   if (FReduction[j] < MIN_VALUE_FOR_STDDEV)
    then FCoef[j]:= 0.0
    else
     begin
      FCoef[j]:= FCoef[j] / FReduction[j];
      FCoef[0]:= FCoef[0] - FCoef[j] * FCentrage[j];
     end;
  end;
 //new -- 03/05/2006 -- corriger les coefs. pour qu'ils fournissent en priorit la proba P(Y=1/X) de la premire modalit donc...
 for j:= 0 to self.Descriptors.Count do
  FCoef[j]:= -1.0 * FCoef[j];
end;

procedure TCalcLogisticReg.directGrad(Func: TFuncNVar; X: TVector; Lbound,
  Ubound: Integer; G: TVector);
var j,i: integer;
    z,y,value: extended;
    example: integer;
begin

 for j:= LBound to UBound do G[j]:= 0.0;

 //**************************
 //calcul direct du gradient en une seule passe
 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   //produit scalaire trs simple
   value:= X[0];
   for j:= 1 to self.Descriptors.Count do
    //value:= value + X[j] * self.getCorrectedValue(pred(j),example);//
    value:= value + X[j] * FXData[i,j];//<<

   //fonction logistique
   value:= 1.0 / (1.0 + exp(-1.0 * value));
   y:= ClassAttribute.cValue[example] - 1.0;
   //gradient
   for j:= LBound to UBound do
    begin
     //if (j > 0) then z:= self.getCorrectedValue(pred(j),example) else z:= 1.0;//
     z:= FXData[i,j];//<<
     G[j]:= G[j] + (y - value) * z;
    end;
  end;
  
end;

procedure matvec_prod(A: TMatrix; v: TVector; LBound, UBound: integer; var u: TVector);
var i,j: integer;
    sum: float;
begin
 for i:= LBound to UBound do
  begin
   sum:= 0.0;
   for j:= LBound to UBound do
    sum:= sum + A[i,j]*v[j];
   u[i]:= sum;
  end;
end;

function isParamConv(OldX, X        : TVector;
                   Lbound, Ubound : Integer;
                   Tol            : Float) : Boolean;
var
  I : Integer;
  Conv : Boolean;
begin
  I := Lbound;
  Conv := True;
  repeat
    Conv := Conv and (Abs(X[I] - OldX[I]) < Max(Tol, Tol * Abs(OldX[I])));
    Inc(I);
  until (Conv = False) or (I > Ubound);
  result := Conv;
end;

function isValueConv(FOld, FNew: float): boolean;
begin
 if ((FOld - FNew) < TOLERANCE_DECREASE_LOGISTIC * FOld)
  then result:= true
  else result:= false;
end;

function TCalcLogisticReg.optimizationFixedHessian_plus_Marquardt(Func: TFuncNVar; Grad: TGradientObject; HessGrad : THessGradObject; X: TVector; Lbound, Ubound : Integer; MaxIter : Integer; Tol: Float; var F_min : Float; H_inv : TMatrix) : Integer;
var j1,j2,i:integer;
    //example: integer;
    v1,v2,sum: float;
    H: TMatrix;
    iter: integer;
    U,G,XOld: TVector;
const  FirstMaxIter: integer = 5;
       lambda: float = 0.0001;//terme de rgularisation
begin
 //*************************************************************************
 //commencer par une mthode o H est fix  l'avance (FIXED-HESSIAN METHOD)
 dimMatrix(H,Descriptors.Count,Descriptors.Count);
 //calculer le X'X dans H_inv
 for j1:= LBound to UBound do
  for j2:= LBound to UBound do
   begin
    sum:= 0.0;
    for i:= 1 to FCurExamples.Size do
     begin
      //example:= FCurExamples.Number[i];
      //if (j1 = LBound) then v1:= 1.0 else v1:= self.getCorrectedValue(pred(j1),example);//
      v1:= FXData[i,j1];//<<
      //if (j2 = LBound) then v2:= 1.0 else v2:= self.getCorrectedValue(pred(j2),example);//
      v2:= FXData[i,j2];//<<
      sum:= sum + v1 * v2;
     end;
    H[j1,j2]:= -0.25*sum;
    if (j1 = j2) then H[j1,j2]:= H[j1,j2] - lambda;//appliquer la rgularisation
   end;
 //inverser
 invMat(H,LBound,UBound,H_Inv);
 //prparer les vecteurs intermdiaires
 dimVector(U,UBound);
 dimvector(G,UBound);
 dimVector(XOld,UBound);
 //commencer les itrations
 iter:= 0;
 REPEAT
  inc(iter);
  Grad(Func,X,LBound,UBound,G);
  matvec_prod(H_Inv,G,LBound,UBound,U);
  CopyVector(XOld,X,LBound,UBound);
  for j1:= LBound to UBound do X[j1]:= XOld[j1] - U[j1];
 UNTIL (iter >= FirstMaxIter) or (isParamConv(XOld,X,LBound,UBound,Tol) = true);
 TraceLog.WriteToLogFile(format('[FIXED-HESSIAN] convergence on %d iterations (max authorized = %d)',[iter,FirstMaxIter]));
 //enchaner avec un Levenberg-Marquardt c'est tout aussi bien... (surtout dans les espaces sparses !!!)
 result:= MARQUARDT_OBJECT(Func,HessGrad,X,LBound,UBound,MaxIter,Tol,F_Min,H_INV);
 //vider les matrices et vecteurs locaux
 H:= nil; U:= nil; G:= nil; XOld:= nil;
end;

function TCalcLogisticReg.optimizationFHNewtonRaphson(Func: TFuncNVar;
  Grad: TGradientObject; HessGrad: THessGradObject; X: TVector; Lbound,
  Ubound, MaxIter: Integer; Tol: Float; var F_min: Float;
  H_inv: TMatrix): Integer;
var j1,j2,i:integer;
    //example: integer;
    v1,v2,sum: float;
    H: TMatrix;
    iter: integer;
    U,G,XOld,DeltaX: TVector;
    FOld: float;
    Det: float;
const  FirstMaxIter: integer = 5;
       lambda: float = 0.0001;//terme de rgularisation
begin
 //*************************************************************************
 //commencer par une mthode o H est fix  l'avance (FIXED-HESSIAN METHOD)
 dimMatrix(H,Descriptors.Count,Descriptors.Count);
 //calculer le X'X dans H_inv
 for j1:= LBound to UBound do
  for j2:= LBound to UBound do
   begin
    sum:= 0.0;
    for i:= 1 to FCurExamples.Size do
     begin
      //example:= FCurExamples.Number[i];
      //if (j1 = LBound) then v1:= 1.0 else v1:= self.getCorrectedValue(pred(j1),example);//
      v1:= FXData[i,j1];//<<
      //if (j2 = LBound) then v2:= 1.0 else v2:= self.getCorrectedValue(pred(j2),example);//
      v2:= FXData[i,j2];//<<
      sum:= sum + v1 * v2;
     end;
    H[j1,j2]:= -0.25*sum;
    if (j1 = j2) then H[j1,j2]:= H[j1,j2] - lambda;//appliquer la rgularisation
   end;
 //inverser
 invMat(H,LBound,UBound,H_Inv);
 //prparer les vecteurs intermdiaires
 dimVector(U,UBound);
 dimvector(G,UBound);
 dimVector(XOld,UBound);
 //commencer les itrations
 iter:= 0;
 REPEAT
  inc(iter);
  Grad(Func,X,LBound,UBound,G);
  matvec_prod(H_Inv,G,LBound,UBound,U);
  CopyVector(XOld,X,LBound,UBound);
  for j1:= LBound to UBound do X[j1]:= XOld[j1] - U[j1];
 UNTIL (iter >= FirstMaxIter) or (isParamConv(XOld,X,LBound,UBound,Tol) = true);
 TraceLog.WriteToLogFile(format('[FIXED-HESSIAN] convergence on %d iterations (max authorized = %d)',[iter,FirstMaxIter]));
 //********************************
 //enchaner avec un Newton-Raphson
 dimVector(DeltaX,UBound);
 F_Min:= Func(X);
 iter:= 0;
 REPEAT
  inc(iter);
  FOld:= F_Min;
  HessGrad(Func,X,LBound,UBound,G,H);
  for j1:= LBound to UBound do H[j1,j1]:= H[j1,j1] - lambda;//rgularisation
  CopyVector(XOld,X,LBound,UBound);
  GaussJordan(H, G, Lbound, Ubound, H_inv, DeltaX, Det);
  for j1:= LBound to UBound do X[j1]:= XOld[j1] + DeltaX[j1];
  F_Min:= Func(X);
 UNTIL (iter >= maxIter) or (isParamConv(XOld,X,LBound,UBound,Tol) = true) or (isValueConv(FOld,F_min) = true);
 TraceLog.WriteToLogFile(format('[NEWTON-RAPHSON] convergence on %d iterations (max authorized = %d)',[iter,MaxIter]));
 //vider les matrices et vecteurs locaux
 H:= nil; U:= nil; G:= nil; XOld:= nil; DeltaX:= nil;
 //renvoyer le rsultat
 if (iter >= maxIter) then result:= OPT_NON_CONV else result:= OPT_OK;
end;

function TCalcLogisticReg.optimizationFast(X: TVector; MaxIter: integer; Tol: float; var F_Min: float; H_inv: TMatrix): integer;
var LLp, LL: extended;
    iter: integer;
    i,j,k: integer;
    example: integer;
    s,v,y,proba,q,xij,xijprim: extended;
    Arr: array of extended; //dans cette structure est la hessienne ET le gradient
    nC, nR, nP, nP1: integer;
    indx,indx2,indx3: integer;

const lambda: float = 0.0001;//terme de rgularisation     

    function ix(j,k,nCols: integer): integer;
    begin
     result:= j * nCols + k;
    end;
begin
 //init...
 nR:= self.Descriptors.Count;
 nC:= FCurExamples.Size;
 nP:= nR + 1;
 nP1:= nP + 1;
 setLength(Arr,(nP + 1) * (nP1 + 1));
 //
 LL:= Math.MaxSingle;
 iter:= 0;

 TRY

 REPEAT
  inc(iter);
  LLp:= LL;
  LL:= 0.0;
  //
  for j:= 0 to nR do
   for k:= j to nR + 1 do
    begin
     indx:= ix(j,k,nR+2);
     Arr[indx]:= 0.0;
    end;
  //calcul de la vraisemblance et de la matrice hessienne
  for i:= 1 to nC do
   begin
   example:= FCurExamples.Number[i];
   y:= ClassAttribute.cValue[example]-1.0;
   
   //pour le calcul de la vraisemblance
   v:= X[0];
   for j:= 1 to nR do
    //v:= v + X[j] * self.getCorrectedValue(pred(j),example);//
    v:= v + X[j] * FXData[i,j];//<<

   if (v > 708) //!\protection
    then proba:= 1.0
    else if (v < -708)
     then proba:= 0.0
     else proba:= 1.0 / (1.0 + exp(-1.0 * v));

   q:= proba * (1.0 - proba);
   if (proba > 0.0) and (proba < 1.0)//petite protection contre les bugs...
    then LL:= LL - 2.0 * y * ln(proba) - 2.0 * (1.0 - y) * ln(1.0 - proba);

   //matrice hessienne -- eq. suivante
   for j:= 0 to nR do
    begin
     //if (j > 0) then xij:= self.getCorrectedValue(pred(j),example) else xij:= 1.0;//
     xij:= FXData[i,j];//<<
     indx:= ix(j,nR + 1,nR + 2);
     //le gradient est ici ???
     Arr[indx]:= Arr[indx] + xij * (y * (1.0 - proba) + (1.0 - y) * (-1.0 * proba));
     for k:= j to nR do
      begin
       //if (k > 0) then xijprim:= self.getCorrectedValue(pred(k),example) else xijprim:= 1.0;//
       xijprim:= FXData[i,k];//<<
       indx:= ix(j,k,nR + 2);
       //la matrice hessienne
       Arr[indx]:= Arr[indx] + xij * xijprim * q;
      end;
    end;
  end;

  //triangulaire inf. de la matrice hessienne
  for j:= 1 to nR do
   for k:= 0 to j-1 do
    Arr[ix(j,k,nR+2)]:= Arr[ix(k,j,nR+2)];

  //R.R. -- ajouter un terme de rgularisation ?
  for i:= 0 to nR do
   begin
    indx:= ix(i,i,nR+2);
    Arr[indx]:= Arr[indx] - lambda;
   end;

  //*******************
  //les mises  jour...
  for i:= 0 to nR do
   begin
    indx:= ix(i,i,nR+2);
    s:= Arr[indx];
    Arr[indx]:= 1.0;
    //
    for k:= 0 to nR+1 do
     begin
      indx:= ix(i,k,nR+2);
      if (s <> 0.0) //!\protection
       then Arr[indx]:= Arr[indx] / s;
     end;
    //
    for j:= 0 to nR do
     begin
      if (i <> j)
       then
        begin
         indx:= ix(j,i,nR+2);
         s:= Arr[indx];
         Arr[indx]:= 0.0;
         //
         for k:= 0 to nR+1 do
          begin
           indx2:= ix(j,k,nR+2);
           indx3:= ix(i,k,nR+2);
           Arr[indx2]:= Arr[indx2] - s * Arr[indx3];
          end;
        end;
     end;
   end;
  //mj des coefficients
  for j:= 0 to nR do
   begin
    indx:= ix(j,nR+1,nR+2);
    X[j]:= X[j] + Arr[indx];
   end;
 UNTIL (abs(LLp-LL) < 0.00001) or (iter > MaxIter);

 EXCEPT
  on e: Exception do
   begin
    TraceLog.WriteToLogFile(format('[FAST] exception = %s during optimization',[e.message]));
    Raise e;
   end;
 END;

 TraceLog.WriteToLogFile(format('[FAST] convergence in %d iterations (max = %d)',[iter,MaxIter]));
 //rcupration des infos
 F_Min:= LL;//qui tait optimis
 H_inv[0,0]:= 1.0;//on ne s'en sert pas de toute manire
 for j:= 1 to nR do//uniquement la diagonale de la matrice hessienne est renvoye correctement
  begin
   indx:= ix(j,j,nP+1);
   H_Inv[j,j]:= Arr[indx];
  end;
 Arr:= nil;
 //and then...
 if (iter >= maxIter) then result:= OPT_NON_CONV else result:= OPT_OK;
end;

procedure TCalcLogisticReg.destroyTempTabProba;
begin
 //adios vite fait, bon boulot...
 //finalize(FTabPk);
 finalize(FXData);
end;

function TCalcLogisticReg.optimizationBFGSNewtonRaphson(Func: TFuncNVar;
  Grad: TGradientObject; HessGrad: THessGradObject; X: TVector; Lbound,
  Ubound, MaxIter: Integer; Tol: Float; var F_min: Float;
  H_inv: TMatrix): Integer;

const  FirstMaxIter: integer = 5;
var err: integer;
    plantage: boolean;
    FirstX: TVector;
begin
 err:= OPT_OK;
 dimvector(FirstX,UBound);
 d6_optim.BFGS_OBJECT(Func,Grad,X,LBound,UBound,FirstMaxIter,Tol,F_Min,H_Inv);
 CopyVector(FirstX,X,LBound,Ubound);
 try
 plantage:= false;
 err:= d6_optim.Marquardt_Object(Func,HessGrad,X,LBound,UBound,MaxIter,Tol,F_Min,H_Inv);
 except
 plantage:= true;
 TraceLog.WriteToLogFile('[WARNING] LEV-MQRDT a planted, passer la main to NEWTON');
 end;
 //si plantage, on se tourne vers une mthode plus robuste...
 if (plantage = true)
  then
   begin
    err:= OPT_OK;
    CopyVector(X,FirstX,LBound,UBound);
    TRY
    plantage:= false;
    err:= self.optimizationNewtonRaphson(Func,Grad,HessGrad,X,LBound,UBound,MaxIter,Tol,F_Min,H_Inv);
    EXCEPT
    plantage:= true;
    CopyVector(X,FirstX,LBound,UBound);
    TraceLog.WriteToLogFile('[WARNING] NEWTON a planted, retour sur BFGS');
    END;
   end;
 if (plantage = true) then err:= d6_optim.BFGS_OBJECT(Func,Grad,X,LBound,UBound,MaxIter,Tol,F_Min,H_Inv);//on reste l dessus  donf. alors... 
 FirstX:= nil;
 //advienne que pourra
 result:= err;
end;

function TCalcLogisticReg.optimizationNewtonRaphson(Func: TFuncNVar;
  Grad: TGradientObject; HessGrad: THessGradObject; X: TVector; Lbound,
  Ubound, MaxIter: Integer; Tol: Float; var F_min: Float;
  H_inv: TMatrix): Integer;
var
    j1: integer;
    H: TMatrix;
    iter: integer;
    G,XOld,DeltaX: TVector;
    FOld: float;
    Det: float;
const   lambda: float = 1e-4;//terme de rgularisation      
begin
 //********************************
 //enchaner avec un Newton-Raphson
 dimMatrix(H,UBound,UBound);
 dimVector(DeltaX,UBound);
 dimVector(XOld,UBound);
 dimvector(G,UBound);
 F_Min:= Func(X);
 iter:= 0;
 REPEAT
  inc(iter);
  FOld:= F_Min;
  HessGrad(Func,X,LBound,UBound,G,H);
  for j1:= LBound to UBound do H[j1,j1]:= H[j1,j1] - lambda;//rgularisation
  CopyVector(XOld,X,LBound,UBound);
  GaussJordan(H, G, Lbound, Ubound, H_inv, DeltaX, Det);
  for j1:= LBound to UBound do X[j1]:= XOld[j1] + DeltaX[j1];
  F_Min:= Func(X);
 UNTIL (iter >= maxIter) or (isParamConv(XOld,X,LBound,UBound,Tol) = true) or (isValueConv(FOld,F_min) = true);
 TraceLog.WriteToLogFile(format('[NEWTON-RAPHSON] convergence on %d iterations (max authorized = %d)',[iter,MaxIter]));
 //vider les matrices et vecteurs locaux
 H:= nil; G:= nil; XOld:= nil; DeltaX:= nil;
 //renvoyer le rsultat
 if (iter >= maxIter) then result:= OPT_NON_CONV else result:= OPT_OK;
end;

function TCalcLogisticReg.computeDevianceInterceptOnly: double;
var tmpVec: TVector;
    a,v: float;
begin
 v:= StatClassAtt.TabFreq.Frequence[2];
 a:= ln(v/(1.0-v));
 //calculer la vraisemblance de ce modle
 dimVector(tmpVec,Descriptors.Count);
 tmpVec[0]:= a;
 result:= self.Log_Likelihood(tmpVec);
 finalize(tmpVec);
end;

procedure TCalcLogisticReg.computeSpecificStats;
var i: integer;
begin
 //l'cart-type
 // >> on peut le faire sans recalcul, mais attention souci avec l'cart-type de la constante
 for i:= 0 to descriptors.Count do
  if (FHInv[i][i]<=0.0) or (FReduction[i] < MIN_VALUE_FOR_STDDEV)
   then FStdDev[i]:= 0.0
   else FStdDev[i]:= SQRT(1.0*FHInv[i][i]) / FReduction[i];
 //Wald et p-value associe
 for i:= 0 to length(FCoef)-1 do
  begin
   FWald[i]:= 0.0;
   FPValueWald[i]:= 1.0;
   if (FStdDev[i] > 0.0)
    then
     begin
      FWald[i]:= SQR(FCoef[i]/FStdDev[i]);
      if (FWald[i] > 0.0) then FPValueWald[i]:= fmath.PKHI2(1,FWald[i]); 
     end;
  end;
end;

procedure TCalcLogisticReg.computeStdStats;
var n,LL_0,LL_B: double;
begin
 n:= 1.0*FCurExamples.Size;
 FChi2_ddl:= NbParam-(NbClasses-1);
 //le chi-2
 FChi2:= FNullLogLikelihood - FMinLogLikelihood;
 //proba associe
 FPrChi2:= fmath.PKHI2(FChi2_ddl,FChi2);
 //new -- 19/08/2007 -- calcul des Pseudo-R2 (voir -- ftp://ftp.spss.com/pub/spss/statistics/spss/algorithms/logregre.pdf)
 LL_0:= FNullLogLikelihood/(-2.0);
 LL_B:= FMinLogLikelihood/(-2.0);
 FR2M:= 1.0 - LL_B/LL_0;
 FR2CS:= 1.0 - exp(-(2.0/n)*(LL_B-LL_0));
 FR2N:= FR2CS/(1.0-exp((2.0/n)*LL_0));
 //new -- 05/12/2007 -- AIC et cie.
 AICIntercept:= FNullLogLikeLihood + 2.0 * (-1.0 + NbClasses);
 AICModel:= FMinLogLikelihood + 2.0 * NbParam;
 BICIntercept:= FNullLogLikeLihood + LN(1.0*FCurExamples.Size) * (-1.0 + NbClasses);
 BICModel:= FMinLogLikelihood+LN(1.0*FCurExamples.Size) * NbParam;
end;

function TCalcLogisticReg.getHTMLCoefficients: string;
var s: string;
    i: integer;
begin
//les coefficients
 s:= '<P><H3>Attributes in the equation</H3><BR>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH width=120>Attribute</TH><TH width=80>Coef.</TH><TH width=80>Std-dev</TH><TH width=80>Wald</TH><TH width=80>Signif</TH></TR>';
 for i:= 0 to descriptors.count do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY;
   if (i=0)
    then s:= s+'<TD>constant</TD>'
    else s:= s+'<TD>'+descriptors.Attribute[pred(i)].Name+'</TD>';
   //envoyer la sauce
   if (i = 0)
    then s:= s + format('<TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD>',[FCoef[i]])
    else
     s:= s+format('<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD>'+
                  '<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>'+
                  '<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>'+
                  '<TD align=right %s>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>',
                  [FCoef[i],FStdDev[i],FWald[i],colorCodeForPValue(FPValueWald[i]),FPValueWald[i]]);
   s:= s+'</TR>';
  end;
 s:= s+'</table>';
 result:= s;
end;

function TCalcLogisticReg.getHTMLGlobalResult: string;
var s: string;
begin
 s:= '<P><H3>Adjustement quality</H3><BR>';
 // complter la premire partie des rsultats
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Predicted attribute</TD><TD colspan=2 align=right>%s</TD></TR>',[ClassAttribute.Name]);
 s:= s+self.getHTMLRefValue();
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Number of examples</TD><TD colspan=2 align=right>%d</TD></TR>',[FCurExamples.Size]);
 s:= s+HTML_TABLE_COLOR_DATA_GREEN+'<TD align="center" colspan=3>Model Fit Statistics</TD></TR>';
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TD>Criterion</TD><TD align=center width=70>Intercept</TD><TD align=center width=70>Model</TD></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>AIC</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD>',[AICIntercept,AICModel]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>SC</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD>',[BICIntercept,BICModel]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>-2LL</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD>',[FNullLogLikelihood,FMinLogLikelihood]);
 s:= s+HTML_TABLE_COLOR_DATA_GREEN+'<TD align="center" colspan=3>Model Chi test (LR)</TD></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>Chi-2</TD>'+format('<TD colspan=2 align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD></TR>',[FChi2]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>d.f.</TD><TD colspan=2 align=right>%d</TD></TR>',[FChi2_ddl]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>P(>Chi-2)</TD>'+format('<TD colspan=2 align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD></TR>',[FPrChi2]);
 s:= s+HTML_TABLE_COLOR_DATA_GREEN+'<TD align="center" colspan=3>R-like</TD></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>McFadden''s R</TD>'+format('<TD colspan=2 align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD></TR>',[FR2M]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>Cox and Snell''s R</TD>'+format('<TD colspan=2 align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD></TR>',[FR2CS]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD>Nagelkerke''s R</TD>'+format('<TD colspan=2 align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD></TR>',[FR2N]);
 s:= s+'</table>';
 result:= s;
end;

function TCalcLogisticReg.getHTMLPostResults: string;
var s: string;
    i: integer;
begin
 s:= '';
 //odds ratio et intervalle de confiance  5%
 s:= s+'<P><H3>Odds ratios and 95% confidence intervals</H3><BR>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH width=120>Attribute</TH><TH width=80>Coef.</TH><TH width=80>Low</TH><TH width=80>High</TH></TR>';
 for i:= 1 to descriptors.count do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY;
   s:= s+'<TD>'+descriptors.Attribute[pred(i)].Name+'</TD>';
   s:= s+format('<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>'+
                '<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>'+
                '<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</TD>',
                [exp(FCoef[i]),exp(FCoef[i]-fmath.InvNorm(0.975)*FStdDev[i]),exp(FCoef[i]+fmath.InvNorm(0.975)*FStdDev[i])]);
  end;
 s:= s + '</table>';
 result:= s;
end;


function TCalcLogisticReg.getHTMLRefValue: string;
begin
 result:= HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Positive value</TD><TD colspan=2 align=right>%s</TD></TR>',[ClassAttribute.LstValues.getDescription(1)]);
end;

initialization
 Classes.RegisterClass(TMLGCompLogisticReg);
end.
