(******************************************************************)
(* UCompDemonstration.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Module de dmonstration : insertion d'un nouveau composant)
@author(Ricco Rakotomalala)
@created(21/04/2008)

Cette unit montre comment insrer un nouveau composant dans TANAGRA.
Nous nous contentons de calculer la moyenne des variables INPUT. Nous
crons des classes hritires (4 principalement) des classes pr-existantes
pour cela. Quelques mthodes doivent imprativement tre surcharges. 
}

unit UCompDemonstration;

interface

USES
  //units std delphi
  Classes, IniFiles, Forms, Sysutils,
  //gestion des composants
  UCompDefinition, UOperatorDefinition,
  //gestion de l'accs aux donnes
  UCompManageDataset, UDatasetDefinition, UDatasetImplementation,
  UDatasetExamples;

TYPE
  //***********************************************************************
  //** gnrateur de composant, son rle est de crer une instance de composant
  //** qui est insr dans le diagramme de traitements
  //***********************************************************************
  TMLGenCompDemo = class(TMLGenComp)
  protected
  //spcifier dans quel onglet de la palette doit apparatre l'icne
  procedure   genCompInitializations(); override;
  public
  //spcifier quelle est la classe du composant  gnrer sur le diagramme
  function    getClassMLComponent: TClassMLComponent; override;
  end;

  //***********************************************************************
  //** classe du composant graphique insr dans le diagramme
  //***********************************************************************
  TMLCompDemo = class(TMLCompLocalData)
  protected
  //spcifier l'oprateur de calcul
  function    getClassOperator: TClassOperator; override;
  end;

  //***********************************************************************
  //** classe oprateur de calcul : prend en charge le calcul
  //** et l'affichage des rsultats
  //***********************************************************************
  TOpDemo = class(TOpLocalData)
  private
  //tableau des moyennes INPUT
  FTabAvg: array of double;
  protected
  //spcifier la classe de paramtrage -- surcharge obligatoire
  function    getClassParameter: TClassOperatorParameter; override;
  //contrle de cohrence des attributs avant les calculs
  function    checkAttributes(): boolean; override;
  //lancer effectivement les calculs -- surcharge obligatoire
  function    coreExecute(): boolean; override;
  public
  //libration de la mmoire alloue
  destructor  destroy(); override;
  //affichage des rsultats au format HTML -- surcharge obligatoire
  function    getHTMLResultsSummary(): string; override;
  end;

  //***********************************************************************
  //** classe paramtrage de l'oprateur de calcul
  //** gre la bote de paramtrage et l'E/S des paramtres
  //***********************************************************************
  TOpPrmDemo = class(TOperatorParameter)
  protected
  //cration de la bote de dialogue de paramtrage -- surcharge obligatoire
  //(doit renvoyer "nil" si pas de bote  gnrer)
  function    CreateDlgParameters(): TForm; override;
  //spcifier les paramtres par dfaut de la mthode -- surcharge obligatoire
  procedure   SetDefaultParameters(); override;
  end;



implementation

{ TMLGenCompDemo }

procedure TMLGenCompDemo.genCompInitializations;
begin
 //onglet STATISTICS
 FMLComp:= mlcDescriptiveStat;
end;

function TMLGenCompDemo.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompDemo;
end;

{ TMLCompDemo }

function TMLCompDemo.getClassOperator: TClassOperator;
begin
 result:= TOpDemo; 
end;

{ TOpDemo }

function TOpDemo.checkAttributes: boolean;
var ok: boolean;
begin
 //premier test, pas de TARGET
 ok:= (self.WorkData.LstAtts[asTarget].Count = 0);
 //second test, au moins 1 INPUT et tous continus
 ok:= ok and
      ((self.WorkData.LstAtts[asInput].Count > 0) and
        self.WorkData.LstAtts[asInput].isAllCategory(caContinue));
 //renvoyer la rponse
 result:= ok;
end;

function TOpDemo.coreExecute: boolean;
var j,i: integer;
    nbAtt,nbExamples: integer;
    avg: double;
    att: TAttribute;
begin
 TRY
 //prparer le tableau de rsultats
 nbAtt:= self.WorkData.LstAtts[asInput].Count;
 setLength(FTabAvg,nbAtt);
 //nb. d'individus
 nbExamples:= self.WorkData.Examples.Size;
 //pour chaque variable, calculer la moyenne
 for j:= 0 to nbAtt-1 do
  begin
   //se brancher sur la variable
   att:= self.WorkData.LstAtts[asInput].Attribute[j];
   //calcul de la somme
   avg:= 0.0;
   for i:= 1 to nbExamples do
    avg:= avg + att.cValue[self.WorkData.Examples.Number[i]];
   //moyenne
   FTabAvg[j]:= avg / (1.0 * nbExamples); 
  end;
 //pas de soucis
 result:= true;
 EXCEPT
 //il y a eu des soucis
 result:= false;
 END;
end;

destructor TOpDemo.destroy;
begin
 //dsallouer le tableau
 setLength(FTabAvg,0);
 inherited;
end;

function TOpDemo.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmDemo;
end;

function TOpDemo.getHTMLResultsSummary: string;
var s: string;
    j: integer;
begin
 s:= '<table border = 1><tr><th>Attribute</th><th>Average</th></tr>';
 for j:= 0 to self.WorkData.LstAtts[asInput].Count-1 do
  begin
   s:= s + '<tr>';
   s:= s + format('<td align=left>%s</td><td align=right>%.4f</td>',
           [self.WorkData.LstAtts[asInput].Attribute[j].Name,FTabAvg[j]]);
   s:= s + '</tr>';
  end;
 s:= s + '</table>';
 //
 result:= s;
end;

{ TOpPrmDemo }

function TOpPrmDemo.CreateDlgParameters: TForm;
begin
 //pas de bote de dialogue de paramtrage
 result:= nil;
end;

procedure TOpPrmDemo.SetDefaultParameters;
begin
 //>> pas de paramtres
end;

initialization
 RegisterClass(TMLGenCompDemo);
end.
