(***************************************************************)
(* UCompVarClusHCA.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(***************************************************************)

{
@abstract(Composant de base pour le clustering de variables)
@author(Ricco)
@created(19/02/2007)

Clustering sur les variables. Mthode des CAH.

23/02/2007 -- attention la distance d'agrgation -- Nakache, page.231
}

unit UCompVarClusHCA;

interface

USES
    Classes,
    UCompVarClusBase, UCalcVarClusBase,
    Forms,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompClusteringDefinition,
    UDatasetDefinition,
    UDatasetExamples,
    UCalcStatDes,
    contnrs;

TYPE
    //gnrateur
    TMLGenVarClusHCA = class(TMLGenVarClus)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLVarClusHCA = class(TMLCompVarClusBase)
    protected
    function    getClassOperator: TClassOperator; override;
    function    getGenericAttName(): string; override;
    end;

    //oprateur
    TOpVarClusHCA = class(TOpVarClusBase)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    getClassCalc(): TClassCalcVarClus; override;
    function    getNewForm(): TForm; override;
    end;

    //paramtrage
    TPrmOpVarClusHCA = class(TOpPrmVarClus)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //Mode de dtermination de la coupure optimale : 0->automatique, 1->manuel, df. 0
    ModeBestSelection: integer;
    //si choix manuel, nombre de clusters, df. 4
    NbClusters: integer;
    //nombre de clusters de dpart, df. 20 -- permet de rduire considrablement le temps de calcul
    StartingNbClusters: integer;
    //I-O
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    //HTML output
    function    getHTMLParameters(): string; override;
    end;

    //dclaration forward
    TCalcHCA = class;

    {Un sommet de CAH}
    TNodeHCA = class(TObject)
    public
    {oprateur de calcul}
    FCalc: TCalcHCA;
    {pre}
    FDad: TNodeHCA;
    {enfants gauche et droite}
    FChildL,FChildR: TNodeHCA;
    {les informations locales, propritaire}
    FClus: TVarCluster;
    {numro de sommet pour suivre les fusions}
    FNumNode: integer;
    {Distance d'agrgation}
    FDist: double;
    {coordonne sur l'axe des X, pour le dessin du dendrogramme}
    FXPos: double;
    {coordonne sur l'axe des ordonnes, pour le dessin du dendrogramme}
    FYPos: double;
    {indique si le noeud est une solution}
    FIsCluster: boolean;
    public
    {construire un noeud  partir d'un cluster}
    constructor  Create(prmCalc: TCalcHCA; clus: TVarCluster; dist: double);
    {dtruire}
    destructor   Destroy(); override;
    {tester si le noeud est une feuille}
    function     isLeaf(): boolean;
    {recursif pour rcuprer les labels sur les feuilles d'un noeud}
    function     getLabel(): string;
    //proprits
    property     XPos: double read FXPos;
    property     YPos: double read FYPos;
    property     ChildL: TNodeHCA read FChildL;
    property     ChildR: TNodeHCA read FChildR;
    property     isCluster: boolean read FIsCluster;
    end;

    {Structure complte de CAH}
    TSetNodeHCA = class(TObject)
    private
    {Calculateur HAC}
    FCalc: TCalcHCA;
    {racine de l'arbre}
    FRootNode: TNodeHCA;
    {tous les noeuds - liste propritaire}
    FLstAllNodes: TObjectList;
    {les feuilles de l'arbre - liste non propritaire}
    FLstLeaves: TObjectList;
    {les noeuds  fusionner - les noeuds orphelins - liste non propritaire}
    FLstMergeNodes: TObjectList;
    {Indicateur global du dernier numro de noeud ajout}
    //FLastNumNode: integer;
    {Tableau de suivi des fusions ascendantes}
    FLstNewNodes: TObjectList;
    {Tableau des sommets candidats  la fusion -- non propritaire}
    FCandidates: TObjectList;
    {Hauteur dans le dendrogramme,  mettre  jour graduellement}
    FGlobalYPos: double;
    protected
    {ajouter un noeud dans la liste initiale}
    procedure   AddLeaf(node: TNodeHCA);
    public
    {cration de la structure}
    constructor Create(prmCalc: TCalcHCA);
    {dtruire la structure}
    destructor  Destroy(); override;
    {initialiser les noeuds}
    procedure   InitializeNodes(prmExamples: TExamples);
    {fusionner deux noeuds et l'ajouter dans la structure}
    procedure   MergeNodes(examples: TExamples);
    {calculer les coordonnes des X pour les feuilles}
    procedure   setLeavesCoords(node: TNodeHCA; var curCoord: double);
    {calculer les coordonnes des noeuds intrieurs de l'arbre}
    procedure   setNodesCoords(node: TNodeHCA);
    {rcuprer le nombre de noeuds}
    function    getCountNodes(): integer;
    {rcuprer le noeud numro index}
    function    getNode(index: integer): TNodeHCA;
    //proprits
    property    RootNode: TNodeHCA read FRootNode;
    end;


    //oprateur de calcul
    TCalcHCA = class(TCalcVarClus)
    private
    FPrm: TPrmOpVarClusHCA; 
    {inertie totale}
    FTSS: double;
    {Indice du palier max.}
    FIdMaxPalier: integer;
    {les noeuds de la solution adopte}
    FLstNodesSolution: TObjectList;
    {liste des fusions possibles - propritaire}
    FLstComputedFusions: TObjectList;
    {structure d'arbre}
    FSetNode:TSetNodeHCA;
    {Tableau des variances expliques}
    FTabVarExplained: TTabVarExplained;
    {tableau des diffrences premires, du cosinus, de l'angle et du repre}
    FTabPalier: TTabPalier;
    {Seuil pour les moving average intressants}
    FSeuilMV: double;
    {recalcul des variances expliques pour chaque squence}
    procedure   refreshVarExplained();
    {initialiser les noeuds}
    procedure   InitializeNodes(prmExamples: TExamples);
    {calculer les coordonnes pour le dessin du dendrogramme}
    procedure   computeNodeCoords();
    protected
    procedure   detectPalier();
    procedure   buildBestNodes(examples: TExamples);
    procedure   computeClusters(prmExamples: TExamples); override;
    //affichage HTML
    function    getHTMLDetailedResults(): string; override;
    public
    constructor create(prmInputs: TLstAttributes; param: TOperatorParameter); override;
    destructor  Destroy(); override;
    property    SetNodeHCA: TSetNodeHCA read FSetNode;
    end;

implementation

uses
    Sysutils, StrUtils, Math,
    UCompVarClusKMeans, ULogFile, UFrmOpVarClusHCA, UDlgOpPrmVarClusHCA,
    UConstConfiguration;

{ TMLGenVarClusHCA }

function TMLGenVarClusHCA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLVarClusHCA;
end;

{ TMLVarClusHCA }

function TMLVarClusHCA.getClassOperator: TClassOperator;
begin
 result:= TOpVarClusHCA;
end;

function TMLVarClusHCA.getGenericAttName: string;
begin
 result:= 'VCHca';
end;

{ TOpVarClusHCA }

function TOpVarClusHCA.getClassCalc: TClassCalcVarClus;
begin
 result:= TCalcHCA;
end;

function TOpVarClusHCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpVarClusHCA;
end;

function TOpVarClusHCA.getNewForm: TForm;
begin
 result:= TFrmOpViewVarClusHCA.CreateFromOperator(self);
end;

{ TPrmOpVarClusHCA }

function TPrmOpVarClusHCA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmVarClusHCA.CreateFromOpPrm(self);
end;

function TPrmOpVarClusHCA.getHTMLParameters: string;
var s: string;
begin
 s:= inherited getHTMLParameters();
 s:= s + format('User-defined number of clusters : %d<br>',[ModeBestSelection]);
 s:= s + format('Number of clusters, if user-defined : %d<br>',[NbClusters]);
 s:= s + format('Initial number of clusters in the search algorithm : %d<br>',[StartingNbClusters]);
 //
 result:= s;
end;

procedure TPrmOpVarClusHCA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 ModeBestSelection:= prmINI.ReadInteger(prmSection,'cluster_selection',ModeBestSelection);
 NbClusters:= prmINI.ReadInteger(prmSection,'user_defined_nb_cluster',NbClusters);
 StartingNbClusters:= prmINI.ReadInteger(prmSection,'initial_cluster',StartingNbClusters);
end;

procedure TPrmOpVarClusHCA.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.ReadBuffer(NbClusters,sizeof(NbClusters));
 prmStream.ReadBuffer(StartingNbClusters,sizeof(StartingNbClusters));
end;

procedure TPrmOpVarClusHCA.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'cluster_selection',ModeBestSelection);
 prmINI.WriteInteger(prmSection,'user_defined_nb_cluster',NbClusters);
 prmINI.WriteInteger(prmSection,'initial_cluster',StartingNbClusters);
end;

procedure TPrmOpVarClusHCA.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.WriteBuffer(NbClusters,sizeof(NbClusters));
 prmStream.WriteBuffer(StartingNbClusters,sizeof(StartingNbClusters));
end;

procedure TPrmOpVarClusHCA.SetDefaultParameters;
begin
 inherited;
 //Mode de dtermination de la coupure optimale : 0->automatique, 1->manuel, df. 0
 ModeBestSelection:= 0;
 //si choix manuel, nombre de clusters, df. 4
 NbClusters:= 4;
 //nombre de clusters de dpart, df. 20 -- permet de rduire considrablement le temps de calcul
 StartingNbClusters:= 20;
end;

{ TNodeHCA }

constructor TNodeHCA.Create(prmCalc: TCalcHCA; clus: TVarCluster; dist: double);
begin
 inherited create();
 FCalc:= prmCalc;
 FClus:= clus;
 FDad:= NIL;
 FChildL:= NIL;
 FChildR:= NIL;
 FXPos:= -1.0;
 FIsCluster:= false;
 FDist:= dist; //distance d'agrgation
end;

destructor TNodeHCA.Destroy;
begin
 FClus.Free();
 inherited;
end;

const
    CRLF : string = #10#13;

function TNodeHCA.getLabel: string;
var s,sTemp,aName: string;
    j: integer;
    r2,corr: double;
const l = 45;
begin
 s:= '';
 sTemp:= format('Aggreg. Distance    : %7.2f',[self.YPos]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 sTemp:= format('Merging Distance    : %7.2f',[self.FDist]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 s:= s + DupeString('-',l) + CRLF;
 //sTemp:= format('Total Var. Expl. (%s): %7.2f',['%',100.0*(self.FCalc.LstVariables.getCount()-self.YPos)/(1.0*self.FCalc.LstVariables.getCount())]);
 //s:= s + format('%36s',[sTemp]) + CRLF;
 sTemp:= format('Local Var. Expl. (%s): %7.2f',['%',100.0*self.FClus.EigVal/(1.0*self.FClus.getCount())]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 s:= s + DupeString('-',l) + CRLF;
 s:= s + format('%29s %8s %6s',['Att.Name','R-Square','Corr.']) + CRLF;
 s:= s + DupeString('-',l) + CRLF;
 for j:= 0 to pred(self.FClus.Lst.getCount()) do
  begin
   aName:= self.FClus.Lst.getVarClus(j).Att.Name;
   r2:= SQR(self.FClus.TabOwnCorrelation[j]);
   corr:= self.FClus.TabAllCorrelation[pred(self.FClus.Lst.getVarClus(j).Numero)];
   s:= s + format('%29s %8.3f %6.3f',[aName,r2,corr]) + CRLF;
  end;
 result:= copy(s,1,length(s)-2);
end;

function TNodeHCA.isLeaf: boolean;
begin
 result:= (self.FChildL = nil) and (self.FChildR = nil);
end;

{ TSetNodeHCA }

procedure TSetNodeHCA.AddLeaf(node: TNodeHCA);
begin
 FLstAllNodes.Add(node);
 FLstLeaves.Add(node);
 FCandidates.Add(node);
 node.FNumNode:= FLstLeaves.Count;
 node.FDist:= 0.0;
end;

constructor TSetNodeHCA.Create(prmCalc: TCalcHCA);
begin
 inherited Create();
 FCalc:= prmCalc;
 //les listes internes
 FLstAllNodes:= TObjectList.Create(TRUE);
 FLstLeaves:= TObjectList.Create(FALSE);
 FLstMergeNodes:= TObjectList.Create(FALSE);
 FLstNewNodes:= TObjectList.Create(FALSE);
 FCandidates:= TObjectList.Create(false);
 FGlobalYPos:= 0.0;
end;

destructor TSetNodeHCA.Destroy;
begin
 FCandidates.Free();
 FLstNewNodes.Free();
 FLstMergeNodes.Free();
 FLstLeaves.Free();
 FLstAllNodes.Free();
 inherited;
end;

function TSetNodeHCA.getCountNodes: integer;
begin
 result:= self.FLstAllNodes.Count;
end;

function TSetNodeHCA.getNode(index: integer): TNodeHCA;
begin
 result:= self.FLstAllNodes[index] as TNodeHCA;
end;

//critre de tri selon la hauteur de fusion i.e. la valeur propre
function compareOnEigVal(n1,n2: Pointer): integer;
var node1, node2: TNodeHCA;
begin
 node1:= TNodeHCA(n1);
 node2:= TNodeHCA(n2);
 if (node1.FClus.EigVal < node2.FClus.EigVal)
  then result:= -1
  else
    if (node1.FClus.EigVal > node2.FClus.EigVal)
     then result:= +1
     else result:= 0;
end;

//critre de tri selon la distance d'agrgation -- plus le R est grand, plus ils sont proches...
function compareOnDist(n1,n2: Pointer): integer;
var node1, node2: TNodeHCA;
begin
 node1:= TNodeHCA(n1);
 node2:= TNodeHCA(n2);
 if (node1.FDist < node2.FDist)
  then result:= -1
  else
    if (node1.FDist > node2.FDist)
     then result:= +1
     else result:= 0;
end;

//tester si le nombre de variables est suprieur au noeud initial demand
procedure TSetNodeHCA.InitializeNodes(prmExamples: TExamples);
var j,i: integer;
    clus: TVarCluster;
    node,node_i,node_j: TNodeHCA;
    prmKMeans: TPrmOpVarClusKMeans;
    calcKMeans: TCalcVarClusKMeans;
    dist: double;
begin
 if (FCalc.LstVariables.getCount() <= FCalc.FPrm.StartingNbClusters)
  then
   begin
    for j:= 0 to pred(FCalc.LstVariables.getCount()) do
     begin
      clus:= TVarCluster.create(FCalc.Inputs.Size,FCalc);
      clus.addVar(FCalc.LstVariables.getVarClus(j));
      clus.refreshClusVar(prmExamples);
      node:= TNodeHCA.Create(FCalc,clus,0);
      self.AddLeaf(node);
      node.FClus.majCorrelation(prmExamples);
     end;
   end
  else
   begin
    //lancer le K-Means
    prmKMeans:= TPrmOpVarClusKMeans.Create(nil);
    prmKMeans.AskedNbClusters:= FCalc.FPrm.StartingNbClusters;
    //on lui passe notre LstVariables pour ne pas tre dtruit plus tard
    calcKMeans:= TCalcVarClusKMeans.create(FCalc,prmKMeans);
    calcKMeans.execute(prmExamples);
    //rcupration de la solution
    for j:= 0 to pred(calcKMeans.StrucClusters.getCount()) do
     begin
      //il y a bien une variable dedans...
      if (calcKMeans.StrucClusters.get(j).EigVecSize > 0)
       then
        begin
         clus:= calcKMeans.StrucClusters.get(j).duplicate(FCalc);
         node:= TNodeHCA.Create(FCalc,clus,0);
         self.AddLeaf(node);
         node.FClus.majCorrelation(prmExamples);
        end;
     end;
    //and free...
    calcKMeans.Free();
    prmKMeans.Free();
   end;
 //*******************************
 //calculer les noeuds  fusionner
 //*******************************
 for i:= 0 to self.getCountNodes()-2 do
  for j:= succ(i) to self.getCountNodes()-1 do
   begin
    node_i:= self.getNode(i);
    node_j:= self.getNode(j);
    clus:= TVarCluster.fusion(node_i.FClus,node_j.FClus,prmExamples,dist);
    node:= TNodeHCA.Create(Fcalc,clus,dist);
    node.FDad:= nil;
    node.FChildL:= node_i;
    node.FChildR:= node_j;
    FLstMergeNodes.Add(node);
   end;
  //FLstMergeNodes.Sort(compareOnEigVal);
  FLstMergeNodes.Sort(compareOnDist);
end;

procedure TSetNodeHCA.MergeNodes(examples: TExamples);
var solution,node,other: TNodeHCA;
    j: integer;
    clus: TVarCluster;
    dist: double;
begin
 //la premire est la bonne
 solution:= FLstMergeNodes[0] as TNodeHCA;
 FLstMergeNodes.Delete(0);
 //calculer les corrlations internes
 solution.FClus.majCorrelation(examples);
 //supprimer de la liste donc les solutions contenant une des deux feuilles
 for j:= pred(FLstMergeNodes.Count) downto 0 do
  begin
   node:= FLstMergeNodes[j] as TNodeHCA;
   if (node.FChildL = solution.FChildL) or (node.FChildR = solution.FChildR) or (node.FChildR = solution.FChildL) or (node.FChildL = solution.FChildR) 
    then
     begin
      FLstMergeNodes.Delete(j);
      node.Free();
     end;
  end;
 //supprimer de la liste des nouveaux noeuds ceux qui deviennent des enfants
 for j:= pred(FCandidates.Count) downto 0 do
  begin
   node:= FCandidates[j] as TNodeHCA;
   if (node = solution.FChildL) or (node = solution.FChildR)
    then FCandidates.Delete(j);
  end;
 //recalculer la liste des candidats  partir du noeud solution
 for j:= 0 to pred(FCandidates.Count) do
  begin
    other:= FCandidates[j] as TNodeHCA;
    clus:= TVarCluster.fusion(solution.FClus,other.FClus,examples,dist);
    node:= TNodeHCA.Create(Fcalc,clus,dist);
    node.FDad:= nil;
    node.FChildL:= solution;
    node.FChildR:= other;
    FLstMergeNodes.Add(node);
  end;
 //trier ces solutions candidates
 FLstMergeNodes.Sort(compareOnDist);
 //rfrencer ce nouveau noeud solution
 FLstAllNodes.Add(solution);
 FLstNewNodes.Add(solution);
 FCandidates.Add(solution);
 //dire aux enfants qui est leur pre
 solution.FChildL.FDad:= solution;
 solution.FChildR.FDad:= solution;
 //et lui affecter un numro
 solution.FNumNode:= FLstAllNodes.Count;
 //mj de la hauteur dans l'arbre
 FGlobalYPos:= FGlobalYPos + solution.FDist;
 solution.FYPos:= FGlobalYPos;
 //c'est la dernire ?
 if (FCandidates.count = 1) then self.FRootNode:= solution;
end;

CONST
    //dcalage de coordonnes des feuilles
    DECALAGE_COORD_LEAF : double = 10.0;

procedure TSetNodeHCA.setLeavesCoords(node: TNodeHCA;
  var curCoord: double);
begin
 if not(node.isLeaf())
  then
   begin
    self.setLeavesCoords(node.FChildL,curCoord);
    self.setLeavesCoords(node.FChildR,curCoord);
   end
  else
   begin
    node.FXPos:= curCoord;
    curCoord:= curCoord + DECALAGE_COORD_LEAF;
   end;
end;

procedure TSetNodeHCA.setNodesCoords(node: TNodeHCA);
begin
 //doit-on en calculer les coordonnes ?
 if (node.FXPos < 0.0)
  then
   begin
    //tester si fils gauche n'est pas bien plac
    if (node.FChildL.FXPos < 0.0)
     then setNodesCoords(node.FChildL);
    //tester  droite
    if (node.FChildR.FXPos < 0.0)
     then setNodesCoords(node.FChildR);
    //se positionner au milieu de ses enfants
    node.FXPos:= 0.5 * (node.FChildL.FXPos + node.FChildR.FXPos);
   end;
end;

{ TCalcHCA }

procedure TCalcHCA.buildBestNodes(examples: TExamples);
var i: integer;
    node: TNodeHCA;
begin
 FLstNodesSolution:= TObjectList.Create(FALSE);
 FLstNodesSolution.Add(FSetNode.FRootNode);
 for i:= 1 to FIdMaxPalier do
  begin
   node:= FSetNode.FLstNewNodes.Items[FSetNode.FLstNewNodes.Count-i] as TNodeHCA;
   FLstNodesSolution.Extract(node);
   FLstNodesSolution.Add(node.FChildL);
   FLstNodesSolution.Add(node.FChildR);
  end;
 //new -- indiquer au noeud qu'il est un cluster
 StrucClusters.clearClusters();
 for i:= 0 to pred(FLstNodesSolution.Count) do
  begin
   (FLstNodesSolution.Items[i] as TNodeHCA).FIsCluster:= TRUE;
   StrucClusters.add((FLstNodesSolution.Items[i] as TNodeHCA).FClus.duplicate(self));
  end;
 StrucClusters.majCorrelation(examples);
end;

procedure TCalcHCA.computeClusters(prmExamples: TExamples);
begin
 FTSS:= self.Inputs.Count;
 //construire les noeuds de base
 self.InitializeNodes(prmExamples);
 //processus de remonte
 WHILE (FSetNode.FCandidates.Count > 1) DO
  begin
   //TraceLog.WriteToLogFile(format('[VARCLUS-HCA] nombre de fusions evaluees = %d',[FSetNode.FCandidates.Count]));
   FSetNode.MergeNodes(prmExamples);
   //recalcul des variances expliques
   self.refreshVarExplained();
  end;
 //mj des coordonnes pour le dessin
 self.computeNodeCoords();
 //*************************************************
 //dtecter les paliers et dfinir la bonne solution
 //*************************************************
 self.detectPalier();
 //forcer la solution si demande par l'utilisateur
 if (FPrm.ModeBestSelection = 1)
  then self.FIdMaxPalier:= FPrm.NbClusters - 1;
 //dtecter les noeuds du clustering dfinitif
 self.buildBestNodes(prmExamples);
end;

procedure TCalcHCA.computeNodeCoords;
var curCoord: double;
begin
 curCoord:= 0.0;
 //numroter les feuilles
 self.FSetNode.setLeavesCoords(self.FSetNode.FRootNode,curCoord);
 //numroter les noeuds
 self.FSetNode.setNodesCoords(self.FSetNode.FRootNode);
end;

constructor TCalcHCA.create(prmInputs: TLstAttributes;
  param: TOperatorParameter);
begin
 inherited;
 FPrm:= param as TPrmOpVarClusHCA;
 FLstComputedFusions:= TObjectList.Create(true); 
end;

destructor TCalcHCA.Destroy;
begin
 Finalize(FTabPalier);
 Finalize(FTabVarExplained);
 FLstComputedFusions.Free();
 FSetNode.Free();
 if assigned(FLstNodesSolution) then FLstNodesSolution.Free();
 inherited;
end;

(*
//ancienne procdure qui calcule le saut maximal
procedure TCalcHCA.detectPalier();
var i: integer;
    node: TNodeHCA;
    max: double;
begin
 setLength(FTabDifIndex,FSetNode.FLstNewNodes.Count);
 setLength(FTabIndex,FSetNode.FLstNewNodes.Count);
 //tableau des distances d'agrgation
 for i:= 0 to pred(FSetNode.FLstNewNodes.Count) do
  begin
   node:= FSetNode.FLstNewNodes.Items[pred(FSetNode.FLstNewNodes.Count)-i] as TNodeHCA;
   FTabDifIndex[i]:= node.FDist;
   FTabIndex[i]:= 1.0 * Self.LstVariables.getCount() - node.YPos;
  end;
 //calcul des diffrences 1res
 for i:= low(FTabDifIndex) to high(FTabDifIndex)-1 do
  FTabDifIndex[i]:= FTabDifIndex[i]-FTabDifIndex[succ(i)];
 //on peut dtecter le palier
 FIdMaxPalier:= -1;
 max:= -1.0e308;
 for i:= low(FTabDifIndex) to high(FTabDifIndex)-1 do
  begin
   if (i>0) and (FTabDifIndex[i]>max)
    then
     begin
      max:= FTabDifIndex[i];
      FIdMaxPalier:= i;
     end;
  end;
end;
*)

procedure TCalcHCA.detectPalier;
begin
 FIdMaxPalier:= self.getOptimalNbClusters(FTabVarExplained,FTabPalier,FSeuilMV)-1;
end;

function TCalcHCA.getHTMLDetailedResults: string;
begin
 result:= self.getHTMLVarExplained(FTabVarExplained,FTabPalier,succ(FIdMaxPalier),FSeuilMV);
end;

procedure TCalcHCA.InitializeNodes(prmExamples: TExamples);
begin
 FSetNode:= TSetNodeHCA.Create(self);
 FSetNode.InitializeNodes(prmExamples);
 setLength(FTabVarExplained,self.FSetNode.FCandidates.Count);
 self.refreshVarExplained();
end;

procedure TCalcHCA.refreshVarExplained;
var j: integer;
    sum: single;
begin
 sum:= 0.0;
 for j:= 0 to pred(self.FSetNode.FCandidates.Count) do
  sum:= sum + (self.FSetNode.FCandidates[j] as TNodeHCA).FClus.EigVal;
 FTabVarExplained[pred(self.FSetNode.FCandidates.Count)]:= sum;
end;

initialization
 RegisterClass(TMLGenVarClusHCA);
end.
