(******************************************************************)
(* UCompVarClusKMeans.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Composant de base pour le clustering de variables)
@author(Ricco)
@created(19/02/2007)

Clustering sur les variables. Mthode des K-Means.
}

unit UCompVarClusKMeans;

interface

USES
    Classes,
    UCompVarClusBase, UCalcVarClusBase,
    Forms,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompClusteringDefinition,
    UDatasetDefinition,
    UDatasetExamples,
    UCalcStatDes,
    contnrs;

TYPE
    //gnrateur
    TMLGenVarClusKMeans = class(TMLGenVarClus)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLVarClusKMeans = class(TMLCompVarClusBase)
    protected
    function    getClassOperator: TClassOperator; override;
    function    getGenericAttName(): string; override;
    end;

    //oprateur
    TOpVarClusKMeans = class(TOpVarClusBase)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    getClassCalc(): TClassCalcVarClus; override; 
    end;

    //paramtrage
    TPrmOpVarClusKMeans = class(TOpPrmVarClus)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //nombre de clusters demands, df. 5
    AskedNbClusters: integer;
    //maximum d'itrations, df. 10
    MaxIter: integer;
    //Nombdre d'essais, df. 5
    NbTrials: integer;
    //I-O
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;

    //classe de calcul
    TCalcVarClusKMeans = class(TCalcVarClus)
    protected
    //nombre de clusters rellement calculs
    UsedNbClusters: integer;
    //liste des rsultats -- liste des clusters solutions
    LstResult: TObjectList;
    //algorithme de construction
    procedure   computeClusters(examples: TExamples); override;
    //description de la liste des optimisations
    function    getHTMLDetailedResults(): string; override;
    public
    //constructeur
    constructor create(prmInputs: TLstAttributes; param: TOperatorParameter); override;
    constructor create(otherCalc: TCalcVarClus; param: TOperatorParameter); override;
    //destructeur
    destructor  destroy(); override;
    end;

    
implementation

USES
    SysUtils,
    Math, ULogFile, UCalcRndGenerator, UDlgOpPrmVarClusKMeans,
  UConstConfiguration;


{ TMLGenVarClusKMeans }

function TMLGenVarClusKMeans.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLVarClusKMeans; 
end;

{ TMLVarClusKMeans }

function TMLVarClusKMeans.getClassOperator: TClassOperator;
begin
 result:= TOpVarClusKMeans;
end;

function TMLVarClusKMeans.getGenericAttName: string;
begin
 result:= 'VCKMeans';
end;

{ TOpVarClusKMeans }

function TOpVarClusKMeans.getClassCalc: TClassCalcVarClus;
begin
 result:= TCalcVarClusKMeans; 
end;

function TOpVarClusKMeans.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpVarClusKMeans; 
end;

{ TPrmOpVarClusKMeans }

function TPrmOpVarClusKMeans.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmVarClusKMeans.CreateFromOpPrm(self);
end;

function TPrmOpVarClusKMeans.getHTMLParameters: string;
var s: string;
begin
 s:= inherited getHTMLParameters();
 s:= s + format('Number of clusters : %d<br>',[AskedNbClusters]);
 s:= s + format('Max number of iterations : %d<br>',[MaxIter]);
 s:= s + format('Number of attempts : %d<br>',[NbTrials]);
 result:= s;
end;

procedure TPrmOpVarClusKMeans.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 AskedNbClusters:= prmINI.ReadInteger(prmSection,'nb_clusters',AskedNbClusters);
 MaxIter:= prmINI.ReadInteger(prmSection,'max_iter',MaxIter);
 NbTrials:= prmINI.ReadInteger(prmSection,'nb_trials',NbTrials);
end;

procedure TPrmOpVarClusKMeans.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(AskedNbClusters,sizeof(AskedNbClusters));
 prmStream.ReadBuffer(MaxIter,sizeof(MaxIter));
 prmStream.ReadBuffer(NbTrials,sizeof(NbTrials));
end;

procedure TPrmOpVarClusKMeans.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'nb_clusters',AskedNbClusters);
 prmINI.WriteInteger(prmSection,'max_iter',MaxIter);
 prmINI.WriteInteger(prmSection,'nb_trials',NbTrials);
end;

procedure TPrmOpVarClusKMeans.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(AskedNbClusters,sizeof(AskedNbClusters));
 prmStream.WriteBuffer(MaxIter,sizeof(MaxIter));
 prmStream.WriteBuffer(NbTrials,sizeof(NbTrials));
end;

procedure TPrmOpVarClusKMeans.SetDefaultParameters;
begin
 inherited;
 AskedNbClusters:= 4;
 MaxIter:= 10;
 NbTrials:= 5;
end;

{ TCalcVarClusKMeans }

procedure TCalcVarClusKMeans.computeClusters(examples: TExamples);
var j,k,iteration,trial: integer;
    prevExpl,curExpl,bestExpl: double;
    curClusters: TStrucVarCluster;
    shuffled: TExamples;
begin
 //nombre de clusters  produire
 UsedNbClusters:= MATH.Min(Inputs.Count,(OpParameter as TPrmOpVarClusKMeans).AskedNbClusters);
 //liste des variables  shuffler pour varier les dparts
 shuffled:= TExamples.Create(Inputs.Count);
 shuffled.Initialize();
 //*****************
 //pour chaque essai
 //*****************
 for trial:= 1 to (OpParameter as TPrmOpVarClusKMeans).NbTrials do
  begin
   //cluster rsultat
   curClusters:= TStrucVarCluster.create(self);
   for j:= 1 to UsedNbClusters do
    curClusters.add(TVarCluster.create(Inputs.Size,self));
   //initialiser alatoire avec le shuffled
   shuffled.procRandomizeExamples(RndGenerator);
   for j:= 0 to pred(Inputs.Count) do
    begin
     k:= j mod UsedNbClusters;
     curClusters.get(k).addVar(self.LstVariables.getVarClus(pred(shuffled.Number[succ(j)])));
    end;
   curClusters.refreshClusVar(examples);
   curExpl:= curClusters.VarExplained;
   //lancer l'algorithme
   prevExpl:= 0;
   iteration:= 1;
   WHILE (CurExpl > prevExpl) and (iteration <= (OpParameter as TPrmOpVarClusKMeans).MaxIter) DO
    BEGIN
     //TraceLog.WriteToLogFile(format('[VARCLUS-KMEANS] Iteration %d, explained : %.4f',[iteration,curClusters.VarExplained]));
     //vider les clusters
     curClusters.clearClusters();
     for j:= 0 to pred(self.LstVariables.getCount()) do
      curClusters.addVarToNearest(self.LstVariables.getVarClus(j),examples);
     //recalcul des centres de clusters
     curClusters.refreshClusVar(examples);
     prevExpl:= curExpl;
     curExpl:= curClusters.VarExplained;
     //passage au suivant
     inc(iteration);
    END;
   //calculer les diffrentes corrlations
   curClusters.majCorrelation(examples);
   //test
   TraceLog.WriteToLogFile(format('[VARCLUS-KMEANS] Trial : %d, convergence in %d iterations, explained : %.4f',[trial,iteration,curExpl]));
   //ajouter  la liste des rsultats
   LstResult.Add(curClusters);
  end;
 shuffled.Free();
 //******************************
 //dtecter la meilleure solution
 //******************************
 if assigned(StrucClusters) then StrucClusters.Free();
 bestExpl:= -1e308;
 for trial:= 0 to pred((OpParameter as TPrmOpVarClusKMeans).NbTrials) do
  begin
   curClusters:= LstResult[trial] as TStrucVarCluster;
   if (curClusters.VarExplained > bestExpl)
    then
     begin
      bestExpl:= curClusters.VarExplained;
      StrucClusters:= curClusters;
     end;
  end;
end;

constructor TCalcVarClusKMeans.create(prmInputs: TLstAttributes;
  param: TOperatorParameter);
begin
 inherited;
 LstResult:= TObjectList.Create(true);
end;

constructor TCalcVarClusKMeans.create(otherCalc: TCalcVarClus;
  param: TOperatorParameter);
begin
 inherited;
 LstResult:= TObjectList.Create(true);
end;

destructor TCalcVarClusKMeans.destroy;
begin
 LstResult.Free();
 StrucClusters:= nil;//il n'est plus propritaire ici...
 inherited;
end;

function TCalcVarClusKMeans.getHTMLDetailedResults: string;
var s,sTemp: string;
    soluce: TStrucVarCluster;
    j,k: integer;
begin
 //produire la liste des solutions
 s:= '<H4>Details of trials</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Trial</TH><TH>Var.Explained</TH><TH>Clusters''<br>Composition</TH></TR>';
 for j:= 0 to pred(LstResult.Count) do
  begin
   soluce:= LstResult[j] as TStrucVarCluster;
   sTemp:= '(';
   for k:= 0 to pred(soluce.getCount()) do
    sTemp:= sTemp + IntToStr(soluce.get(k).getCount()) + ',';
   sTemp:= copy(sTemp,1,pred(length(sTemp))) + ')';
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%d</TD><TD align=right>%.4f</TD><TD align=right>%s</TD>',[succ(j),soluce.VarExplained,sTemp]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

initialization
 RegisterClass(TMLGenVarClusKMeans);
end.
