(*******************************************************************)
(* UCompVarClusTopDown.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*******************************************************************)

{
@abstract(Composant de base pour le clustering de variables)
@author(Ricco)
@created(19/02/2007)

Clustering sur les variables. Mthode Top-Down.

new -- 23/02/2007 -- VARCLUS telle qu'elle est implmente dans SAS est ingrable, les procdures
de r-affectation  chaque tape sont trop lourdes et  mon avis peu efficaces.
---> on va en driver une version simplifie avec une rotation VARIMAX qui semble trs efficace,
le tout est de trouver une rgle d'arrt simple, la valeur propre 1 comme valeur limite semble un peu limite...

new -- 24/02/2006
Voici le mode de fonctionnement au final :
(1) on fixe un nombre maximum de cluster dans la descente
(2) une option permet de savoir si c'est ce nombre que l'on veut ou si on demande 
l'agorithme de dterminer automatiquement le bon nombre lui mme.
(3) autre particularit, on refuse un dcoupage si le gain est ngatif

}
unit UCompVarClusTopDown;

interface

USES
    Classes,
    UCompVarClusBase, UCalcVarClusBase,
    Forms,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompClusteringDefinition,
    UDatasetDefinition,
    UDatasetExamples,
    UCalcStatDes,
    UCompVarClusHCA,
    Matrices,
    contnrs;

TYPE
    //gnrateur
    TMLGenVarClusTD = class(TMLGenVarClus)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLVarClusTD = class(TMLCompVarClusBase)
    protected
    function    getClassOperator: TClassOperator; override;
    function    getGenericAttName(): string; override;
    end;

    //oprateur
    TOpVarClusTD = class(TOpVarClusBase)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    getClassCalc(): TClassCalcVarClus; override;
    function    getNewForm(): TForm; override;
    end;

    //paramtrage
    TPrmOpVarClusTD = class(TOpPrmVarClus)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //valeur propre seuil pour le 2eme facteur, df. 1.0
    MinEigVal: double;
    //nombre maximum de clusters, df. 20
    MaxNbClusters: integer;
    //Mode de dtermination de la coupure optimale : 0->automatique, 1->manuel, df. 0
    ModeBestSelection: integer;
    //si choix manuel, nombre de clusters, df. 4
    UserNbClusters: integer;
    //I-O
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    //html
    function    getHTMLParameters(): string; override;
    end;

    //tableau des corrlations
    TTabRSquare = array of double;

    //forward
    TCalcVarClusTD = class;

    //noeud de l'arbre
    TNodeTD = class(TObject)
    protected
    //recalculer le vecteur propre aprs rotation
    procedure    recomputeEigVec(MCorr: PMatrix; MQ: PMatrix);
    //obtenir la position
    function     getYPos(): single;
    public
    //numro
    FNumero: integer;
    //oprateur de calcul
    FCalc: TCalcVarClusTD;
    //premire et deuxime valeur propre
    F1EigVal,F2EigVal: double;
    //cluster courant
    FClus: TVarCluster;
    //pre
    PrevNode: TNodeTD;
    //fils droit et gauche
    LeftChild, RightChild: TNodeTD;
    //position du noeud dans le dendrogram
    XPos: single;
    //gain issu de la segmentation
    Goodness: single;
    //permet de savoir si c'est un cluster ou pas
    IsCluster: boolean;
    //constructeur
    constructor  Create(prmCalc: TCalcVarClusTD; clus: TVarCluster);
    //calculer si on peut le spliter
    function     canISplit(const examples: TExamples): boolean;
    //calculer les corrlations au carr pour l'axe naxis
    procedure    computeCorrelation(MVec: PMatrix; axis: integer; examples: TExamples; var tab: TTabRSquare);
    //construire des noeuds enfants  partir d'une liste de variables
    function     getNewNode(lst: TObjectList; examples: TExamples): TNodeTD;
    //destructeur
    destructor   destroy(); override;
    //est-ce une feuille ?
    function     isLeaf(): boolean;
    //rcuprer une description du noeud
    function     getLabel(): string;
    //hauteur dpend du numro dans l'arbre
    property     YPos: single read getYPos;
    end;

    //structure d'arbre
    TSetNodeTD = class(TObject)
    protected
    //rordonner les feuilles
    procedure   reArrangeLeaves(curNode: TNodeTD; var coord: single);
    //rordonner les noeuds intermdiaires
    procedure   reArrangeNodes(curNode: TNodeTD);
    //re-positionner les noeuds dans l'arbre
    procedure   reArrangeTree();
    //re-arranger les sequences selon le gain du split
    procedure   reArrangeSequenceOnGoodness();
    public
    //numro global
    FGlobalNumero: integer;
    //calculateur
    FCalc: TCalcVarClusTD;
    //la racine
    FRoot: TNodeTD;
    //liste des candidats  la coupure
    FCandidates: TObjectList;
    //liste des squences
    FSequence: TObjectList;
    //les feuilles de l'arbre
    FLeaves: TObjectList;
    //tableau de suivi des inerties expliques
    FTabVarExplained: TTabVarExplained;
    //recalculer les inerties expliques
    procedure   refreshVarExplained();
    //dtruire la structure
    procedure   FreeNodes(var node: TNodeTD);
    //constructeur
    constructor create(calc: TCalcVarClusTD);
    //destructeur
    destructor  destroy(); override;
    //ajouter les noeuds qui n'ont pas produit de feuilles aux squences
    procedure AddCandidateToSequence(examples: TExamples);
    end;
    

    //classe de calcul
    TCalcVarClusTD = class(TCalcVarClus)
    protected
    //liste des solutions en 2, 3, etc. clusters... pour viter les re-calcul -- liste propritaire, mais les listes internes non
    FListeSolutions: TObjectList;
    //seuil du nombre de clusters lors de la construction
    SeuilNbClusters: integer;
    //nombre de cluster final
    FinalNbClusters: integer;
    //seuil Moving average pour la dtection de la zone d'angle la plus intressante
    FSeuilMV: double;
    //structure de noeuds
    FSetNode: TSetNodeTD;
    //tableau des dif. de var explained, cos et angle pour chaque solution  valuer
    FTabPalier: TTabPalier;
    //construire les clusters
    procedure   computeClusters(prmExamples: TExamples); override;
    //construire l'arbre vers le bas
    //procedure   computeSpliting(curNode: TNodeTD; prmExamples: TExamples);
    //autre approche de construction de l'arbre
    procedure   computeNextSpliting(prmExamples: TExamples);
    //slection le bon nombre de clusters -- si option demande
    procedure   detectBestSolution(andApply: boolean);
    //rcuprer les feuilles pour les mettre dans une structure de cluster
    procedure   recupLeaves(node: TNodeTD);
    //dcrire la squence des coupures
    function    getHTMLDetailedResults(): string; override;
    //nouvelles initialisations internes
    procedure   commonInitialisations(); override;
    public
    //
    destructor  destroy(); override;
    //arbre
    property    SetNode: TSetNodeTD read FSetNode;
    end;


implementation

uses
    Math, SysUtils, StrUtils,
    UCalcMatrixAdditionalFunctions, UCalcFactRotateFactors,
    UDatasetImplementation, ULogFile, UConstConfiguration,
    UFrmOpViewVarClusTopDown, UDlgOpPrmVarClusTopDown;

{ TMLGenVarClusTD }

function TMLGenVarClusTD.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLVarClusTD; 
end;

{ TMLVarClusTD }

function TMLVarClusTD.getClassOperator: TClassOperator;
begin
 result:= TOpVarClusTD; 
end;

function TMLVarClusTD.getGenericAttName: string;
begin
 result:= 'VarClus';
end;

{ TOpVarClusTD }

function TOpVarClusTD.getClassCalc: TClassCalcVarClus;
begin
 result:= TCalcVarClusTD;
end;

function TOpVarClusTD.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpVarClusTD; 
end;

function TOpVarClusTD.getNewForm: TForm;
begin
 result:= TFrmOpViewVarClusTopDown.CreateFromOperator(self);
end;

{ TPrmOpVarClusTD }

function TPrmOpVarClusTD.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmVarClusTopDown.CreateFromOpPrm(self);
end;

function TPrmOpVarClusTD.getHTMLParameters: string;
var s: string;
begin
 s:= inherited getHTMLParameters();
 s:= s + format('Threshold eigen value on second factor : %.2f<br>',[MinEigVal]);
 s:= s + format('Maximum number of the leaves in the search algorithm : %d<br>',[MaxNbClusters]);
 s:= s + format('User-defined number of clusters : %d<br>',[ModeBestSelection]);
 s:= s + format('Number of clusters, if user-defined : %d<br>',[UserNbClusters]);
 result:= s; 
end;

procedure TPrmOpVarClusTD.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 MinEigVal:= prmINI.ReadFloat(prmSection,'min_eig_val',MinEigVal);
 MaxNbClusters:= prmINI.ReadInteger(prmSection,'max_nb_clusters',MaxNbClusters);
 ModeBestSelection:= prmINI.ReadInteger(prmSection,'cluster_selection',ModeBestSelection);
 UserNbClusters:= prmINI.ReadInteger(prmSection,'user_defined_nb_cluster',UserNbClusters);
end;

procedure TPrmOpVarClusTD.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(MinEigVal,sizeof(MinEigVal));
 prmStream.ReadBuffer(MaxNbClusters,sizeof(MaxNbClusters));
 prmStream.ReadBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.ReadBuffer(UserNbClusters,sizeof(UserNbClusters));
end;

procedure TPrmOpVarClusTD.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteFloat(prmSection,'min_eig_val',MinEigVal);
 prmINI.WriteInteger(prmSection,'max_nb_clusters',MaxNbClusters);
 prmINI.WriteInteger(prmSection,'cluster_selection',ModeBestSelection);
 prmINI.WriteInteger(prmSection,'user_defined_nb_cluster',UserNbClusters);
end;

procedure TPrmOpVarClusTD.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(MinEigVal,sizeof(MinEigVal));
 prmStream.WriteBuffer(MaxNbClusters,sizeof(MaxNbClusters));
 prmStream.WriteBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.WriteBuffer(UserNbClusters,sizeof(UserNbClusters));
end;

procedure TPrmOpVarClusTD.SetDefaultParameters;
begin
 inherited;
 MinEigVal:= 1.0;
 MaxNbClusters:= 15;
 ModeBestSelection:= 0;
 UserNbClusters:= 4;
end;

{ TCalcVarClusTD }

//comparer sur les hauteurs, tri dcroissant
function compareOnEigenValue(item1, Item2: Pointer): integer;
var n1,n2: TNodeTD;
begin
 n1:= TNodeTD(Item1);
 n2:= TNodeTD(Item2);
 if (n1.FClus.EigVal > n2.FClus.EigVal)
  then result:= -1
  else
   if (n1.FClus.EigVal < n2.FClus.EigVal)
    then result:= +1
    else result:= 0;
end;

//comparer sur la proportion de variance explique -- tri croissant
function compareOnProportion(item1, Item2: Pointer): integer;
var pv1, pv2: double;
begin
 pv1:= TNodeTD(Item1).FClus.getPropotionVarExplained();
 pv2:= TNodeTD(Item2).FClus.getPropotionVarExplained();
 if (pv1 < pv2)
  then result:= -1
  else
   if (pv1 > pv2)
    then result:= +1
    else result:= 0;
end;

//comparer selon l'ordre de cration
function compareOnNumero(item1, Item2: Pointer): integer;
var n1,n2: TNodeTD;
begin
 n1:= TNodeTD(Item1);
 n2:= TNodeTD(Item2);
 if (n1.FNumero < n2.FNumero)
  then result:= -1
  else
   if (n1.FNumero > n2.FNumero)
    then result:= +1
    else result:= 0;
end;

procedure TCalcVarClusTD.commonInitialisations;
begin
 inherited;
 //liste des solutions
 FListeSolutions:= TObjectList.Create(true);
end;

procedure TCalcVarClusTD.computeClusters(prmExamples: TExamples);
var clus: TVarCluster;
begin
 //le seuil en nombre de clusters
 if ((self.OpParameter as TPrmOpVarClusTD).ModeBestSelection = 0)
  then SeuilNbClusters:= (self.OpParameter as TPrmOpVarClusTD).MaxNbClusters
  else SeuilNbClusters:= MATH.Min((self.OpParameter as TPrmOpVarClusTD).MaxNbClusters,(self.OpParameter as TPrmOpVarClusTD).UserNbClusters);
 //structure d'arbre
 if assigned(FSetNode) then FSetNode.Free();
 FSetNode:= TSetNodeTD.Create(self);
 //le premier cluster -- la racine de l'arbre
 clus:= TVarCluster.create(self.Inputs.Size,self);
 clus.Lst.assign(self.LstVariables);
 clus.refreshClusVar(prmExamples);
 FSetNode.FRoot:= TNodeTD.Create(self,clus);
 FSetNode.FCandidates.Add(FSetNode.FRoot);
 FSetNode.FLeaves.Add(FSetNode.FRoot);
 FSetNode.refreshVarExplained();
 //assigner un numro
 FSetNode.FRoot.FNumero:= succ(FSetNode.FGlobalNumero);
 inc(FSetNode.FGlobalNumero);
 //on enchane vers le bas maintenant
 //self.computeSpliting(FSetNode.FRoot,prmExamples);
 WHILE (FSetNode.FCandidates.Count > 0) and (FSetNode.FLeaves.Count < SeuilNbClusters) DO
  self.computeNextSpliting(prmExamples);
 //vrifier que les sommets feuilles n'ont pas de branche
 if (FSetNode.FCandidates.Count > 0) then FSetNode.AddCandidateToSequence(prmExamples);
 //trier la squence par importance du gain du au split
 FSetNode.reArrangeSequenceOnGoodness();
 //>> rcuprer l'arbre optimal voulu
 if ((self.OpParameter as TPrmOpVarClusTD).ModeBestSelection = 0)
  then self.detectBestSolution(true)
  else
   begin
    self.detectBestSolution(false);//pour avoir les tableaux
    //rcuprer  partir des feuilles
    StrucClusters.clearClusters();
    self.recupLeaves(FSetNode.FRoot);
   end;
 //et donc calculer les corrlations et rcuprer le nombre de clusters final...
 StrucClusters.majCorrelation(prmExamples);
 FinalNbClusters:= StrucClusters.getCount();
 //mise en forme de l'arbre
 FSetNode.reArrangeTree();
end;

procedure TCalcVarClusTD.computeNextSpliting(prmExamples: TExamples);
var toSplit: TNodeTD;
begin
 //tester si le premier noeud est splitable
 toSplit:= FSetNode.FCandidates.First as TNodeTD;
 //on peut le splitter ?
 if (toSplit.canISplit(prmExamples) = true)
  then
   begin
    //Ok, ses enfants deviennent candidats
    FSetNode.FCandidates.Add(toSplit.LeftChild);
    FSetNode.FCandidates.Add(toSplit.RightChild);
    //assigner des numros
    toSplit.LeftChild.FNumero:= FSetNode.FGlobalNumero + 1;
    toSplit.RightChild.FNumero:= FSetNode.FGlobalNumero + 2;
    inc(FSetNode.FGlobalNumero,2);
    //gestion des feuilles
    FSetNode.FLeaves.Add(toSplit.LeftChild);
    FSetNode.FLeaves.Add(toSplit.RightChild);
    FSetNode.FLeaves.Extract(toSplit);
    FSetNode.refreshVarExplained();
    //re-trier pour mettre en comptition -- on cherche  splitter ceux qui ont forte proportion de variance explique
    FSetNode.FCandidates.Sort(compareOnProportion);
    //indiquer le gain issu de la segmentation
    toSplit.Goodness:= toSplit.LeftChild.FClus.EigVal + toSplit.RightChild.FClus.EigVal - toSplit.FClus.EigVal;  
   end;
 //il est vir de toute manire
 FSetNode.FCandidates.Extract(toSplit);
 //l'indiquer dans la squence
 FSetNode.FSequence.Add(toSplit);
 //on aura besoin quand mme de ses corrlations
 toSplit.FClus.majCorrelation(prmExamples);
end;

destructor TCalcVarClusTD.destroy;
begin
 if assigned(FSetNode) then FSetNode.Free();//GongYu
 FListeSolutions.Free();
 inherited;
end;

procedure TCalcVarClusTD.detectBestSolution(andApply: boolean);
var soluce: TObjectList;
    k: integer;
begin
 //rcuprer le bon nombre de clusters
 FinalNbClusters:= self.getOptimalNbClusters(FSetNode.FTabVarExplained,FTabPalier,FSeuilMV);
 if (andApply = true)
  then
   begin
    //dfinir la solution retenue et former les clusters
    soluce:= self.FListeSolutions[FinalNbClusters-1] as TObjectList;
    Self.StrucClusters.clearClusters();
    for k:= 0 to pred(soluce.Count) do
     begin
      Self.StrucClusters.add((soluce[k] as TNodeTD).FClus.duplicate(self));
      (soluce[k] as TNodeTD).isCluster:= true;
     end;
   end;
end;

function TCalcVarClusTD.getHTMLDetailedResults: string;
var s: string;
    j: integer;
    node: TNodeTD;
    sChild: string;
    sTRColor: string;
begin
 //sequence de construction de l'arbre
 s:= '<H4>Clustering sequence</H4>';
 s:= s + '<i>Note : 2nd eigen value is computed on rotated axis</i><br>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Node<br>number</TH><TH>Childs</TH><TH>Members</TH><TH>Var.<br>Explained</TH><TH>Prop.<br>Explained</TH><TH>Second<br>Eigen value</TH></TR>';
 for j:= 0 to pred(FSetNode.FSequence.Count) do
  begin
   node:= FSetNode.FSequence[j] as TNodeTD;
   if node.isLeaf()
    then
     begin
      sChild:= '()';
      sTRColor:= HTML_TABLE_COLOR_HEADER_GRAY;
     end
    else
     begin
      sChild:= format('(%d,%d)',[node.LeftChild.Fnumero,node.RightChild.FNumero]);
      sTRColor:= HTML_TABLE_COLOR_DATA_GRAY;
     end;
   s:= s + sTRColor + format('<TD>%d</TD><TD align=center>%s</TD><TD>%d</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD></TR>',
           [node.FNumero,sChild,node.FClus.GetCount(),node.FClus.EigVal,node.FClus.getPropotionVarExplained(),node.F2EigVal]);
  end;
 s:= s + '</table>';
 //suite : variances. expliques pour chaque nombre de clusters
 s:= s + self.getHTMLVarExplained(FSetNode.FTabVarExplained,FTabPalier,FinalNbClusters,FSeuilMV);
 // 
 result:= s;
end;

procedure TCalcVarClusTD.recupLeaves(node: TNodeTD);
begin
 if (node.isLeaf() = false)
  then
   begin
    self.recupLeaves(node.LeftChild);
    self.recupLeaves(node.RightChild);
   end
  else Self.StrucClusters.add(node.FClus.duplicate(self));
end;

{ TNodeTD }

function TNodeTD.canISplit(const examples: TExamples): boolean;
var MCorr: PMatrix;
    eVal: PVector;
    eVec: PMatrix;
    n: integer;
    j,i: integer;
    t1, t2: TTabRSquare;
    lst1, lst2: TObjectList;
    MQ: PMatrix;
    left,right: TNodeTD;
    ok: boolean;
    gain: double;
begin
 TRY
 n:= self.FClus.EigVEcSize;
 if (n > 1)
  then
   begin
   MCorr:= self.FClus.getCorrMatrix();
   GetEigensFromSymetricMatrix(MCorr,n,eVal,eVec);
   //TraceLog.WriteToLogFile('ok getEigen');
   //il faut : projeter, calculer les corrlations
   self.computeCorrelation(eVec,1,examples,t1);
   self.computeCorrelation(eVec,2,examples,t2);
   //TraceLog.WriteToLogFile('ok correlation');
   //copier dans une matrice
   dimMatrix(MQ,n,2);
   for i:= 1 to n do MQ^[i]^[1]:= t1[i-1];
   for i:= 1 to n do MQ^[i]^[2]:= t2[i-1];
   F1EigVal:= eVal^[1];
   F2EigVal:= eVal^[2];
   //******************
   //* rotation VARIMAX
   //******************
   TRY
   //>>
   //QUARTIMAX(MQ,n,2);
   VARIMAX(MQ,n,2);
   //TraceLog.WriteToLogFile('ok varimax');
   //>>
   //re-calcul des valeurs propres
   F1EigVal:= 0.0;
   for i:= 1 to n do F1EigVal:= F1EigVal + SQR(MQ^[i]^[1]);
   F2EigVal:= 0.0;
   for i:= 1 to n do F2EigVal:= F2EigVal + SQR(MQ^[i]^[2]);
   //rcuprer les tableaux -- comme cela tous les tests se font sur les tableaux t1 et t2
   //consquence : si la rotation a plant, on effectue les valuations sur les axes non-rotationns !!! et c'est trs bien !
   for i:= 1 to n do t1[i-1]:= MQ^[i]^[1];
   for i:= 1 to n do t2[i-1]:= MQ^[i]^[2];
   EXCEPT
   on e: Exception do
    TraceLog.WriteToLogFile(format('[VARCLUS-TD] error during VARIMAX, rotation not achieved (Eig1:%.4f ; Eig2:%.4f)',[F1EigVal,F2EigVal]));
   END;
   //*********** attention **************************************************
   //* on effectue la rotation pour obtenir une valeur plus tranche des r2 avec les axes
   //* la deuxime valeur propre augmente, certes...
   //* mais on ne doit pas recalculer le vecteur propre et l'axe qui "reprsente" le groupe de variables
   //* c'est trs important pour la variable construite que l'on peut utiliser par la suite...
   //************************************************************************
   (*
   //recalculer le vecteur de projection
   self.recomputeEigVec(MCorr,MQ);
   //refaire la projection...
   self.FClus.setProjectionFromEigVector();
   *)
   //est-ce que la deuxime valeur propre passe ?
   if (F2EigVal >= (self.FCalc.OpParameter as TPrmOpVarClusTD).MinEigVal)
    then
     begin
      //*******************************************
      //voyons si on peut spliter sur les deux fils
      //*******************************************
      lst1:= TObjectList.Create(false);
      lst2:= TObjectList.Create(false);
      for j:= 0 to pred(self.FClus.getCount()) do
       begin
        if (SQR(t1[j]) > SQR(t2[j]))
         then lst1.Add(self.FClus.Lst.getVarClus(j))
         else lst2.Add(self.FClus.Lst.getVarClus(j));
       end;
      //a-t-on une solution satisfaisante pour tout le monde
      if (lst1.Count > 0) and (lst2.Count > 0)
       then
        begin
         left:= self.getNewNode(lst1,examples);
         right:= self.getNewNode(lst2,examples);
         gain:= (left.FClus.EigVal + right.FClus.EigVal - self.FClus.EigVal);
         ok:= (gain > 0.0) or ((self.FCalc.OpParameter as TPrmOpVarClusTD).ModeBestSelection = 1); 
         if ok
          then
           begin
            //TraceLog.WriteToLogFile(format('[VARCLUS-TD] %d is splitted into (%d, %d) --> %.4f',[n,lst1.count,lst2.count,gain]));
            self.LeftChild:= left;
            self.LeftChild.PrevNode:= self;
            self.RightChild:= right;
            self.RightChild.PrevNode:= self;
           end
          else
           begin
            //TraceLog.WriteToLogFile(format('[VARCLUS-TD] splittind %d into (%d, %d) refused --> %.4F',[n,lst1.count,lst2.count,gain]));
            left.Free();
            right.Free();
           end;
         //and then...
         result:= ok;
        end
       else result:= false;
      //vider le tout
      lst1.Free(); lst2.Free();
     end
    else
     begin
      result:= false;//fin de l'exploration  ce stade
     end;
    //librer
    delMatrix(MCorr,n,n);
    delVector(eVal,n);
    delMatrix(eVec,n,n);
    delMatrix(MQ,n,2);
    //librer les tableaux
    finalize(t1); finalize(t2);
   end
  else result:= false;
 EXCEPT
 on e: Exception do
  begin
   TraceLog.WriteToLogFile(format('[VARCLUS-TD] error in CanISplit --> %s',[e.message]));
   result:= false;
  end;
 END;
end;

procedure TNodeTD.computeCorrelation(MVec: PMatrix; axis: integer; examples: TExamples; var tab: TTabRSquare);
var newAtt: TAttribute;
    i,j: integer;
    value,sum: single;
    attV: TAttVarClus;
begin
 //tableau de corrlations
 setLength(tab,self.FClus.getCount());
 //nouvelle variable de projection
 newAtt:= TAttContinue.Create('clus',self.FClus.VarClus.Size);
 for i:= 1 to newAtt.Size do
  begin
   sum:= 0.0;
   for j:= 1 to self.FClus.getCount() do
    begin
     attV:= self.FClus.Lst.getVarClus(pred(j));
     value:= (attV.Att.cValue[i] - attV.StatDes.Average) / attV.StatDes.StdDevPop;
     sum:= sum + MVec^[j]^[axis] * value;
    end;
   newAtt.cValue[i]:= sum;
 end;
 //calculer les corrlations
 for j:= 1 to self.FClus.getCount() do
  begin
   attV:= self.FClus.Lst.getVarClus(pred(j));
   tab[pred(j)]:= attV.getCorrelation(newAtt,examples);
  end;
 //...
 newAtt.Free();
end;

const
    MIN_VALUE_GOODNESS : single = -1.0 * MATH.MaxSingle;

constructor TNodeTD.Create(prmCalc: TCalcVarClusTD; clus: TVarCluster);
begin
 inherited create();
 FCalc:= prmCalc;
 FClus:= clus;//propritaire...
 XPos:= -1;
 PrevNode:= NIL;
 Goodness:= MIN_VALUE_GOODNESS;
 IsCluster:= false;
end;

destructor TNodeTD.destroy;
begin
 FClus.Free();
 inherited;
end;


const
    CRLF : string = #10#13;

function TNodeTD.getLabel: string;
var s,sTemp: string;
    j: integer;
    aName: string;
    r2,corr: double;
const l = 45;
begin
 sTemp:= format('1st Eig.Val. : %7.2f',[self.FClus.EigVal]);
 s:= format('%45s',[sTemp]) + CRLF;
 sTemp:= format('2nd Eig.Val. : %7.2f',[self.F2EigVal]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 sTemp:= format('%svar. expl. : %7.2f',['%',100.0*self.FClus.getPropotionVarExplained()]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 if (self.Goodness = MIN_VALUE_GOODNESS)
  then sTemp:= format('Goodness of Split : %7s',['-'])
  else sTemp:= format('Goodness of Split : %7.2f',[self.Goodness]);
 s:= s + format('%45s',[sTemp]) + CRLF;
 s:= s + DupeString('-',l) + CRLF;
 s:= s + format('%29s %8s %6s',['Att.Name','R-Square','Corr.']) + CRLF;
 s:= s + DupeString('-',l) + CRLF;
 for j:= 0 to pred(self.FClus.Lst.getCount()) do
  begin
   aName:= self.FClus.Lst.getVarClus(j).Att.Name;
   r2:= SQR(self.FClus.TabOwnCorrelation[j]);
   corr:= self.FClus.TabAllCorrelation[pred(self.FClus.Lst.getVarClus(j).Numero)];
   s:= s + format('%29s %8.3f %6.3f',[aName,r2,corr]) + CRLF;
  end;
 result:= copy(s,1,length(s)-2);
end;

function TNodeTD.getNewNode(lst: TObjectList; examples: TExamples): TNodeTD;
var clus: TVarCluster;
    j: integer;
begin
 clus:= TVarCluster.create(self.FClus.VarClus.Size,self.FCalc);
 for j:= 0 to pred(lst.Count) do
  clus.addVar(lst[j] as TAttVarClus);
 clus.refreshClusVar(examples);
 result:= TNodeTD.Create(self.FCalc,clus);
end;

function TNodeTD.getYPos: single;
begin
 result:= self.FCalc.LstVariables.getCount()-self.FNumero;
 if (result < 0) then result:= 0;
end;

function TNodeTD.isLeaf: boolean;
begin
 result:= (self.LeftChild = nil) and (self.RightChild = nil);
end;

const
 //cartement entre deux feuilles
 DECALAGE_COORD_LEAF : single = 10;

procedure TSetNodeTD.reArrangeLeaves(curNode: TNodeTD; var coord: single);
begin
 if not(curNode.isLeaf())
  then
   begin
    self.reArrangeLeaves(curNode.LeftChild,coord);
    self.reArrangeLeaves(curNode.RightChild,coord);
   end
  else
   begin
    curNode.XPos:= coord;
    coord:= coord + DECALAGE_COORD_LEAF;
   end;
end;

procedure TSetNodeTD.reArrangeNodes(curNode: TNodeTD);
begin
 //doit-on en calculer les coordonnes ?
 if (curNode.XPos < 0.0)
  then
   begin
    //tester si fils gauche n'est pas bien plac
    if (curNode.LeftChild.XPos < 0.0)
     then reArrangeNodes(curNode.LeftChild);
    //tester  droite
    if (curNode.RightChild.XPos < 0.0)
     then reArrangeNodes(curNode.RightChild);
    //se positionner au milieu de ses enfants
    curNode.XPos:= 0.5 * (curNode.LeftChild.XPos + curNode.RightChild.XPos);
    //curNode.YPos:= curNode.FClus.EigVal;
   end;
end;

procedure TNodeTD.recomputeEigVec(MCorr, MQ: PMatrix);
var invCorr: PMatrix;
    n,i: integer;
    NewEgVec: PMatrix;
    sum,newValue: single;
begin
 n:= self.FClus.EigVecSize;
 //inverser la matrice de corrlation
 Matrices.DimMatrix(invCorr,n,n);
 Matrices.CopyMatrix(invCorr,MCorr,1,1,n,n);
 UCalcMatrixAdditionalFunctions.SVDInverse(invCorr,n);
 UCalcMatrixAdditionalFunctions.MatrixProd(invCorr,n,n,MQ,n,2,NewEgVec);
 //norme
 sum:= 0.0;
 for i:= 1 to n do sum:= sum + NewEgVec[i,1] * NewEgVec[i,1];
 sum:= SQRT(sum);
 //normer maintenant et recopier
 if (sum > 0)
  then
   for i:= 1 to n do
    begin
     newValue:= NewEgVec[i,1] / sum;
     self.FClus.EigVector^[i]:= newValue;
    end;
 //librer
 delMatrix(NewEgVec,n,2);
 delMatrix(invCorr,n,n);
end;

{ TSetNodeTD }

constructor TSetNodeTD.create(calc: TCalcVarClusTD);
begin
 inherited create();
 FCalc:= calc;
 FCandidates:= TObjectList.Create(false);
 FSequence:= TObjectList.Create(false);
 FLeaves:= TObjectList.Create(false);
 FGlobalNumero:= 0;
end;

destructor TSetNodeTD.destroy;
begin
 Finalize(FTabVarExplained);
 FLeaves.Free();
 FCandidates.Free();
 FSequence.Free();
 //dtruire en profondeur l'arbre
 FreeNodes(FRoot);
 inherited;
end;

procedure TSetNodeTD.FreeNodes(var node: TNodeTD);
begin
 if (node.isLeaf() = false)
  then
   begin
    FreeNodes(node.LeftChild);
    FreeNodes(node.RightChild);
   end;
 FreeAndNil(node);
end;

procedure TSetNodeTD.reArrangeTree();
var coord: single;
begin
 coord:= 0.0;
 self.reArrangeLeaves(FRoot,coord);
 self.reArrangeNodes(FRoot);
end;

procedure TSetNodeTD.AddCandidateToSequence(examples: TExamples);
var j: integer;
    node: TNodeTD;
begin
 for j:= 0 to pred(FCandidates.Count) do
  begin
   node:= FCandidates[j] as TNodeTD;
   node.FClus.majCorrelation(examples);//crucial...
   FSequence.Add(node);
  end;
end;

procedure TSetNodeTD.refreshVarExplained;
var j,n: integer;
begin
 n:= FLeaves.Count;
 setlength(FTabVarExplained,n);
 FTabVarExplained[pred(n)]:= 0.0;
 for j:= 0 to pred(FLeaves.Count) do
  FTabVarExplained[pred(n)]:= FTabVarExplained[pred(n)] + (FLeaves[j] as TNodeTD).FClus.EigVal;
end;

//critre de tri dcroissant selon le goodness
function compareOnGoodness(Item1, Item2: Pointer): integer;
begin
 //comparaison sur le goodness d'abord, en cas d'galit -> on passe au Eigen value
 if (TNodeTD(Item1).Goodness < TNodeTD(Item2).Goodness)
  then result:= -1
  else if (TNodeTD(Item1).Goodness > TNodeTD(Item2).Goodness)
   then result:= +1
   else
    begin
     if (TNodeTD(Item1).FClus.EigVal < TNodeTD(Item2).FClus.EigVal)
      then result:= -1
      else if (TNodeTD(Item1).FClus.EigVal > TNodeTD(Item2).FClus.EigVal)
       then result:= +1
       else result:= 0;
    end;
end;

procedure TSetNodeTD.reArrangeSequenceOnGoodness;
var tmpCandidate: TObjectList;
    solution: TNodeTD;
    nbCluster: integer;
    localSolution: TObjectList;
begin
 self.FSequence.Clear();
 //prendre la premire
 tmpCandidate:= TObjectList.Create(false);
 self.FTabVarExplained[0]:= FRoot.FClus.EigVal;
 //solution triviale  un seul cluster
 localSolution:= TObjectList.Create(false);
 localSolution.Add(self.FRoot);
 self.FCalc.FListeSolutions.Add(localSolution);
 //prparation de l'exploration
 tmpCandidate.Add(self.FRoot);
 //allons-y maintenant
 nbCluster:= 1;
 while (tmpCandidate.Count > 0) do
  begin
   solution:= tmpCandidate.Last as TNodeTD;
   self.FSequence.Add(solution);
   //solution.FNumero:= self.FSequence.Count; //non, ne pas modifier les numros parce qu'on s'y perd
   tmpCandidate.Delete(pred(tmpCandidate.Count));
   if (solution.isLeaf() = false)
    then
     begin
      //produire la solution  nbCluster+1 suivante et trier selon le gain
      tmpCandidate.Add(solution.LeftChild);
      tmpCandidate.Add(solution.RightChild);
      tmpCandidate.Sort(compareOnGoodness);
      self.FTabVarExplained[nbCluster]:= self.FTabVarExplained[nbCluster-1] - solution.FClus.EigVal + solution.LeftChild.FClus.EigVal + solution.RightChild.FClus.EigVal;
      //rcuprer la solution locale  nbCluster
      localSolution:= TObjectList.Create(false);
      localSolution.Assign(tmpCandidate);
      self.FCalc.FListeSolutions.Add(localSolution);
      //au suivant...
      inc(nbCluster);
     end;
  end;
 tmpCandidate.Free();
end;

initialization
 RegisterClass(TMLGenVarClusTD);
end.
