(*****************************************************************)
(* UCompPLSFactorial.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(PLS -- Comme une analyse factorielle)
@author(Ricco)
@created(31/07/2006)

Lors de la projection, on cre les axes factoriels et non pas la prvision de la rgression elle-mme.
Rien ne nous empche par la suite d'exploiter ces axes pour la rgression.

}

unit UCompPLSFactorial;

interface

USES
    Classes, Forms,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    JFGCompPLS;

TYPE
       //***********************************************************************
       //** gnrateur de composant, son rle est de crer une instance de composant qui est insr dans le diagramme de traitements
       //***********************************************************************
       TMLGenCompPLSFact = class(TMLGenCompPLS)
       public
       //spcifier quel est la classe du composant  gnrer sur le diagramme --  surcharger obligatoirement chez les hritiers
       function    getClassMLComponent: TClassMLComponent; override;
       end;

       //***********************************************************************
       //** classe du composant graphique insr dans le diagramme
       //***********************************************************************
       TMLCompPLSFact = class(TMLCompPLS)
       //spcifier l'oprateur de calcul --  surcharger obligatoirement
       function    getClassOperator: TClassOperator; override;
       end;

       //***********************************************************************
       //** classe oprateur de calcul : prend en charge le calcul et l'affichage des rsultats
       //***********************************************************************
       TOpPLSFact = class(TOpPLS)
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       function    getAttName(): string; override;
       procedure   SetProjections(); override;
       public
       function    getHTMLResultsSummary(): string; override;
       end;

       //***********************************************************************
       //** classe paramtrage de l'oprateur de calcul : gre la bote de paramtrage et l'E/S des paramtres
       //** si la mthode de requiert pas de paramtres, il suffit de crer une classe vide hritire de "TOperatorParameter"
       //***********************************************************************
       TOpPrmPLSFact = class(TOpPrmPLS)
       protected
       function    CreateDlgParameters(): TForm; override;
       procedure   SetDefaultParameters(); override;
       public
       //si pas de detailed result, que faut-il afficher dans ce cas ?
       ShowRedundancy: boolean;
       ShowCoefAxis: boolean;
       ShowVIP: boolean;
       //sauvegarde
       procedure   LoadFromStream(prmStream: TStream); override;
       procedure   SaveToStream(prmStream: TStream); override;
       procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
       procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
       end;


implementation

USES
    Math, Sysutils,
    UDatasetDefinition, UDatasetImplementation, UDlgOpPrmPlsFactorial;

{ TMLGenCompPLSFact }

function TMLGenCompPLSFact.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompPLSFact;
end;

{ TMLCompPLSFact }

function TMLCompPLSFact.getClassOperator: TClassOperator;
begin
 result:= TOpPLSFact;
end;

{ TOpPLSFact }

function TOpPLSFact.getAttName: string;
begin
 result:= 'Axis';
end;

function TOpPLSFact.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmPLSFact;
end;

function TOpPLSFact.getHTMLResultsSummary: string;
begin
 if (self.PrmOp as TOpPrmPLSFact).DetailedResult
  then result:= inherited getHTMLResultsSummary()
  else result:= FCalc.getHTMLFactorialResults((self.PrmOp as TOpPrmPLSFact).ShowRedundancy,(self.PrmOp as TOpPrmPLSFact).ShowCoefAxis,(self.PrmOp as TOpPrmPLSFact).ShowVIP);
end;

procedure TOpPLSFact.SetProjections;
var i: integer;
    att: TAttribute;
begin
 self.GenAtts.Clear();
 for i:= 1 to MATH.Min(self.WorkData.LstAtts[asInput].Count,(self.PrmOp as TOpPrmPLSFact).NbAxisCreated)  do
  begin
   att:= TAttContinue.Create(Format('PLS_%d_%s_%d',[(MLOwner as TMLComponent).Number,self.getAttName(),i]),WorkData.LstAtts[asAll].Size);
   GenAtts.Add(att);
  end;
 //puis demander leur remplissage par projection dans le nouvel espace
 //TraceLog.WriteToLogFile('[NIPALS] >>> begin projection');
 FCalc.setProjectionAxis(GenAtts);
end;

{ TOpPrmPLSFact }

function TOpPrmPLSFact.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmPLSFactorial.CreateFromOpPrm(self);
end;

procedure TOpPrmPLSFact.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 ShowRedundancy:= prmINI.ReadBool(prmSection,'show_redundancy',ShowRedundancy);
 ShowCoefAxis:= prmINI.ReadBool(prmSection,'show_coef_axis',ShowCoefAxis);
 ShowVIP:= prmINI.ReadBool(prmSection,'show_vip',ShowVIP);
end;

procedure TOpPrmPLSFact.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(ShowRedundancy,sizeof(ShowRedundancy));
 prmStream.ReadBuffer(ShowCoefAxis,sizeof(ShowCoefAxis));
 prmStream.ReadBuffer(ShowVIP,sizeof(ShowVIP)); 
end;

procedure TOpPrmPLSFact.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteBool(prmSection,'show_redundancy',ShowRedundancy);
 prmINI.WriteBool(prmSection,'show_coef_axis',ShowCoefAxis);
 prmINI.WriteBool(prmSection,'show_vip',ShowVIP);
end;

procedure TOpPrmPLSFact.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(ShowRedundancy,sizeof(ShowRedundancy));
 prmStream.WriteBuffer(ShowCoefAxis,sizeof(ShowCoefAxis));
 prmStream.WriteBuffer(ShowVIP,sizeof(ShowVIP));
end;

procedure TOpPrmPLSFact.SetDefaultParameters;
begin
 inherited;
 self.ShowRedundancy:= true;
 self.ShowCoefAxis:= false;
 self.ShowVIP:= false;
end;

initialization
 RegisterClass(TMLGenCompPLSFact);
end.
